--
--  Debug Logging
--
--===========================================================================--
--[[ @sdk
--- LRLib_DebugLog defines the debugging functions utilised by all of the Photographer's
-- Toolbox library functions.
--
	module 'LRLib_DebugLog' -- not actually executed, but suffices to trick LuaDocs
--]]
--============================================================================--

-- Load Photographer's Toolbox libraries if not already defined
require "LRLib_InterOp"

local LrPathUtils = import 'LrPathUtils'
local LrFileUtils = import 'LrFileUtils'
local LrView = import 'LrView'

require 'LRLib_TableEx'

local PROP
local PREFS

local pluginName = nil
local logFileName = nil
local callback = nil

LRLib_DebugLog = {}

--- Initialise the debug log.
-- Typically called as part of plugin initialisation.
-- @param name (string) Plugin name
-- @param logfilename (string) Log file name
-- @param modLogFn (function) Optional callback function to pre-process log entries
function LRLib_DebugLog.init(name, logfilename, modLogFn)
	pluginName = name
	logFileName = logfilename
	callback = modLogFn
end

--- Adds a Debugging section to control the debug log functionality from within the Plugin Info Provider.
-- @param pt (LrObservableTable) Preferences table to store results.
-- @param extraText (string) Optional extra text to be added after the introduction to this dialog section.
function LRLib_DebugLog.PluginInfoSection(pt, extraText)
	local f = LrView.osFactory()
	return {
		title = LOC "$$$/LRLib/DebugLog/PluginInfo/Title=Debugging",
		f:column {
			fill_horizontal = 1,
			spacing = f:control_spacing(),
			f:static_text {
				width = 512,
				height_in_lines = -1,
				title = LOC("$$$/LRLib/DebugLog/Text=If you're having issues with the PhotoShelter plugin we'll " ..
												"ask you to activate the debug logging.  This will save all sorts " ..
												"of useful information into a file for us to use in helping you through any problems.")
			},
			LrView.conditionalItem(extraText, f:static_text {
				width = 512,
				height_in_lines = -1,
				title = extraText
			}),
			f:column {
				fill_horizontal = 0,
				place_horizontal = 0.5,
				spacing = f:label_spacing(),
				f:radio_button {
					title = LOC "$$$/LRLib/DebugLog/Off=Do not log debug information",
					bind_to_object = pt,
					value = LrView.bind "LRLib_DebugLog_dest",
					checked_value = "none"
				},
				f:radio_button {
					title = LOC("$$$/LRLib/DebugLog/File=Log debug information to a file (^1) on your desktop", logFileName),
					bind_to_object = pt,
					value = LrView.bind "LRLib_DebugLog_dest",
					checked_value = "file"
				},
				f:radio_button {
					title = LOC "$$$/LRLib/DebugLog/Log=Send debug to system log",
					bind_to_object = pt,
					value = LrView.bind "LRLib_DebugLog_dest",
					checked_value = "log"
				}
			}
		}
	}
end

--- Cleans the log files and sets the default logging destination (if necessary) when starting the Plugin info Provider..
-- @param pt (LrObservableTable) Preferences table to store results.
-- @param extraText (string) Optional extra text to be added after the introduction to this dialog section.
function LRLib_DebugLog.startDialog(pt)
	DBG = function(msg, debug, lvl)
		if not lvl then
			lvl = 'INFO'
		end

		msg = lvl .. " - " .. msg

		if debug then
			LRLib_DebugLog.log(pt, msg)
			LRLib_DebugLog.logRaw(pt, LRLib_TableEx.dump(debug or {}))
		else
			LRLib_DebugLog.logRaw(pt, msg)
		end
	end

	LRLib_DebugLog.reset()

	local prefs = LRLib_InterOp.prefs()

	if prefs.LRLib_DebugLog_dest ~= nil then
		pt.LRLib_DebugLog_dest = prefs.LRLib_DebugLog_dest
	else
		pt.LRLib_DebugLog_dest = "none"
		prefs.LRLib_DebugLog_dest = "none"
	end
end

--- Updates the plugin preferences with the result from this dialog.
-- @param pt (LrObservableTable) Preferences table to store results.
-- @param why (string) Reason this function was called.
function LRLib_DebugLog.endDialog(pt, why)
	local prefs = LRLib_InterOp.prefs()
	prefs.LRLib_DebugLog_dest = pt.LRLib_DebugLog_dest
end

--- Deletes the log file if it exists.
function LRLib_DebugLog.reset()
	local errorFile = LrPathUtils.child(LrPathUtils.getStandardFilePath("desktop"), logFileName)

	if LrFileUtils.exists(errorFile) then
		LrFileUtils.delete(errorFile)
	end
end

--- Writes a string to the log.
-- Debugging method "log" relies upon the LRLib_Logger LOG() function.
-- Debugging method "file" writes to log file on the desktop.  If a callback
-- was specified during initialisation, this is used to pre-process the message
-- before being logged.
-- @param pt (LrObservableTable) Preferences table containing the debug log configuration.
-- @param msg (string)  Message to write to the debug log.
-- @see LRLib_Logger
function LRLib_DebugLog.logRaw(pt, msg)
	local prefs = LRLib_InterOp.prefs()

	if not prefs.LRLib_DebugLog_dest or prefs.LRLib_DebugLog_dest == "none" then
		return
	elseif prefs.LRLib_DebugLog_dest == "log" then
		LOG(msg)
	elseif prefs.LRLib_DebugLog_dest == "file" then
		if callback then
			msg = callback(pt, msg)
		end

		local errorFile = LrPathUtils.child(LrPathUtils.getStandardFilePath("desktop"), logFileName)
		local handle = io.open(errorFile, "a")

		handle:write(msg)
		handle:write('\n')
		handle:close()
	end
end

--- Writes a formatted string to the log
-- @param pt (LrObservableTable) Preferences table containing the debug log configuration.
-- @param ... (any) string.format() compatible arguments to be written to the log.
function LRLib_DebugLog.log(pt, ...)
	LRLib_DebugLog.logRaw(pt, string.format(...))
end
