--[[----------------------------------------------------------------------------

Copyright (c) 2010, Timothy Armes (http://www.timothyarmes.com)

All rights reserved.

See included file LICENSE.txt for details.

------------------------------------------------------------------------------]]

local LrColor = import 'LrColor'
local LrView = import 'LrView'
local LrApplication = import 'LrApplication'
local LrFunctionContext = import 'LrFunctionContext'
local LrBinding = import 'LrBinding'
local LrDialogs = import 'LrDialogs'

LrPrefs = import 'LrPrefs'

require "LRLib_Logger"
require "LRLib_InterOp"

LRLib_DialogUtils = {}
LRLib_DialogUtils.canExportList = {}

local prefs = LRLib_InterOp.prefs()

-- Centralise the management of the LR_canExport property

function LRLib_DialogUtils.globalCanExport(pt, section, canExport, because)
   
   local k, v
   
   -- Set the property for the module
   LRLib_DialogUtils.canExportList[section] = { canExport = canExport, because = because }

   -- Return the AND of all the properties
   for k, v in pairs(LRLib_DialogUtils.canExportList) do

      if not v.canExport then 
         pt.LR_canExport = false
         pt.LR_cantExportBecause = v.because
         return
      end
   end
 
   pt.LR_canExport = true
	pt.LR_cantExportBecause = nil

end

-- The the given variable and return the exiftool options if it's not empty

function LRLib_DialogUtils.testAndQuote(item, tag)
   if not item or #item == 0 then return "" end
   
   if not tag then
      if WIN_ENV == true then           
         return item..'\n'
      else
         return item..' '
      end
   else
      if WIN_ENV == true then           
         return "-"..tag..'='..item..'\n'
      else
         return "-"..tag..'="'..item..'" '
      end
   end
end

-- Save the given preset fields in the prefs.
--
-- Used in conjunction with LRLib_DialogUtils.loadSettings to reload the previous
-- settings when switching export plug-ins since LR doesn't do this (a bug)

function LRLib_DialogUtils.saveSettings(pt, section, presets)

   -- Save the changed settings in case LR loses them
   
   local lastSettingsUsed = {}
   for k, v in pairs(presets) do
      lastSettingsUsed[v.key] = pt[v.key]
   end

   prefs[section .. "_lastSettingsUsed"] = lastSettingsUsed
end

function LRLib_DialogUtils.loadSettings(pt, section, presets)

   if prefs[section .. "_lastSettingsUsed"] ~= nil then
      for k, v in pairs(prefs[section .. "_lastSettingsUsed"]) do
         if v ~= nil then pt[k] = v end
      end      
   end
   
end

-- Display a field in red if it isn't a number

function LRLib_DialogUtils.LrColorValidatedNumber(value)
   return LrView.bind {
      key = value,
      transform = function(value, fromTable)
         if tonumber(value) ~= nil then return LrColor(0, 0, 0)
         else return LrColor(1, 0, 0) end
      end
   }
end

-- Used by post-process plug-in to detect and avert the user whenever a 
-- panel is closed (which removes *all* observers!)

function LRLib_DialogUtils.startPostProcessSection(pt)

   if not pt.broken or LRLib_InterOp.majorVersion >= 3  then
      pt.healthy = true
      pt.broken = false
      pt.numClosed = 0
   end
   
end

function LRLib_DialogUtils.endPostProcessSection(pt)
   
   -- It's fixed in LR3
   if LRLib_InterOp.majorVersion >= 3 then return end
   
   -- Closing a post process section will tickle a Lightroom bug that removes ALL the bindings for all the
   -- post process sections of this plugin.  Once section is closed, we consider that the bug has occurred.
   
   pt.numClosed = pt.numClosed + 1
   pt.healthy = false
   pt.broken = true
   
end


function LRLib_DialogUtils.brokenCheck(f, view)

   local bind = LrView.bind
   local share = LrView.share

   return f:view {
   
      fill_horizontal = 1,
   
      place = "overlapping",
      f:view {
         visible = bind "healthy",
         fill_horizontal = 1,
         view,
      },
      f:view {
         visible = bind "broken",
         fill_horizontal = 1,
         place_vertical = 0.5,
         f:column {
            fill_horizontal = 1,
            
            f:static_text {
               width = 500,
               height_in_lines = -1,
               font = "<system/bold>",
               alignment = "center",
               text_color = LrColor(1, 0, 0),
               title = LOC "$$$/LRPPPBug/Title=Please close and reopen the export dialog to continue"
            },
            
            f:spacer {
               height = 20,
            },
            
            f:static_text {
               width = 500,
               height_in_lines = -1,
               alignment = "center",
               title = LOC "$$$/LRPPPBug/Title=You've just tickled a Lightroom bug that can't be avoided.  You'll need to close and reopen the export dialog box to continue.  Don't panic, you're settings won't be lost."
            },
         }
      }
   }
   
end

-- Helper function -- given a table of { value = ... } tables, return true
-- if the given value is found in one of the subtable's 'value' entries,
-- false otherwise.

function LRLib_DialogUtils.popupmenuItemsHasValue(t, value)
   
   if not value then
      return false
   end

   for i, set in ipairs(t) do
      if set.value == value then
         return true
      end
   end

   return false
end

-- Helper function -- given a table of { title = ...., value = ... } tables, return the 'title'
-- from the table with the given 'value'. Returns nil if not found.
 
function LRLib_DialogUtils.popupmenuTitleFromValue(t, value)
 
   if not value then
      return nil
   end

   for i, set in ipairs(t) do
      if set.value == value then
         return set.title
      end
   end
   return nil
end

-- Popup menu items might have "&amp;", "&lt;", "&gt;" that need to be decoded.
-- Also, all remaining ampersands should be doubled, because LrView strips single ampersands.

function LRLib_DialogUtils.popupmenuTidyItems(items)

   for k, v in ipairs(items) do
      local text = v.title
      text = text:gsub('&lt;',  '<')
      text = text:gsub('&gt;',  '>')
      text = text:gsub('&amp;', '&')
      text = text:gsub('&',     '&&')
      v.title = text
   end

end

-- A do not show again dialog without the silly cancel button

function LRLib_DialogUtils.getStoredResultForpromptForActionWithDoNotShow(actionPrefKey)
   return prefs["LRLib_DialogUtils_dontshow_" .. actionPrefKey]
end

function LRLib_DialogUtils.resetDoNotShowFlag(actionPrefKey)
   prefs["LRLib_DialogUtils_dontshow_" .. actionPrefKey] = nil
end

function LRLib_DialogUtils.promptForActionWithDoNotShow(args)

   local result
   local catalog = LrApplication.activeCatalog()

   LrFunctionContext.callWithContext("newPresetDialog", function(context)
   
      local f = LrView.osFactory()
      local pt = LrBinding.makePropertyTable(context)
      
      local stored = LRLib_DialogUtils.getStoredResultForpromptForActionWithDoNotShow(args.actionPrefKey)
      if stored then
         result = stored
         return
      end
      
      pt.dont = false
            
      -- Set up the accessory view
      
      local checkbox = f:checkbox {
         bind_to_object = pt,
         title = LOC "$$$/LRLib/DialogUtils/DontShow=Don't show again",
         value = LrView.bind("dont")
      }
      
      local row = {
         fill_horizontal = 1,
         checkbox,
         f:spacer {
            fill_horizontal = 1
         }
      }
      
      local n = 2
      while args.verbBtns[n] do
   
         local button = args.verbBtns[n]
         table.insert(row, 
            f:push_button {
               title = button.label,
               action = function()
                  LrDialogs.stopModalWithResult(checkbox, button.verb)
               end
            }
         )
   
         n = n + 1
      end
   
      result = LrDialogs.presentModalDialog {
         title = "",
         resizable = false,
         contents = f:column {
            spacing = f:control_spacing(),
            f:static_text {
               width = 500,
               height_in_lines = -1,
               title = args.message
            },
            f:static_text {
               width = 500,
               height_in_lines = -1,
               title = args.info,
               size = "small"
            }
         },
         cancelVerb = "< exclude >", -- this magic, unpublished setting removes the cancel button, and may not work in the future
         actionVerb = args.verbBtns[1].label,
         actionBinding = args.verbBtns[1].verb,
         accessoryView = f:row(row)
      }
      
      if result == "ok" then result = args.verbBtns[1].verb end
      
      if pt.dont then
         prefs["LRLib_DialogUtils_dontshow_" .. args.actionPrefKey] = result
      end
      
   end)
   
   return result
      
end

