--
-- Functions that vary between different versions of LR
--
--===========================================================================--
--[[ @sdk
--- LRLib_InterOp defines the version interoperability functions utilised by all of the Photographer's
-- Toolbox library functions.
--
	module 'LRLib_InterOp' -- not actually executed, but suffices to trick LuaDocs
--]]
--============================================================================--

local LrApplication = import 'LrApplication'
local LrPrefs = import 'LrPrefs'

LRLib_InterOp = {}

local majorVersion
if LrApplication.versionTable then
   majorVersion = LrApplication.versionTable().major
   minorVersion = LrApplication.versionTable().minor
else
   majorVersion = 1
   minorVersion = 3
end

local catalog

local lr3RawFields = {
   countStackInFolderMembers = 3,
   countVirtualCopies = 3,
   isInStackInFolder = 3,
   isVirtualCopy = 3,
   masterPhoto = 3,
   path = 3,
   stackInFolderMembers = 3,
   stackPositionInFolder = 3,
   topOfStackInFolderContainingPhoto = 3,
   uuid = 3,
   virtualCopies = 3
}

--- LRLib_InterOp properties
-- @name LRLib_InterOp
-- @class table
-- @field majorVersion (read only) Major Lightroom version
-- @field minorVersion (read only) Minor Lightroom version
LRLib_InterOp.majorVersion = majorVersion
LRLib_InterOp.minorVersion = minorVersion

-- Get RawMetaData from the right place depending on the LR version

--- Retrieves the value of the raw metadata field.
-- @param photo (LrPhoto) Photo being queried
-- @param field (string) Raw metadata field to query
-- @return (any) Value for the field being queried
function LRLib_InterOp.getRawMetadata(photo, field)

   if lr3RawFields[field] and lr3RawFields[field] <= majorVersion then
      return photo:getRawMetadata(field)
   else
      local result
      LRLib_InterOp.withReadAccessDo(function()
         result = photo[field]
      end)
      
      return result
   end

end

-- Catalog functions

--- Retrieves the target photo.
-- @return (LrPhoto) Target photo
function LRLib_InterOp.getTargetPhoto() 

   if not catalog then catalog = LrApplication.activeCatalog() end

   if majorVersion >= 3 then
      return catalog:getTargetPhoto()
   else
      local photo
      if not catalog.hasReadAccess then
         catalog:withCatalogDo(function()
            photo = catalog.targetPhoto
         end)
      else
         photo = catalog.targetPhoto
      end
      return photo
   end

end

--- Retrieves the target photos.
-- @return (array of LrPhoto) Target photos
function LRLib_InterOp.getTargetPhotos() 

   if not catalog then catalog = LrApplication.activeCatalog() end

   if majorVersion >= 3 then
      return catalog:getTargetPhotos()
   else
      local photo
      if not catalog.hasReadAccess then
         catalog:withCatalogDo(function()
            photo = catalog.targetPhotos
         end)
      else
         photo = catalog.targetPhotos
      end
      return photo
   end

end

--- Retrieves the path for the current catalog.
-- @return (string) Catalog path
function LRLib_InterOp.getCatalogPath() 

   if not catalog then catalog = LrApplication.activeCatalog() end
   
   if majorVersion >= 3 then
      return catalog:getPath()
   else
      return catalog.getPath
   end

end

--- Executes a function requiring read access to the catalog.
-- @param fn (function) Function that reads the catalog
function LRLib_InterOp.withReadAccessDo(fn)

   if majorVersion >= 3 then
      fn()
      return
   end
   
   if majorVersion == 2 then
      if not catalog then catalog = LrApplication.activeCatalog() end
   
      if not catalog.hasReadAccess then
         catalog:withReadAccessDo(fn)
      else
         fn()
      end
      
      return
   end
   
   -- LR 1.x
   catalog:withCatalogDo(fn)
   
end

--- Retrieves the preferences for this plugin.
-- @return (LrObservableTable) Preferences for the plugin
function LRLib_InterOp.prefs()

   if majorVersion >= 3 then 
      return LrPrefs.prefsForPlugin()
   else
      return LrPrefs.prefsForPlugin(_PLUGIN)
   end
   
end   