--[[----------------------------------------------------------------------------

Copyright (c) 2007, Timothy Armes (http://www.timothyarmes.com)

All rights reserved.

See included file LICENSE.txt for details.

------------------------------------------------------------------------------]]

-- Principles to bear in mind
--
-- The scroller maintains a list of all the fields for each row in 'mappings'.
-- The exportPresetFields contain serialised versions of 'mappings'.
-- The fields are now reserialised for every change made, only the serialised versions are saved.
--     For this reason all edit fields should be "immmediate"

local LrTasks = import "LrTasks"
local LrPathUtils = import 'LrPathUtils'
local LrView = import 'LrView'

require "LRLib_Serialise"

-- Namespace

LRLib_Scroller = {}

-- Create a new generic scoller
--
-- baseid             : prefix for the fields stored in presets
-- fields             : table of field names and default values, like preset field table, without prefixes
-- createRowFn        : Callback function to create a row of the dialog, takes form fn(factory, scroller, row_index)
--                      Note that all edit fields should be immediate
-- numVisibleMappings : number of visible items

function LRLib_Scroller:new(baseid, fields, createRowFn, numVisibleMappings)

   local o = {}

   o.baseid = baseid
   o.numVisibleMappings = numVisibleMappings
   o.fields = fields
   o.atLeastOne = false
   o.createRowFn = createRowFn
   o.maxRows = 0
   o.isDynamic = true
   o.canAddRows = true
   o.displayControls = true
   o.displayQuickMenu = false
   o.scrollBarAtSide = false
   o.headerRowOnTop = false
   o.horizontal = false
   o.customIndexDisplayCallback = LRLib_Scroller.defaultIndexDisplay
   o.indexClickCallback = nil
   o.customQuickMenuIndexDisplayCallback = LRLib_Scroller.defaultQuickMenuIndexDisplay
   o.customQuickMenuWidthInCharsCallback = LRLib_Scroller.defaultQuickMenuWidthInCharsCallback
   
   o.deleteButtonText = LOC "$$$/LRLib/Scroller/Delete=-"
   o.upArrowActive = LrPathUtils.child(_PLUGIN.path, (LrPathUtils.child("LRLib_Scroller", "LRLib_Scroller_UpActive.png")))
   o.upArrowGrey = LrPathUtils.child(_PLUGIN.path, (LrPathUtils.child("LRLib_Scroller", "LRLib_Scroller_UpGrey.png")))
   o.downArrowActive = LrPathUtils.child(_PLUGIN.path, (LrPathUtils.child("LRLib_Scroller", "LRLib_Scroller_DownActive.png")))
   o.downArrowGrey = LrPathUtils.child(_PLUGIN.path, (LrPathUtils.child("LRLib_Scroller", "LRLib_Scroller_DownGrey.png")))
   o.xItemsTextPosition = 0.5

   o.setPropertyTable = LRLib_Scroller.setPropertyTable
   o.exportPresetFields = LRLib_Scroller.exportPresetFields
   o.saveVisible = LRLib_Scroller.saveVisible
   o.loadVisible = LRLib_Scroller.loadVisible
   o.startDialog = LRLib_Scroller.startDialog
   o.endDialog = LRLib_Scroller.endDialog
   o.setDynamic = LRLib_Scroller.setDynamic
   o.setHeaderOnTop = LRLib_Scroller.setHeaderOnTop
   o.setScrollBarAtSide = LRLib_Scroller.setScrollBarAtSide
   o.setCanAddRows = LRLib_Scroller.setCanAddRows
   o.addRow = LRLib_Scroller.addRow
   o.setDisplayControls = LRLib_Scroller.setDisplayControls
   o.setDisplayQuickMenu = LRLib_Scroller.setDisplayQuickMenu  
   o.setHorizontal = LRLib_Scroller.setHorizontal
   o.refreshQuickMenu = LRLib_Scroller.refreshQuickMenu
   o.deleteMapping = LRLib_Scroller.deleteMapping
   o.scroll = LRLib_Scroller.scroll
   o.sectionForDialog = LRLib_Scroller.sectionForDialog
   o.bindID = LRLib_Scroller.bindID
   o.bindIDForMappings = LRLib_Scroller.bindIDForMappings
   o.bindIDForNumberOfMappings = LRLib_Scroller.bindIDForNumberOfMappings
   o.bindIDForScrollOffset = LRLib_Scroller.bindIDForScrollOffset
   o.setMaxRows = LRLib_Scroller.setMaxRows
   o.getRows = LRLib_Scroller.getRows
   o.setRows = LRLib_Scroller.setRows
   o.getNumRows = LRLib_Scroller.getNumRows
   o.setNumRows = LRLib_Scroller.setNumRows
   o.getScrollOffset = LRLib_Scroller.getScrollOffset
   o.setScrollOffset = LRLib_Scroller.setScrollOffset
   o.startFieldUpdate = LRLib_Scroller.startFieldUpdate
   o.endFieldUpdate = LRLib_Scroller.endFieldUpdate
   o.isUpdating = LRLib_Scroller.isUpdating
   o.setCustomIndexDisplayCallback = LRLib_Scroller.setCustomIndexDisplayCallback
   o.setIndexClickCallback = LRLib_Scroller.setIndexClickCallback
   o.setQuickMenuIndexDisplayCallback = LRLib_Scroller.setQuickMenuIndexDisplayCallback
   o.setQuickMenuWidthInCharsCallback = LRLib_Scroller.setQuickMenuWidthInCharsCallback
   o.refreshIndexDisplay = LRLib_Scroller.refreshIndexDisplay
   
   return o

end

-- Set the property table

function LRLib_Scroller:setPropertyTable(pt)
   self.pt = pt
   
   if not pt[self.baseid .. '_refresh'] then
      pt[self.baseid .. '_refresh'] = 0
   end
   
end  

-- Bind id for a field in a row

function LRLib_Scroller:bindID(field, row)
   if row == nil then MSG("Scroller error: empty 'row' value passed to bindID for field '"..field.."'") end
      return self.baseid .. "_mapping_" .. row .. "_" ..field
end

function LRLib_Scroller:bindIDForMappings()
   return self.baseid .. '_serialised_mappings'
end

function LRLib_Scroller:bindIDForNumberOfMappings()
   return self.baseid .. '_num_mappings'
end

function LRLib_Scroller:bindIDForScrollOffset()
   return self.baseid .. '_scrollOffset'
end

-- Access to other scroller variables

function LRLib_Scroller:getRows()
   local success, theError = LrTasks.pcall(loadstring("scrollerMappings = " .. self.pt[self.baseid .. '_serialised_mappings']))
   if not success then
      MSG(theError)
      scrollerMappings = {}
   end
   
   return scrollerMappings
end

function LRLib_Scroller:setRows(mappings)
   self:startFieldUpdate()
   self.pt[self.baseid .. '_mappings'] = mappings
   self.pt[self.baseid .. '_serialised_mappings'] = LRLib_Serialise.serialise(mappings)
   self:setNumRows(#mappings)
   self:endFieldUpdate()
   self:loadVisible()
end

function LRLib_Scroller:getNumRows()
   return self.pt[self.baseid .. '_num_mappings']
end

function LRLib_Scroller:setNumRows(rows)
  self.pt[self.baseid .. '_num_mappings'] = rows
end

function LRLib_Scroller:setMaxRows(rows)
   self.maxRows = rows
end

function LRLib_Scroller:setDynamic(isDynamic)
   self.isDynamic = isDynamic
end

function LRLib_Scroller:setHeaderOnTop(onTop)
   self.headerRowOnTop = onTop
end

function LRLib_Scroller:setScrollBarAtSide(atSide)
   self.scrollBarAtSide = atSide
end

function LRLib_Scroller:setCanAddRows(canAddRows)
   self.canAddRows = canAddRows
end

function LRLib_Scroller:setDisplayControls(shouldDisplay)
   self.displayControls = shouldDisplay
end

function LRLib_Scroller:setDisplayQuickMenu(shouldDisplay)
   self.displayQuickMenu = shouldDisplay
end

function LRLib_Scroller:setHorizontal()
   self.horizontal = true
end

function LRLib_Scroller:getScrollOffset()
   return self.pt[self.baseid .. '_scrollOffset']
end

function LRLib_Scroller:setScrollOffset(offset)
  self.pt[self.baseid .. '_scrollOffset'] = offset
end

function LRLib_Scroller:exportPresetFields()

   local defaultNumMappings = 0
   if self.atLeastOne then defaultNumMappings = 1 end

	return {
	   { key = self.baseid .. '_num_mappings', default = defaultNumMappings },					
		{ key = self.baseid .. '_serialised_mappings', default = "{}" },					
		{ key = self.baseid .. '_scrollOffset', default = 0 },					
   }

end

-- Functions to allow the caller to change the field contents

function LRLib_Scroller:startFieldUpdate()
   self.pt[self.baseid .. '_internalUpdate'] = true
end

function LRLib_Scroller:endFieldUpdate()
   self.pt[self.baseid .. '_internalUpdate'] = false
end

function LRLib_Scroller:isUpdating()
   return self.pt[self.baseid .. '_internalUpdate'] or self.pt[self.baseid .. '_loading']
end

-- Setting up call backs

function LRLib_Scroller:setCustomIndexDisplayCallback(callback)
   self.customIndexDisplayCallback = callback
end

function LRLib_Scroller:setIndexClickCallback(callback)
   self.indexClickCallback = callback
end

function LRLib_Scroller:setQuickMenuIndexDisplayCallback(callback)
   self.customQuickMenuIndexDisplayCallback = callback
end

function LRLib_Scroller:setQuickMenuWidthInCharsCallback(callback)
   self.customQuickMenuWidthInCharsCallback = callback
end

-- Default callbacks

function LRLib_Scroller.defaultIndexDisplay(index)
   return ""..index
end

function LRLib_Scroller.defaultQuickMenuIndexDisplay(index)
   return ""..index
end

function LRLib_Scroller.defaultQuickMenuWidthInCharsCallback()
   return 3
end

function LRLib_Scroller:refreshQuickMenu()
   
   local items = {}
   local row = 1
   while row <= self:getNumRows() do
      items[row] = { title = self.customQuickMenuIndexDisplayCallback(row), value = row }
      row = row + 1
   end
   
   self.pt[self.baseid .. "_quickMenu"] = items
end

function LRLib_Scroller:refreshIndexDisplay()
   if self.pt[self.baseid .. '_refresh'] then
      self.pt[self.baseid .. '_refresh'] = self.pt[self.baseid .. '_refresh'] + 1
   else
      self.pt[self.baseid .. '_refresh'] = 0
   end
end

function LRLib_Scroller:saveVisible()

   local mappings = self.pt[self.baseid .. '_mappings']
   local n = 1
   
   while n + self:getScrollOffset() <= self:getNumRows() and n <= self.numVisibleMappings do
      
      mappings[n + self:getScrollOffset()] = {}
      for k, v in pairs(self.fields) do
         mappings[n + self:getScrollOffset()][v.key] = self.pt[self:bindID(v.key, n)]
      end
      
      n = n + 1
   end

	-- Serialise the settings now so that if the user saves a preset then there'll be no problems
	--
	-- Mark the fact that we're saving so that LR doesn't react to the change to serialised_mappings

   self.pt[self.baseid .. '_saving'] = true
	self.pt[self.baseid .. '_serialised_mappings'] = LRLib_Serialise.serialise(mappings) 
   self.pt[self.baseid .. '_saving'] = false
   
end

-- Load the mapping into the visible area

function LRLib_Scroller:loadVisible()

   if self.pt[self.baseid .. '_internalUpdate'] == true then return end

   -- Load the available info

   self.pt[self.baseid .. '_loading'] = true

   local mappings = self.pt[self.baseid .. '_mappings']
   local n = 1
   
   while n + self:getScrollOffset() <= self:getNumRows() and n <= self.numVisibleMappings do
      if mappings[n + self:getScrollOffset()] then
         for k, v in pairs(self.fields) do
            if type(mappings[n + self:getScrollOffset()][v.key]) ~= "nil" then
               self.pt[self:bindID(v.key, n)] = mappings[n + self:getScrollOffset()][v.key]
            else
               self.pt[self:bindID(v.key, n)] = v.default
            end
         end
      else
         for k, v in pairs(self.fields) do
            self.pt[self:bindID(v.key, n)] = v.default
         end
      end
         
      n = n + 1
   end

   -- Clear the remaining ones

   while n <= self.numVisibleMappings do
      for k, v in pairs(self.fields) do
         self.pt[self:bindID(v.key, n)] = v.default
      end
      n = n + 1
   end

   self.pt[self.baseid .. '_loading'] = false

end

function LRLib_Scroller:startDialog(pt)

   self.pt = pt

   self.pt[self.baseid .. '_internalUpdate'] = false
   self.pt[self.baseid .. '_saving']   = false
   self.pt[self.baseid .. '_loading']  = false

   -- Set up the default parameters
      
   self.pt[self.baseid .. '_scrollOffset'] = 0
   self.pt[self.baseid .. '_mappings'] = {}

	-- Check for changes to the mappings that can result from the user double clicking on a preset
   --
   -- The important thing is that pt.mappings is kept in sync with the serialised_mappings, so we
   -- need to react to changes caused by a preset being reloaded.
   --
   -- Serialised settings is always up to date, so be reload it
	
	local function reloadSettings()

      if self.pt[self.baseid .. '_internalUpdate'] == true or self.pt[self.baseid .. '_saving'] == true then return end

      -- Load the serialised mappings.  We have to do this every time since they may have been changed behind our back
      -- by the user loading a preset

      if not self.pt[self.baseid .. '_serialised_mappings'] or self.pt[self.baseid .. '_serialised_mappings'] == "" then
         self.pt[self.baseid .. '_mappings'] = {}
      else
         local success, theError = LrTasks.pcall(loadstring("mappings = " .. self.pt[self.baseid .. '_serialised_mappings']))

         if not success then
            MSG(theError)
            self.pt[self.baseid .. '_mappings'] = {}
         else
            self.pt[self.baseid .. '_mappings'] = mappings
         end
      end
      
      self:loadVisible()
      self:refreshQuickMenu()

   end
   
	-- Load any previous settings

	reloadSettings()
	
	-- React to presets being loaded
   
   self.pt:addObserver(self.baseid .. '_serialised_mappings', reloadSettings)
	self.pt:addObserver(self.baseid .. '_num_mappings', reloadSettings)
		
	-- Update the serialised setting for every change made to a field
	
	for n = 1, self.numVisibleMappings, 1 do
   	for k, v in pairs(self.fields) do
   	   self.pt:addObserver(self:bindID(v.key, n), function()
   	      if self.pt[self.baseid .. '_internalUpdate'] == true or self.pt[self.baseid .. '_loading'] == true then return end
   	      self:saveVisible()
   	   end)             
      end
   end
    
end

function LRLib_Scroller:endDialog(pt)
end

-- Delete a mapping

function LRLib_Scroller:deleteMapping(idx)

   if self.atLeastOne and self:getNumRows() == 1 then return end

   self.pt[self.baseid .. '_internalUpdate'] = true

   local mappings = self.pt[self.baseid .. '_mappings']
   local i = idx + self:getScrollOffset()
   
   while i < self:getNumRows() do
      mappings[i] = mappings[i + 1]
      i = i + 1
   end

   mappings[i] = nil
   
   -- Serialise the settings now so that the load visible will work later on
   self.pt[self.baseid .. '_serialised_mappings'] = LRLib_Serialise.serialise(mappings) 
   
   if idx + self:getScrollOffset() <= self:getNumRows() then
      self:setNumRows(self:getNumRows() - 1)
   end

	-- Scroll up if necessary

	if self:getNumRows() - self:getScrollOffset() < self.numVisibleMappings then
		local offset = self:getNumRows() - self.numVisibleMappings
		if offset < 0 then offset = 0 end
		self:setScrollOffset(offset)
	end
	
   -- Initialise the empty rows
   
   self.pt[self.baseid .. '_internalUpdate'] = false     
   
   -- Load the new visible mappings
   
   self:loadVisible()      
   
   -- Reserialise everything
   
   self:saveVisible()
   
   -- Set up the quick menu
   
   self:refreshQuickMenu()
         
end

-- Force a row to be added.  Useful if you're supplying your own add button

function LRLib_Scroller:addRow()

   if self.maxRows > 0 and self:getNumRows() >= self.maxRows then return end
   self:startFieldUpdate()
   self:setNumRows(self:getNumRows() + 1)
   local offset = self.pt[self.baseid .. '_num_mappings'] - self.numVisibleMappings
   if offset < 0 then offset = 0 end	
   self:setScrollOffset(offset)
   self:endFieldUpdate()
   
   self:loadVisible()
   self:saveVisible()
   self:refreshQuickMenu()

end

-- Function to scroll the item list

function LRLib_Scroller:scroll(offset)

   local newOffset = self:getScrollOffset() + offset
   if newOffset < 0 then 
      newOffset = 0
   end

   -- Change the contents if the offset has moved
   
   if newOffset ~= self:getScrollOffset() then
      self:setScrollOffset(newOffset)
      self:loadVisible()
   end

end

function LRLib_Scroller:sectionForDialog(f, style, headerRow)

   local bind = LrView.bind
   local share = LrView.share
   
   local INDEX_WIDTH = 2
   
   -- Create the dialog contents
   
   local function mappingRow(r)
   
      -- Building blocks
      
      local function indexWithWidth(width)
         return f:static_text {
            title = bind {
               keys = {
                  { key = self.baseid .. '_scrollOffset' },
                  { key = self.baseid .. '_refresh' },
               },
               operation = function(binding, values, fromTable)
                 if fromTable then
                    return self.customIndexDisplayCallback(values[self.baseid .. '_scrollOffset'] + r)
                 else
                    return LrBinding.kUnsupportedDirection
                 end
              end,
            },
      		place_vertical = 0.5,
            width_in_chars = width,
            font = "<system/bold>",
         }
      end

      local function index()
         return f:static_text {
            title = bind {
               keys = {
                  { key = self.baseid .. '_scrollOffset' },
                  { key = self.baseid .. '_refresh' },
               },
               operation = function(binding, values, fromTable)
                  if fromTable then
                     return self.customIndexDisplayCallback(values[self.baseid .. '_scrollOffset'] + r)
                  else
                     return LrBinding.kUnsupportedDirection
                  end
              end,
            },
            mouse_down = function()
               if self.indexClickCallback then
                  self.indexClickCallback(self.pt[self.baseid .. '_scrollOffset'] + r)
               end
            end,
      		place_vertical = 0.5,
      		fill_horizontal = 1,
            font = "<system/bold>",
         }
      end
      
      local function deleteButton()
         return LrView.conditionalItem(self.isDynamic, f:push_button {
            title = self.deleteButtonText,
            enabled = bind {
               key = self.baseid .. '_num_mappings',
               transform = function(value, fromTable)
                  if fromTable then
                     if value == 1 and self.atLeastOne then return false end
                     return true
                  else
                     return LrBinding.kUnsupportedDirection
                  end
               end
            },
   			place_vertical = 0.5,
            width = share "addDeleteWidth",
            action = function ()
               self:deleteMapping(r)
            end
         })
      end
      
      -- Styles
      
      local function singleRow()
         return f:row {
            spacing = f:control_spacing(),
            fill_horizontal = 1,
      
            indexWithWidth(INDEX_WIDTH),
            self.createRowFn(f, self, r),		
            deleteButton()
         }
      end
   
      local function ornamentationOnTop()
         return f:column {
            spacing = f:control_spacing(),
            fill_horizontal = 1,
            
            f:row {
               spacing = f:control_spacing(),
               margin  = 4,
               fill_horizontal = 1,
               index(),
               LrView.conditionalItem(self.isDynamic, f:spacer { fill_horizontal = 1 }),
               deleteButton()
            },
   
            self.createRowFn(f, self, r),		
         }
      end
   
      local function noOrnamentation()
         return f:row {
            spacing = f:control_spacing(),
            fill_horizontal = 1,

            self.createRowFn(f, self, r),		
         }
      end
   
      local view = {
			
			spacing = f:label_spacing(),
			fill_horizontal = 1,
			margin_bottom = 4,

         visible = bind {
            keys = {
               { key = self.baseid .. '_scrollOffset' },
               { key = self.baseid .. '_num_mappings' },
            },
            operation = function(binding, values, fromTable)
               if r + values[self.baseid .. '_scrollOffset'] > values[self.baseid .. '_num_mappings'] then
                  return false
               else
                  return true
               end
            end
         }
		}
		
		-- Add row in required style
		
		if style == "ornamentationOnTop" then
		   table.insert(view, ornamentationOnTop())
		elseif style == "noOrnamentation" then
		   table.insert(view, noOrnamentation())				
		else
		   table.insert(view, singleRow())		
		end
		
		return f:view(view)
   
   end

   -- Prepare the header row
   
   local header = {
      spacing = f:control_spacing(),
   }
   
   if headerRow then
   
      -- Add the index number, but make it invisible
   
      if style == "ornamentationOnTop" then
		elseif style == "noOrnamentation" then
		else
		   table.insert(header, f:static_text {
		      visible = false,
            title = "1",
            width_in_chars = INDEX_WIDTH,
            font = "<system/bold>",
		   })	
		end
   
      -- The actual header row
   
      table.insert(header, headerRow)
      
      -- Add an invisible delete button

      if style == "ornamentationOnTop" then
		elseif style == "noOrnamentation" then
		else
		   table.insert(header, f:view {
		      LrView.conditionalItem(self.isDynamic, f:push_button {
		         visible = false,
               title = self.deleteButtonText,
      			place_vertical = 0.5,
               width = share "addDeleteWidth",
            })
         })
		end

      header = f:row {
         fill_horizontal = 1,
         margin = 4,
         spacing = f:label_spacing(),
         f:row(header)
      }  
   
   end

   -- Prepare the mapping rows
         
   local n = 1
   local mappings = {

      fill_horizontal = 1,
      margin = 4,
      spacing = f:label_spacing(),
      
   }

   -- Add the rows
   
   while n <= self.numVisibleMappings do
      table.insert(mappings,  mappingRow(n))
      n = n + 1
   end
   
   local mappingsView
   if self.horizontal then
      mappingsView = f:row(mappings)
   else
      mappingsView = f:view(mappings)
   end

   -- Put it all together

   local quickMenu = {}
   if self.displayQuickMenu then
   
      quickMenu = {
         value = bind {
            key = self.baseid .. '_scrollOffset',
            transform = function(value, fromTable)
               if fromTable then
                  return value + 1
               else
                  local offset = value - 1
                  if self:getNumRows() - value < self.numVisibleMappings then
                     offset = self:getNumRows() - self.numVisibleMappings
                     if offset < 0 then offset = 0 end
                  end
                  self:setScrollOffset(offset)
                  self:loadVisible()
                  return offset
               end
            end
         },
         enabled = LrBinding.keyIsNot(self.baseid .. '_num_mappings', 0),
         items = bind(self.baseid .. "_quickMenu")
      }
      
      local width = self.customQuickMenuWidthInCharsCallback()
      if width == -1 then
         quickMenu.fill_horizontal = 1
      else
         quickMenu.width_in_chars  = width
      end
   end

   local function upArrow()
   
      return f:view {
      
         place = "overlapping",
         place_vertical = 0.5,
         visible = self.maxRows == 0 or self.numVisibleMappings < self.maxRows,
         
         f:picture {
            value = self.upArrowActive,
            frame_width = 0,
            
            visible = bind {
              key = self.baseid .. '_scrollOffset',
              transform = function (value, fromTable)
                 if fromTable then
                    return value > 0
                 else
                    return LrBinding.kUnsupportedDirection
                 end
              end,
            },
            
            mouseDown = function (dummy, event)
               if event.eventType == "leftMouseDown" then
                  self:scroll(-1)
               end
            end
         },
         
         f:picture {
            value = self.upArrowGrey,
            frame_width = 0,
            
            visible = bind {
              key = self.baseid .. '_scrollOffset',
              transform = function (value, fromTable)
                 if fromTable then
                    return value <= 0
                 else
                    return LrBinding.kUnsupportedDirection
                 end
              end,
            }
         },
      }
   end
      
   local function downArrow()
   
      return f:view {
   
         place = "overlapping",
         place_vertical = 0.5,
         visible = self.maxRows == 0 or self.numVisibleMappings < self.maxRows,
         
         f:picture {
            value = self.downArrowActive,
            frame_width = 0,
            
            visible = bind {
               keys = {
                  { key = self.baseid .. '_scrollOffset' },
                  { key = self.baseid .. '_num_mappings' },
               },
               operation = function(binding, values, fromTable)
                  if values[self.baseid .. '_num_mappings'] - values[self.baseid .. '_scrollOffset'] <= self.numVisibleMappings then
                     return false
                  else
                     return true
                  end
               end
            },
            
            mouseDown = function (dummy, event)
               if event.eventType == "leftMouseDown" then
                  self:scroll(1)
               end
            end
         },
         
         f:picture {
            value = self.downArrowGrey,
            frame_width = 0,
            
            visible = bind {
               keys = {
                  { key = self.baseid .. '_scrollOffset' },
                  { key = self.baseid .. '_num_mappings' },
               },
               operation = function(binding, values, fromTable)
                  if values[self.baseid .. '_num_mappings'] - values[self.baseid .. '_scrollOffset'] <= self.numVisibleMappings then
                     return true
                  else
                     return false
                  end
               end
            },
         },
      }

   end
      
   local view = { 
    
      fill_horizontal = 1,
      spacing = f:control_spacing(),

      LrView.conditionalItem(headerRow and self.headerRowOnTop, header),      
      
      -- Buttons
            
      LrView.conditionalItem(self.displayControls, f:view {

         place = "overlapping",
         fill_horizontal = 1,
         margin = 4,
         
         f:row {	
               
            -- Pulldown menu for quick access 
            
            LrView.conditionalItem(self.displayQuickMenu, f:popup_menu(quickMenu)),        

            -- The up arrow
            
            LrView.conditionalItem(self.scrollBarAtSide == false, upArrow()),
            
            -- The down arrow
            
            LrView.conditionalItem(self.scrollBarAtSide == false, downArrow()),
         },
         
         -- Total number text

         f:row {
                           
            LrView.conditionalItem(self.isDynamic, f:static_text {
               place_horizontal = self.xItemsTextPosition,
               font = "<system/small/bold>",
               title = bind {
                  key = self.baseid .. '_num_mappings',
                  transform = function (value, fromTable)
                     if fromTable then 
                        return LOC ("$$$/LRLib/Scroller/Total/Defined=^1 items", value)
                     else
                        return LrBinding.kUnsupportedDirection
                     end
                  end,
               },
            })            
           
         },
         
         -- Add button

         LrView.conditionalItem(self.isDynamic and self.canAddRows, f:push_button {
            title = LOC "$$$/LRLib/Scroller/Add=+",
            width = share "addDeleteWidth",
            place_horizontal = 1,
            enabled = bind {
               key = self.baseid .. '_num_mappings',
               transform = function (value, fromTable)
                  if fromTable then
                     if self.maxRows == 0 then return true end
                     return self:getNumRows() < self.maxRows
                  else
                     return LrBinding.kUnsupportedDirection
                  end
               end
            },
            action = function ()
               self:addRow()
            end
         }),
      }),
      
      LrView.conditionalItem(headerRow and not self.headerRowOnTop, header),

   }
   
   if self.horizontal then
   
      table.insert(view, f:row {
   
         fill_horizontal = 1,
         spacing = f:control_spacing(),

         mappingsView,
         
      })
   
   else
   
      table.insert(view, f:row {
   
         fill_horizontal = 1,
         spacing = f:control_spacing(),

         mappingsView,
         
         -- If the scroll bar is at the side, dispaly it
         
         LrView.conditionalItem(self. scrollBarAtSide, f:column {
         
            fill_vertical = 1,
         
            upArrow(),
            f:spacer { fill_vertical = 1 },
            downArrow(),
         
         })
      })
      
   end
      
   return f:view(view)
   
end
