--[[----------------------------------------------------------------------------

Copyright (c) 2010, Timothy Armes (http://www.timothyarmes.com)

All rights reserved.

See included file LICENSE.txt for details.

------------------------------------------------------------------------------]]

require "LRLib_Logger"

LRLib_TableEx = {}

-- Merge tables together

function LRLib_TableEx.merge(...)
   local res = {}
   for t = 1, select('#', ...) do
      for i, v in pairs(select(t, ...)) do
         if type(i) == "number" then
            table.insert(res, v)
         else
            res[i] = v
         end
      end
   end
   return res
end

function LRLib_TableEx.clone(t)
  local new = {}             -- create a new table
  local i, v = next(t, nil)  -- i is an index of t, v = t[i]
  while i do
    if type(v) == "table" then v = LRLib_TableEx.clone(v) end 
    new[i] = v
    i, v = next(t, i)        -- get next index
  end
  return new
end

function LRLib_TableEx.deepCopy(object)
  local lookup_table = {}
  local function _copy(object)
      if type(object) ~= "table" then
          return object
      elseif lookup_table[object] then
          return lookup_table[object]
      end
      local new_table = {}
      lookup_table[object] = new_table
      for index, value in pairs(object) do
          new_table[_copy(index)] = _copy(value)
      end
      return new_table
  end
  return _copy(object)
end

function LRLib_TableEx.safeTableSort(T)
  return table.sort(T, function(a,b)
                          if type(a) == type(b) then
                             return a < b
                          else
                             return tostring(a) < tostring(b)
                          end
                       end)
end

local tables_dumped = {}

local function table_empty(tbl)
  if not tbl then
     return true
  else
     return not next(tbl)
  end
end

local function dumpTable(level, item, baselevel, toskip)

  if type(item) == 'string' then
     return '"' .. item .. '"\n'
  end

  if not baselevel then
     baselevel = ""
  end

  if #level - #baselevel > 30 then
     return "(LEVEL TOO LARGE)"
  end

  if type(item) == "number" then
     return tostring(item)
  end

  if type(item) == "nil" then
     return tostring(item)
  end

  if type(item) == "boolean" then
     return tostring(item)
  end

  if type(item) == "table" then

     if tables_dumped[item] ~= nil then
        return "[PREVIOUSLY DUMPED " .. tostring(item) .. "]"
     end

     tables_dumped[item] = true

     local string = ""
     local pairs_func = pairs
     if type(item.pairs) == 'function' then
        string = "<OBSERVABLE TABLE>:"
        pairs_func = function(item)
                        return item:pairs()
                     end
     end

     local keys = { }
     for k,v in pairs_func(item) do
        table.insert(keys, k)
     end
     
     --LRLib_TableEx.safeTableSort(keys)

     for _, k in ipairs(keys) do
        local v = item[k]

        local myskip = toskip or item._SKIP_
        if (myskip and myskip[k]) or k == "_SKIP_" then
           v = ">skipped<"
        end

        if level ~= "" then
           string = string .. level .. "| "
        end
        if type(k) == "string" then
           string = string .. k .. " = "
           if k == "_WinClass" then
              v = "(table)"
           end
        elseif type(k) == "number" then
           string = string .. tostring(k) .. " = "
        else
           string = string .. "{\n" .. dumpTable("KEY|" .. level, k, baselevel, myskip) .. "\n} = "
        end

        if type(v) == "table" then
           if tables_dumped[v] ~= nil then
              string = string .. "[[PREVIOUSLY DUMPED " .. tostring(v) .. "]]\n"
           elseif table_empty(v) then
              string = string .. "empty " .. tostring(v) .. "\n"
           else
              string = string .. tostring(v)
              string = string .. ":\n" .. dumpTable(level .. "   ", v, baselevel, myskip)
           end
        elseif type(v) == "function" then
           string = string .. "function\n"
        elseif type(v) == 'string' then
           string = string .. '"' .. v .. '"\n'
        else
           string = string .. tostring(v) .. "\n"
        end
     end

     return string
  end

  return "unknown[" .. type(item) .. "]\n"
end


function LRLib_TableEx.dump(item)
   tables_dumped = {}
   return dumpTable("", item)
end

function LRLib_TableEx.contains(tbl, item)
   for key, value in pairs(tbl) do
      if value == item then return key end
   end
   return false
end