local LrFunctionContext = import 'LrFunctionContext'
local LrApplication = import 'LrApplication'
local LrHttp = import 'LrHttp'
local LrErrors = import 'LrErrors'
local LrPasswords = nil

if LrApplication.versionTable().major >= 3 then
	LrPasswords = import 'LrPasswords'
end

require "LRLib_Spinner"
require "LRLib_InterOp"
require "LRLib_Logger"
require "LRLib_TableEx"
require "LRLib_DialogUtils"
require "LRLib_StringUtils"
require "LRLib_Scroller"
require "LRLib_Serialise"
require "JSON4Lua"
require "PhotoShelterHelper"
require "Trapper"

PSURLV2 = "https://www.photoshelter.com/psapi/v2/mem/"
PSURLV3 = "https://www.photoshelter.com/psapi/v3/mem/"

PhotoShelterAPI = {}

function PhotoShelterAPI.psurl(endpoint, id, func, params, apiVersion)
	local url = PSURLV2

	if apiVersion == 3 then
		url = PSURLV3
	end

	url = url .. endpoint

	if id then
		url = url .. "/" .. id
	end
	if func then
		url = url .. "/" .. func
	end

	if not params then
		params = {}
	end

	local sep = "?"

	for k, v in pairs(params) do
		v = string.gsub( v, '([^ !#-$&-*,%-%.0-;@-~])', function( c ) return string.format( '%%%02X', string.byte( c ) ) end )
		v = string.gsub( v, ' ', '+' )
		url = url .. sep .. k .. "=" .. v
		sep = "&"
	end

	return url
end

function PhotoShelterAPI.errorMessageFromResult(result)
	local success, errorDetails = PhotoShelterHelper.JsonDecode(result)

	if not success then
		return " (Unable to parse JSON input)"
	end

	if errorDetails.error.message then
		if type(errorDetails.error.message) == "table" then
			return " (" .. errorDetails.error.message[1] .. ")"
		else
			return " (" .. errorDetails.error.message .. ")"
		end
	end

	return ""
end

-- Authentication
function PhotoShelterAPI.logon(accountEmail, password)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitAuthenticateEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)

				local params = "mode=token&email=" .. LRLib_StringUtils.urlEncode(accountEmail) .. "&password=" .. LRLib_StringUtils.urlEncode(password)
				local headers = PhotoShelterHelper.authHeaders()

				headers[#headers + 1] = { field = "Content-Type", value = "application/x-www-form-urlencoded" }
				headers[#headers + 1] = { field = "Content-Length", value = #params }

				local url = PhotoShelterAPI.psurl("authenticate");

				return LrHttp.post(url, params .. '\r\n', headers)
			end
	)

	if success and headers  ~= nil and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			return true, JSONDetails.data
		else
			return true
		end
	end

	-- Connection failure
	return false, LOC("$$$/PhotoShelter/API/Error/UnableConnect=Unable to connect to PhotoShelter server.^n^nPlease check your username and password.")
end

function PhotoShelterAPI.logonOrganisation(id, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitAuthenticateOrganizationEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("organization", id, "authenticate");
				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/UnableConnectOrg=Unable to log onto organisation ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.logoffOrganisation(id, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitLogoutOrganizationEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("organization", nil, "logout");

				return LrHttp.get(url, headers)
			end
	)

	return success
end

function PhotoShelterAPI.getSiteURL(authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitUserSiteUrlEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("user", nil, "settings/site-url")

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			return JSONDetails.data.url
		end
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/GetSiteURL=Unable to get site URL ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.galleryOrCollectionExists(set, authToken)
	DBG("Confirming " .. PhotoShelterHelper.pubCollectionNameAndId(set) .. " removal from PhotoShelter.", nil, 'DEBUG')
	local id = set:getRemoteId()
	local success, data

	if PhotoShelterHelper.isCollection(set) then
		success, data = PhotoShelterAPI.queryCollectionById(id, authToken)
	else
		success, data = PhotoShelterAPI.queryGalleryById(id, authToken)
	end

	if success then
		DBG(PhotoShelterHelper.pubCollectionNameAndId(set) .. " still on server but not synced.", nil, 'DEBUG')
		return true
	else
		DBG(PhotoShelterHelper.pubCollectionNameAndId(set) .. " no longer on PhotoShelter.", nil, 'DEBUG')
		return false
	end
end

function PhotoShelterAPI.getChildren(collectionArr, authToken)
	-- Publish feature passes a table containing a PS collection (or 'root') and its LR counterpart
	-- Export feature passes a string 'root' or 'C0000....'/'G0000....'
	local collection = (type(collectionArr) == 'table' and collectionArr[1] or collectionArr)

	if collection == 'root' then
		collection = {}
		collection['id'] = 'root'
	elseif type(collection) == 'string' then
		local collectionID = collection
		collection = {}
		collection['id'] = collectionID
	end

	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitCollectionGetChildrenEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local params = {
					fields = "*",
					extend = '{ "Visibility" : { "params" : {}, "fields":"*" } }'
				}

				local url = PhotoShelterAPI.psurl("collection", collection.id, "children", params, 3);

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data.Children then
			-- V3 of the API is a mess. It can return either of the following formats:
			--
			-- Children
			--   1
			--      Collection
			--   2
			--      Gallery
			--   etc.
			--
			-- or (when requesting the root):
			--
			-- Children
			--   Collection
			--      1
			--      2 etc.
			--   Gallery
			--      1
			--      2 etc.
			--
			-- We reform the second case into the first for consistency
			local structure = JSONDetails.data.Children

			if JSONDetails.data.Children.Gallery or JSONDetails.data.Children.Collection then
				structure = {}

				if JSONDetails.data.Children.Collection then
					for _, collection in ipairs(JSONDetails.data.Children.Collection) do
						table.insert(structure, { Collection = collection })
					end
				end

				if JSONDetails.data.Children.Gallery then
					for _, gallery in ipairs(JSONDetails.data.Children.Gallery) do
						table.insert(structure, { Gallery = gallery })
					end
				end
			end

			return true, structure
		else
			return true, nil
		end
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/GetChildren=Unable to get children ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.updateGalleryOrCollection(childType, id, name, description, listed, visibility, parentID, authToken)
	local requestHeaders, params, url
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitUpdateGalleryOrCollectionEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				requestHeaders = PhotoShelterHelper.authHeaders(authToken)
				params = {
					name = name,
					description = description,
				}

				-- If there's no parent then we should set the f_list flag
				if parentID and parentID == "root" then
					params["f_list"] = PhotoShelterHelper.booleanToString(listed)
				end

				url = PhotoShelterAPI.psurl(childType, id, "update", params, 3)

				return LrHttp.get(url, requestHeaders)
			end
	)

	if not success or (success and headers ~= nil and headers.status ~= 200) then
		DBG('Update ' .. childType .. ' fail.', {
			url = url,
			requestHeaders = requestHeaders,
			params = params,
			success = success,
			result = result,
			responseHeaders = headers,
		},
				'ERROR')
	end

	if headers and headers.status == 200 then
		return PhotoShelterAPI.updateGalleryOrCollectionVisibility(childType, id, visibility, parentID, authToken)
	end

	return false, LOC(
			"$$$/PhotoShelter/API/UnableToUpdate=We was unable to update ^1:^2 (^3)",
			childType,
			name,
			id
	)
end

function PhotoShelterAPI.updateGalleryOrCollectionVisibility(childType, id, visibility, parentID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitUpdateGalleryOrCollectionVisibilityEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)

				local headers = PhotoShelterHelper.authHeaders(authToken)

				-- Top level can't be "inherit"
				local success, h

				if parentID == "root" then
					if visibility == "inherit" then
						visibility = "private"
					end
					success = true
				else
					local visibilityUrl = ""
					local inherit = PhotoShelterHelper.booleanToString(visibility == "inherit")

					-- Turn inheritance on or off
					if childType == "collection" then
						visibilityUrl = PhotoShelterAPI.psurl(childType, id, "permission/inherit",
																									{ collection_id = id, ancestor_id = parentID, inherit = inherit }, 3)
					else
						visibilityUrl = PhotoShelterAPI.psurl(childType, id, "permission/inherit",
																									{ gallery_id = id, parent_id = parentID, inherit = inherit }, 3)
					end

					success, h = LrHttp.get(visibilityUrl, headers)
				end

				-- If we're not inheriting, set the mode

				if success and visibility ~= "inherit" then
					local visibilityUrl = PhotoShelterAPI.psurl(childType, id, "visibility/update", { mode = visibility }, 3);
					success, h = LrHttp.get(visibilityUrl, headers)
				end

				return success, h
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, PhotoShelterAPI.errorMessageFromResult(result)
end

function PhotoShelterAPI.updateGallery(galleryID, name, description, listed, visibility, parentID, authToken)
	local success, errMsg = PhotoShelterAPI.updateGalleryOrCollection(
			"gallery", galleryID, name, description, listed,
			visibility, parentID, authToken
	)

	if success then
		return success
	end

	return false, LOC("$$$/PhotoShelter/API/Error/UpdateGallery=Unable to update gallery. ^n^nDetails: ^1", errMsg)
end

function PhotoShelterAPI.updateCollection(collectionID, name, description, listed, visibility, parentID, authToken)
	local success, errMsg = PhotoShelterAPI.updateGalleryOrCollection("collection", collectionID, name, description,
																																		listed, visibility, parentID, authToken)

	if success then
		return success
	end

	return false, LOC("$$$/PhotoShelter/API/Error/UpdateCollection=Unable to update collection. ^n^nDetails: ^1", errMsg)
end

function PhotoShelterAPI.createGalleryOrCollection(childType, name, description, listed, visibility, parent, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitCreateGalleryOrCollectionEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)

				local params = {
					name = name,
					description = description,
				}

				-- If there's no parent then we should set the f_list flag
				if parent and parent ~= "root" then
					params["parent"] = parent
				else
					params["f_list"] = PhotoShelterHelper.booleanToString(listed)
				end

				local url = PhotoShelterAPI.psurl(childType, nil, "insert", params, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)
		local errMsg

		if JSONDetails.data then
			local id

			if childType == "collection" then
				id = JSONDetails.data.Collection.collection_id
			else
				id = JSONDetails.data.Gallery.gallery_id
			end

			success, errMsg = PhotoShelterAPI.updateGalleryOrCollectionVisibility(childType, id, visibility, parent,
																																						authToken)

			if success then
				return true, id
			else
				return false, errMsg
			end

		end
	end

	return false, PhotoShelterAPI.errorMessageFromResult(result)
end

function PhotoShelterAPI.getGalleryOrCollectionPermissions(childType, id, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetGalleryOrCollectionPermissionsEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl(childType, id, "permission", nil, 2)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data.permission.mode then
			return true, JSONDetails.data.permission.mode
		end
	end

	return false, LOC("$$$/PhotoShelter/API/Error/GetPermissions=Unable to get permissions. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.getGalleryPermissions(id, authToken)
	return PhotoShelterAPI.getGalleryOrCollectionPermissions("gallery", id, authToken)
end

function PhotoShelterAPI.getCollectionPermissions(id, authToken)
	return PhotoShelterAPI.getGalleryOrCollectionPermissions("collection", id, authToken)
end

function PhotoShelterAPI.createGallery(name, description, listed, visibility, parent, authToken)
	local success, idOrMsg = PhotoShelterAPI.createGalleryOrCollection("gallery", name, description, listed, visibility,
																																		 parent, authToken)

	if success then
		return success, idOrMsg
	end

	return false, LOC("$$$/PhotoShelter/API/Error/CreateGallery=Unable to create gallery. ^n^nDetails: ^1", idOrMsg)
end

function PhotoShelterAPI.createCollection(name, description, listed, visibility, parent, authToken)
	local success, idOrMsg = PhotoShelterAPI.createGalleryOrCollection("collection", name, description, listed,
																																		 visibility, parent, authToken)

	if success then
		return success, idOrMsg
	end

	return false, LOC("$$$/PhotoShelter/API/Error/CreateCollection=Unable to create collection. ^n^nDetails: ^1", idOrMsg)
end

function PhotoShelterAPI.addGalleryToCollection(galleryID, collectionID, authToken)
	if collectionID == nil then
		return true
	end

	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitAddChildToParentEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("gallery", galleryID, "parents/" .. collectionID .. "/add", nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/AddGalleryToCollection=Unable to add gallery to collection on PhotoShelter server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.removeGalleryFromCollection(galleryID, collectionID, authToken)
	if collectionID == nil then
		return true
	end

	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitRemoveGalleryFromCollectionEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)

				local url = PhotoShelterAPI.psurl("gallery", galleryID, "parents/" .. collectionID .. "/remove",
																					{ gallery_id = galleryID, collection_id = collectionID }, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/RemoveGalleryFromCollection=Unable to remove gallery from collection on PhotoShelter server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.removeImage(imageId, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitRemoveImageEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("image", imageId, "delete", nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/RemoveGalleryFromCollection=Unable to remove image from PhotoShelter server. ^n^nDetails: ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.moveCollectionToCollection(collectionID, parentID, listed, authToken)
	if parentID == nil then
		parentID = "root"
	end

	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitMoveCollectionToCollectionEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)

				local params = { parent = parentID }
				-- If there's no parent then we should set the f_list flag
				if parentID and parentID == "root" then
					params["f_list"] = PhotoShelterHelper.booleanToString(listed)
					params["parent"] = ""
				end

				local url = PhotoShelterAPI.psurl("collection", collectionID, "update", params, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/AddCollectionToCollection=Unable to add collection to collection on PhotoShelter server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.deleteGallery(galleryID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitDeleteGalleryEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)

				local url = PhotoShelterAPI.psurl("gallery", galleryID, "delete", { gallery_id = galleryID }, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/DeleteGallery=Unable to delete gallery on PhotoShelter server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.deleteCollection(collectionID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitDeleteCollectionEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("collection", collectionID, "delete", { collection_id = collectionID }, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/DeleteCollection=Unable to delete collection on PhotoShelter server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.addImageToGallery(galleryID, photoID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitAddImageToGalleryEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("gallery", galleryID, "images/" .. photoID .. "/add", nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/AddImage=Unable to add image to gallery on PhotoShelter server. ^n^nDetails: ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.queryGalleryById(galleryId, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGalleryByIdEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("gallery", galleryId, nil, nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			return true, JSONDetails.data or {}
		end
	end

	return false, LOC("$$$/PhotoShelter/API/Error/GetGallery=Unable to get gallery from server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.queryImageExists(imageId, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitImageByIdEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("image", imageId, nil, {
					fields = 'image_id'
				}, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	elseif success and headers and headers.status == 404 then
		return false
	end

	return false
end

function PhotoShelterAPI.queryCollectionById(collectionId, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitCollectionByIdEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("collection", collectionId, nil, nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			return true, JSONDetails.data or {}
		end
	end

	return false, LOC("$$$/PhotoShelter/API/Error/GetGallery=Unable to get collection from server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.queryImagesByGallery(galleryID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitImagesByGalleryEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local params = {
					per_page = 100000,
					extend = '{ "Image" : { "params" : {}, "fields": "file_name,width,height,model_released,property_released" } }'
				}
				local url = PhotoShelterAPI.psurl("gallery", galleryID, "images", params, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			return true, JSONDetails.data.GalleryImage or {}
		end
	end

	return false, LOC("$$$/PhotoShelter/API/Error/GetImage=Unable to get images from PhotoShelter server. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.queryImagesByName(galleryID, filename, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitImagesByGalleryEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local params = {
					gallery_id = galleryID,
					file_name = filename,
					fields = 'file_name',
					per_page = 99999 -- hacked because didn't exists an endpoint to ask what I'm asking right now.
				}
				local url = PhotoShelterAPI.psurl("image", nil, "query", params, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			local data = JSONDetails.data or {}

			local filtered = {
				Image = {},
				Total = 0
			}
			-- we need do this because API return name without it be exact, they does a partial match so we could have
			-- false positives.
			for _, image in ipairs(data.Image) do
				if image.file_name == filename then
					filtered.Image[#filtered + 1] = image
					filtered.Total = filtered.Total + 1
				end
			end

			return true, filtered
		end
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/QueryImage=Unable to check how many files with same name exists in PhotoShelter Gallery. ^n^nDetails: ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.removeImageFromGallery(galleryID, photoID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitRemoveImageFromGalleryEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("gallery", galleryID, "images/" .. photoID .. "/remove", nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/RemoveImage=Unable to remove image from gallery on PhotoShelter server. ^n^nDetails: ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)

end

function PhotoShelterAPI.getParentGalleries(imageID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetParentGalleriesEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				-- Using V2 of the API since V3 doesn't return anything!
				local url = PhotoShelterAPI.psurl("image", imageID, "galleries", { image_id = imageID }, 2)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data then
			return true, JSONDetails.data.galleries
		end
	end

	return false, LOC(
			"$$$/PhotoShelter/API/Error/ImageGalleries=Unable to retrieve parent galleries for image data. ^n^nDetails: ^1",
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.updateImage(imageID, searchable, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitUpdateImageEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("image", imageID, "update", { f_searchable = booleanToString(searchable) }, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/UpdateImage=Unable to update image data. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))

end

function PhotoShelterAPI.getIPTC(imageID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetIPTCEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("image", imageID, "iptc", { image_id = imageID }, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.Iptc then
			return true, JSONDetails.data.Iptc
		end
	end

	return false, LOC("$$$/PhotoShelter/API/Error/UpdateImage=Unable to get IPTC data. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.setIPTC(imageID, iptc, authToken)
	local rating = -1
	local parts = {}

	for k, v in pairs(iptc) do
		if k == "rating" then
			rating = v
		else
			table.insert(parts, { name = k, value = v })
		end
	end

	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitUpdateIPTCEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PhotoShelterAPI.psurl("image", imageID, "iptc/update", { image_id = imageID }, 2)

				return LrHttp.postMultipart(url, parts, headers)
			end
	)

	-- Update the rating too
	if success and headers and headers.status == 200 then
		success, result, headers = LrFunctionContext.pcallWithContext(
				'HitUpdateImageEndpoint',
				function(context)
					context:addFailureHandler(PhotoShelterHelper.failureHandler)
					local headers = PhotoShelterHelper.authHeaders(authToken)
					local parts = {}
					local url = PhotoShelterAPI.psurl("image", imageID, "update", { image_id = imageID }, 3)

					table.insert(parts, { name = 'rating', value = rating })

					return LrHttp.postMultipart(url, parts, headers)
				end
		)
	end

	if success and headers and headers.status == 200 then
		return true
	end

	return false, LOC("$$$/PhotoShelter/API/Error/UpdateImage=Unable to set IPTC data. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.getEXIF(imageID, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetEXIFEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)

				local url = PhotoShelterAPI.psurl("image", imageID, "exif", { image_id = imageID }, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		if not result then
			return true, {}
		end

		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.Exif then
			local exif = {}
			for _, data in pairs(JSONDetails.data.Exif) do
				exif[data.hex] = { value = data.value, name = data.name }
			end

			return true, exif
		else
			return true, {}
		end
	end

	return false, LOC("$$$/PhotoShelter/API/Error/UpdateImage=Unable to get EXIF data. ^n^nDetails: ^1",
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.getGalleryAndCollectionCount(authToken)
	local success, result, headers
	local totalCount = 0

	success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetCollectionCountEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PSURLV3 .. 'collection'

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.Collection then
			totalCount = totalCount + JSONDetails.data.Total
		end
	end

	success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetGalleryCountEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local url = PSURLV3 .. 'gallery'

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.Gallery then
			totalCount = totalCount + JSONDetails.data.Total
		end

		return true, totalCount
	end

	return false, LOC('$$$/PhotoShelter/API/Error/getGalleryAndCollectionCount=Unable to get total count of collections and galleries. ^n^nDetails: ^1',
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.getGalleryOrCollectionListed(childType, childId, authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetGalleryOrCollectionEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local params = {
					fields = "f_list"
				}
				local url = (
						childType == 'Collection' and
								PhotoShelterAPI.psurl('collection', childId, nil, params, 3) or
								PhotoShelterAPI.psurl('gallery', childId, nil, params, 3)
				)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local listed
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.Gallery then
			listed = JSONDetails.data.Gallery.f_list == "t"
		else
			listed = JSONDetails.data.Collection.f_list == "t"
		end

		return true, listed
	end

	return false, LOC('$$$/PhotoShelter/API/Error/getGalleryOrCollectionListed=Unable to get çollection or gallery listed status. ^n^nDetails: ^1',
										PhotoShelterAPI.errorMessageFromResult(result))
end

function PhotoShelterAPI.getGalleryListed(galleryId, authToken)
	return PhotoShelterAPI.getGalleryOrCollectionListed('Gallery', galleryId, authToken)
end

function PhotoShelterAPI.getCollectionListed(collectionId, authToken)
	return PhotoShelterAPI.getGalleryOrCollectionListed('Collection', collectionId, authToken)
end

function PhotoShelterAPI.getUserId(authToken)
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetUserIdEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)
				local params = {
					fields = "user_id, org_id"
				}
				local url = PhotoShelterAPI.psurl('user', nil, 'session', params, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.Session then
			local session = JSONDetails.data.Session
			return true, session.user_id, session.org_id
		end
	end

	return false, LOC(
			'$$$/PhotoShelter/API/Error/getUserId=Unable to get user id. ^n^nDetails: ^1',
			PhotoShelterAPI.errorMessageFromResult(result)
	)
end

function PhotoShelterAPI.getUserOrgRole(authToken, orgId, userId)
	if orgId == nil or userId == nil then
		return false
	end

	local success, result, headers = LrFunctionContext.pcallWithContext(
			'HitGetUserOrgRoleEndpoint',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)
				local headers = PhotoShelterHelper.authHeaders(authToken)

				local url = PhotoShelterAPI.psurl('organization', orgId, 'user/' .. userId .. '?fields=*', nil, 3)

				return LrHttp.get(url, headers)
			end
	)

	if success and headers and headers.status == 200 then
		local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

		if JSONDetails.data and JSONDetails.data.OrganizationUser then
			local role = JSONDetails.data.OrganizationUser.role and JSONDetails.data.OrganizationUser.role or nil
			return true, role
		end
	end

	return false
end

function PhotoShelterAPI.upload(pt, path, leafName, contentType, galleryNameOrID, searchable, authToken)
	local f, errMsg = io.open(path, "rb")
	local file_data

	if f then
		file_data = f:read("*a")
		f:close()
	else
		LrErrors.throwUserError(errMsg)
	end

	local uploadBody
	local postHeaders
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'UploadPhoto',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)

				local delimiter = "--PART_DELIMITER"

				if searchable then
					searchable = "t"
				else
					searchable = "f"
				end

				uploadBody = ''
						.. '--' .. delimiter .. "\r\n"
						.. 'Content-Disposition: form-data; name="name"\r\n'
						.. '\r\n'
						.. leafName .. '\r\n'
						.. '--' .. delimiter .. '\r\n'

				if galleryNameOrID then
					if not pt.contributor then
						uploadBody = uploadBody .. 'Content-Disposition: form-data; name="gallery_id"\r\n'
								.. '\r\n'
								.. galleryNameOrID .. '\r\n'
								.. '--' .. delimiter .. '\r\n'
					else
						uploadBody = uploadBody .. 'Content-Disposition: form-data; name="gallery_name"\r\n'
								.. '\r\n'
								.. galleryNameOrID .. '\r\n'
								.. '--' .. delimiter .. '\r\n'
					end
				end

				if pt.photographer ~= nil and pt.org ~= nil then
					uploadBody = uploadBody .. 'Content-Disposition: form-data; name="photographer_id"\r\n'
							.. '\r\n'
							.. pt.photographer .. '\r\n'
							.. '--' .. delimiter .. '\r\n'
				end

				uploadBody = uploadBody .. 'Content-Disposition: form-data; name="f_searchable"\r\n'
						.. '\r\n'
						.. searchable .. '\r\n'
						.. '--' .. delimiter .. '\r\n'

				uploadBody = uploadBody .. 'Content-Disposition: form-data; name="file"; filename="' .. leafName .. '"\r\n'
						.. 'Content-Type: ' .. contentType .. '\r\n'
						.. '\r\n'
						.. file_data .. '\r\n'
						.. '--' .. delimiter .. "--" .. '\r\n'

				postHeaders = PhotoShelterHelper.authHeaders(authToken)

				postHeaders[#postHeaders + 1] = { field = "Content-Type", value = "multipart/form-data; boundary=" .. delimiter }

				return LrHttp.post(
				--"https://ul.photoshelter.com/psapi/v2/mem/image/upload",
						"https://www.photoshelter.com/psapi/v3/mem/image/upload",
						uploadBody,
						postHeaders,
						"POST",
						60 * 30
				)
			end
	)

	if not success then
		LrErrors.throwUserError(LOC("$$$/PhotoShelter/Errors/Upload/UnableConnect=Unable to connect to PhotoShelter server"))
	elseif not headers then
		LrErrors.throwUserError(LOC("$$$/PhotoShelter/Errors/Upload/NoHeaders=No HTTP headers returned from PhotoShelter server"))
	elseif headers.status ~= 200 then
		DBG(
				"Uploading file",
				{
					status = headers.status or 0,
					sentHeader = postHeaders,
					response = result,
					path = path,
					leafName = leafName,
					contentType = contentType,
					galleryNameOrID = galleryNameOrID,
					searchable = searchable
				},
				"ERROR"
		)

		local data = PhotoShelterHelper.throwJsonDecode(result)

		LrErrors.throwUserError(data.error.message or 'Something went wrong calling PhotoShelter but we was unable to read the service error.');
	end

	local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

	if JSONDetails.data and #JSONDetails.data.Image > 0 and JSONDetails.data.Image[1].image_id then
		return JSONDetails.data.Image[1].image_id
	end

	LrErrors.throwUserError(LOC("$$$/PhotoShelter/Errors/Upload/NoIDReturned=No ID returned for uploaded image."))
end

function PhotoShelterAPI.replace(pt, path, leafName, contentType, imageID, searchable, authToken)
	DBG("Replacing " .. imageID, nil, 'DEBUG')

	local f, errMsg = io.open(path, "rb")
	local file_data

	if f then
		file_data = f:read("*a")
		f:close()
	else
		LrErrors.throwUserError(errMsg)
	end

	local uploadBody
	local postHeaders
	local success, result, headers = LrFunctionContext.pcallWithContext(
			'ReplacePhoto',
			function(context)
				context:addFailureHandler(PhotoShelterHelper.failureHandler)

				local delimiter = "--PART_DELIMITER"

				if searchable then
					searchable = "t"
				else
					searchable = "f"
				end

				uploadBody = ''
						.. '--' .. delimiter .. "\r\n"
						.. 'Content-Disposition: form-data; name="f_searchable"\r\n'
						.. '\r\n'
						.. searchable .. '\r\n'
						.. '--' .. delimiter .. '\r\n'

				uploadBody = uploadBody .. 'Content-Disposition: form-data; name="file"; filename="' .. leafName .. '"\r\n'
						.. 'Content-Type: ' .. contentType .. '\r\n'
						.. '\r\n'
						.. file_data .. '\r\n'
						.. '--' .. delimiter .. "--" .. '\r\n'

				postHeaders = PhotoShelterHelper.authHeaders(authToken)
				postHeaders[#postHeaders + 1] = {
					field = "Content-Type",
					value = "multipart/form-data; boundary=" .. delimiter
				}

				return LrHttp.post(
						PhotoShelterAPI.psurl("image", imageID, "replace"),
						uploadBody,
						postHeaders,
						"POST",
						60 * 30
				)
			end
	)

	if not success then
		LrErrors.throwUserError(LOC("$$$/PhotoShelter/Errors/Upload/UnableConnect=Unable to connect to PhotoShelter server"))
	elseif headers == nil or headers.status == nil then
		LrErrors.throwUserError(LOC("$$$/PhotoShelter/Errors/Upload/NoHeaders=No HTTP headers returned from PhotoShelter server"))
	elseif headers.status ~= 201 and headers.status ~= 200 then
		DBG(
				"Replacing file",
				{
					status = headers.status or 0,
					sentHeader = postHeaders,
					response = result,
					path = path,
					leafName = leafName,
					contentType = contentType,
					galleryNameOrID = galleryNameOrID,
					searchable = searchable
				},
				"ERROR"
		)

		local data = PhotoShelterHelper.throwJsonDecode(result)

		LrErrors.throwUserError(data.error.message or 'Something went wrong calling PhotoShelter but we was unable to read the service error.');
	end
end
