local LrFunctionContext = import 'LrFunctionContext'
local LrApplication = import 'LrApplication'
local LrPathUtils = import "LrPathUtils"
local LrDialogs = import 'LrDialogs'
local LrBinding = import 'LrBinding'
local LrErrors = import 'LrErrors'
local LrColor = import 'LrColor'
local LrTasks = import 'LrTasks'
local LrHttp = import 'LrHttp'
local LrView = import 'LrView'

local LrPasswords = nil

if LrApplication.versionTable().major >= 3 then
	LrPasswords = import 'LrPasswords'
end

require "LRLib_Spinner"
require "LRLib_InterOp"
require "LRLib_Logger"
require "LRLib_TableEx"
require "LRLib_DialogUtils"
require "LRLib_StringUtils"
require "LRLib_Scroller"
require "LRLib_Serialise"

require "JSON4Lua"
require "PhotoShelterAPI"
require "Trapper"

local prefs = LRLib_InterOp.prefs()
local NUM_COLS = 2

PhotoShelterUploadExportDialogSections = {}

----- Scroller init -----

-- The fields saved by each scroller
local scrollerFields = {
	{ key = "items", default = {} },
	{ key = "selected", default = "" }
}

local function scrollerRow(f, scroller, row)
	local bind = LrView.bind

	return f:simple_list {
		width = 260,
		height = 180,
		items = bind(scroller:bindID("items", row)),
		value = bind(scroller:bindID("selected", row)),
	}
end

-- Create the scroller
local scroller = LRLib_Scroller:new('folders', scrollerFields, scrollerRow, NUM_COLS)

scroller:setDynamic(false)
scroller:setHorizontal()
scroller:setDisplayControls(false)
scroller.atLeastOne = true

-- Function to retrieve folder hierarchy
local function getFolderContents(pt, id)
	-- Send the request for this collection id
	local success, children = PhotoShelterAPI.getChildren(id, pt.authToken)
	-- Handle the response
	local contents = {}

	if success and children then
		for _, child in pairs(children) do
			if child.Collection then
				local collection = {
					name = child.Collection.name,
					id = child.Collection.collection_id,
					type = "collection",
					children = nil
				}

				-- We generally only fetch children when they are requested, however
				-- we'll need them now if they are in the selected path
				if LRLib_TableEx.contains(pt.folderPath, child.Collection.collection_id) then
					collection.children = getFolderContents(pt, child.Collection.collection_id)
				end

				if child.Collection.f_list then
					collection.listed = child.Collection.f_list == "t"
				else
					collection.listed = false
				end

				table.insert(contents, collection)

				pt.itemIDMap[child.Collection.collection_id] = collection
			elseif child.Gallery then
				local gallery = {
					name = child.Gallery.name,
					type = "gallery",
					id = child.Gallery.gallery_id
				}

				if child.Gallery.f_list then
					gallery.listed = child.Gallery.f_list == "t"
				else
					gallery.listed = false
				end

				table.insert(contents, gallery)

				pt.itemIDMap[child.Gallery.gallery_id] = gallery
			end
		end
	end

	-- Sort the collections alphabetically
	table.sort(contents, function(a, b)
		return a.name < b.name
	end)

	return contents
end

-- Set up the scroller and folder display based on the folder hierarchy and the user's current selection
local function setUpScroller(pt)
	-- Stop recursion
	if pt.settingUpScrollers then
		return
	end

	pt.settingUpScrollers = true

	local columns = {}
	local curFolder = pt.root

	local validFolderPath = {}
	local curDepth = 1
	local nextFolder

	while curFolder do
		-- Get the ID of the selected item at this level
		local selectedID = ""

		if #pt.folderPath >= curDepth then
			selectedID = pt.folderPath[curDepth]
		end

		-- Don't display any more columns (unless we find a collection with our selected index later on)
		nextFolder = nil

		-- Add each item in this folder to the column
		local items = {}
		local confirmedSelectedID = nil

		for _, item in ipairs(curFolder) do
			if curDepth > 1 or (item.listed == pt.showListed) then
				if item.type == "collection" then
					items[#items + 1] = { title = item.name .. " >", value = item.id }

					-- If this collection is selected, we'll display the contents in the next column,
					-- or an empty column if there are no children
					if item.id == selectedID then
						nextFolder = item.children or {}
					end

				elseif item.type == "gallery" then
					items[#items + 1] = { title = item.name, value = item.id }
				end

				-- If this collection or gallery is selected then we'll add it to the validated
				-- folder path
				if item.id == selectedID then
					validFolderPath[#validFolderPath + 1] = selectedID
					confirmedSelectedID = selectedID
				end
			end

		end

		-- Add the folder into the next column in the scroller. Select the selected item if
		-- there is one
		if confirmedSelectedID then
			columns[#columns + 1] = { items = items, selected = { confirmedSelectedID } }
		else
			columns[#columns + 1] = { items = items, selected = {} }
		end

		-- Move on to the next folder
		curDepth = curDepth + 1
		curFolder = nextFolder
	end

	-- Update the stored folderPath with the validated one
	pt.folderPath = validFolderPath

	-- Display the folders and scroll to the last visible column
	local scrollOffset = #columns - NUM_COLS

	if scrollOffset < 0 then
		scrollOffset = 0
	end

	scroller:setScrollOffset(scrollOffset)
	scroller:setRows(columns)

	pt.settingUpScrollers = false
end

-- Returns the name of the selected folder (display above each column)
local function selectedItemName(column)
	local pt = scroller.pt
	local name = ""

	if column == 1 then
		return LOC("$$$/PhotoShelter/ExportDialog/Root=Root")
	end

	if #pt.folderPath < column - 1 then
		return ""
	end

	if pt.itemIDMap[pt.folderPath[column - 1]] then
		name = pt.itemIDMap[pt.folderPath[column - 1]].name
	end

	return name
end

scroller:setCustomIndexDisplayCallback(selectedItemName)

-- Support for clicking the index name (Collection titles above the scroll areas)
--
-- We Remove the selected from the corresponding column
local function indexClicked(column)
	-- Recreate the folder path
	local pt = scroller.pt
	local folderPath = {}

	local n = 1
	while n < column do
		folderPath[n] = pt.folderPath[n]
		n = n + 1
	end

	pt.folderPath = folderPath
	setUpScroller(pt)
end

scroller:setIndexClickCallback(indexClicked)

-- This function is called back whenever the user clicks on a different folder in the scroller
-- We need to update the displayed folders
local function updateFolderSelection(pt)
	if pt.connected and pt.settingUpScrollers == false and pt.scrolling == false then
		-- If we react to the click immediately, which will automatically scroll the columns,
		-- then LR will then proceed to highlight the clicked on item *after* we've updated the
		-- columns, thereby highlighting the corresponding item in the newly displayed
		-- column.
		--
		-- Posting an async task avoids this problem.
		LrFunctionContext.postAsyncTaskWithContext("ConnectTask", function(context)
			context:addFailureHandler(PhotoShelterHelper.failureHandler)
			context:addFailureHandler(
					function(status, theError)
						prefs.lastConnectOK = false
						pt.connecting = false
					end
			)

			Trapper.TrapContextErrors(context)

			-- Reset the folderPath based on the current selection
			scroller:saveVisible()

			local columns = scroller:getRows()

			local folderPath = {}

			for _, column in ipairs(columns) do
				if column.selected then
					folderPath[#folderPath + 1] = column.selected[1]
				else
					break
				end
			end

			pt.folderPath = folderPath

			-- Ensure that we have download the contents of the selected folderPath
			local deepestSelectedID = pt.folderPath[#pt.folderPath]
			local deepestSelectedItem = pt.itemIDMap[deepestSelectedID]
			if deepestSelectedItem and deepestSelectedItem.type == "collection" and deepestSelectedItem.children == nil then
				pt.fetchingContents = true
				LRLib_Spinner.start(pt, "fetchingContents", false)
				deepestSelectedItem.children = getFolderContents(pt, deepestSelectedID)
				pt.fetchingContents = false

			end

			-- Reset the scroller based on this path
			setUpScroller(pt)
			scroller:refreshIndexDisplay()
		end)
	end

end

-- Refresh folder view
local function refreshFolders(pt)
	-- No need to do this for the publish service
	if pt.LR_isExportForPublish then
		return
	end

	pt.itemIDMap = {}

	if not pt.contributor then
		pt.root = getFolderContents(pt, "root", 0)
	else
		pt.root = {}
	end

	setUpScroller(pt)
end

local function refreshFoldersInNewTask(pt)
	if pt.connected then
		LrTasks.startAsyncTask(function()
			Trapper.TRAP(function()
				pt.connected = false
				pt.connecting = true
				LRLib_Spinner.start(pt, "connecting", false)
				refreshFolders(pt)
				pt.connecting = false
				pt.connected = true
			end)
		end)
	end
end

local function organizationChanged(pt)
	if pt.connected == true then
		local f = function()
			pt.connected = false
			pt.connecting = true

			LRLib_Spinner.start(pt, "connecting", false)

			-- First, disconnect from the organization
			if pt.prevOrg then
				PhotoShelterAPI.logoffOrganisation(pt.authToken)
			end

			-- Now log on to the new one
			if pt.org then
				local success = PhotoShelterAPI.logonOrganisation(pt.org, pt.authToken)

				if success then
					pt.contributor = pt.orgDetails[pt.org].fullMember == false
					pt.prevOrg = pt.org

					-- Get the photographer's menu
					local success, result, _ = LrFunctionContext.pcallWithContext('Endpoint',
																																				function(context)
																																					context:addFailureHandler(PhotoShelterHelper.failureHandler)
																																					local headers = PhotoShelterHelper.authHeaders(pt.authToken)
																																					local url = PhotoShelterAPI.psurl("organization/" .. pt.org .. "/photographers");

																																					return LrHttp.get(url, headers)
																																				end
					)

					local photographers = {}
					local selectedPhotographer = nil

					if success then
						local JSONDetails = PhotoShelterHelper.throwJsonDecode(result)

						if JSONDetails.data.photographers then
							photographers = {
								{
									title = LOC "$$$/PhotoShelter/ExportDialog/Photogs/None=- Leave Blank -",
									value = nil
								}
							}

							for _, photographer in ipairs(JSONDetails.data.photographers) do
								photographers[#photographers + 1] = {
									title = photographer.first_name .. " " .. photographer.last_name,
									value = photographer.user_id
								}
								if photographer.user_id == pt.photographer then
									selectedPhotographer = photographer.user_id
								end
							end
						end
					end

					pt.photographers = photographers
					pt.photographer = selectedPhotographer
				else
					-- Unable to authenticate
					pt.org = pt.prevOrg
				end
			else
				-- The is the subscriber's own account
				pt.photographers = {}
				pt.contributor = false
			end

			-- Refresh the folder display
			refreshFolders(pt)
			pt.connecting = false
			pt.connected = true
		end
		-- Call fom a new task if we're not already in one
		if LrTasks.canYield() then
			f()
		else
			LrTasks.startAsyncTask(function()
				Trapper.TRAP(function()
					f()
				end)
			end)
		end
	end
end

-- Connect and authorise the user
local function connect(pt)
	pt.orgs = {}
	pt.orgDetails = {}
	pt.contributor = false

	LrFunctionContext.postAsyncTaskWithContext("ConnectTask", function(context)
		context:addFailureHandler(PhotoShelterHelper.failureHandler)
		context:addFailureHandler(function(status, theError)
			prefs.lastConnectOK = false
			pt.connecting = false
		end)

		pt.connected = false
		pt.connecting = true

		LRLib_Spinner.start(pt, "connecting", false)

		local success, dataOrMsg = PhotoShelterAPI.logon(pt.accountEmail, pt.password)

		if success then
			if dataOrMsg then
				pt.authToken = dataOrMsg.token;

				-- Limit organization account logins to the Libris version of the plugin.
				if PhotoShelterInfo.isLibris then
					local orgToConnectTo = nil

					if dataOrMsg.org then
						-- Multi-user account, get the organisations
						local orgs = {}

						if not PhotoShelterInfo.isLibris then
							orgs[#orgs + 1] = {
								title = LOC("$$$/PhotoShelter/ExportDialog/Orgs/MyAccount=- My Subscriber Account -"),
								value = nil
							}
						end

						for _, org in ipairs(dataOrMsg.org) do
							pt.orgDetails[org.id] = { name = org.name, fullMember = org.member == "t" }
							orgs[#orgs + 1] = { title = org.name, value = org.id }

							-- If this is for Libris and there's nothing currently selected, then choose the first one by default
							if PhotoShelterInfo.isLibris and not pt.org then
								pt.org = org.id
							end

							-- Reconnect to last used organization if we find item
							if org.id == pt.org then
								orgToConnectTo = org.id
							end
						end

						pt.orgs = orgs
						pt.isMUMember = true
					else
						pt.isMUMember = false
					end

					pt.org = orgToConnectTo
				end
			end
		end

		-- Report failure or connect
		if not success then
			LrErrors.throwUserError(dataOrMsg)
			prefs.lastConnectOK = false
		else
			pt.connected = nil
			pt.connected = true
			prefs.lastConnectOK = true
			-- Refresh the folders immediately
			organizationChanged(pt)
		end

		-- We're no longer connecting
		pt.connecting = false
	end)
end

local function logout(pt)
	pt.orgs = {}
	pt.orgDetails = {}
	pt.contributor = false
	pt.connected = nil
	pt.connected = false
	pt.connecting = false
	pt.authToken = nil
	prefs.lastConnectOK = false
end

local function updateExportStatus(pt)
	local message = nil

	-- Make sure that the user's logged on
	if message == nil then
		if not pt.connected then
			message = LOC("$$$/PhotoShelter/ExportDialog/Messages/NotConnected=Please log onto your PhotoShelter account")
		end
	end

	-- A contributor can't create a publish service
	if pt.LR_isExportForPublish and message == nil and pt.contributor then
		message = LOC("$$$/PhotoShelter/ExportDialog/Messages/ContributorCantPublish=Publish Services can't be created for contributors. Use Export instead.")
	end

	-- Ensure that the user's selected a gallery to export to
	if not pt.LR_isExportForPublish and (message == nil and not pt.contributor) then
		if #pt.folderPath == 0 or (pt.itemIDMap[pt.folderPath[#pt.folderPath]] and pt.itemIDMap[pt.folderPath[#pt.folderPath]].type == "collection") then
			message = LOC("$$$/PhotoShelter/ExportDialog/Messages/NoGallery=Please select the gallery that you wish to export your photos to...")
		end
	end

	-- Enable/disable export
	if message then
		pt.message = message
		pt.hasError = true
		pt.hasNoError = false
		pt.LR_cantExportBecause = message
	else
		pt.message = nil
		pt.hasError = false
		pt.hasNoError = true
		pt.LR_cantExportBecause = nil
	end
end

function PhotoShelterUploadExportDialogSections.exportPresetFields()
	return LRLib_TableEx.merge(
	-- The presets as defined by the scroller
			scroller:exportPresetFields(),
	-- Other fields
			{
				{ key = "accountEmail", default = "" },
				{ key = "serialisedFolderPath", default = "{}" },
				{ key = "showListed", default = true },
				{ key = "searchable", default = false },
				{ key = "ifexists", default = "upload" },
				{ key = "org", default = nil },
				{ key = "photographer", default = nil }
			}
	)
end

function PhotoShelterUploadExportDialogSections.startDialog(pt)
	-- Set up default values
	if not prefs.InitDone then
		prefs.InitDone = true
		pt.LR_export_destinationType = 'tempFolder'
	end

	-- Reset some settings
	local password = LrPasswords.retrieve(pt.accountEmail, pt.accountEmail, nil) or ""
	pt.password = password

	-- Observe for export enabling changes
	pt.suffixUnique = true

	pt:addObserver('email', updateExportStatus)
	pt:addObserver('password', updateExportStatus)
	pt:addObserver('connected', updateExportStatus)
	pt:addObserver('folderPath', updateExportStatus)

	updateExportStatus(pt)

	-- Initialise scroller
	scroller:startDialog(pt)
	scroller:setRows({})

	pt.settingUpScrollers = false
	pt.scrolling = false

	-- Initialise variables
	pt.authToken = ""
	pt.connected = false
	pt.connecting = false
	pt.itemIDMap = {}
	pt.orgs = {}
	pt.isMUMember = false
	pt.orgDetails = {}
	pt.contributor = false
	pt.prevOrg = nil
	pt.fetchingContents = false

	--  Set default file format options
	if pt.LR_editingExistingPublishConnection == false then
		pt.LR_format = "JPEG"
		pt.LR_jpeg_quality = 1
		pt.LR_size_resolution = 300
		pt.LR_removeLocationMetadata = false
	end

	-- Observer scroller selection changes
	for n = 1, NUM_COLS do
		pt:addObserver(scroller:bindID("selected", n), updateFolderSelection)
	end

	-- Refresh scroller when listed/unlisted is toggled
	pt:addObserver("showListed", setUpScroller)

	-- Refresh connection when organization is changed
	pt:addObserver("org", organizationChanged)

	-- Extract the saved folder path
	if not pt.serialisedFolderPath or pt.serialisedFolderPath == "" then
		pt.serialisedFolderPath = "{}"
	end

	local success, theError = LrTasks.pcall(loadstring("folderPath = " .. pt.serialisedFolderPath))

	if not success then
		MSG(theError)
		pt.folderPath = {}
	else
		pt.folderPath = folderPath
	end

	-- Auto connect
	if prefs.lastConnectOK and #pt.accountEmail > 0 and #pt.password then
		connect(pt)
	end
end

function PhotoShelterUploadExportDialogSections.endDialog(pt)
	if pt.accountEmail and #pt.accountEmail > 0 and pt.password ~= nil then
		LrPasswords.store(pt.accountEmail, pt.password, pt.accountEmail, nil)
	end

	scroller:endDialog(pt)
	pt.serialisedFolderPath = LRLib_Serialise.serialise(pt.folderPath or {})
end

-- New gallery/collection dialog
local function createItemOnServer(pt, dialogToClose, name, description, visibility, forCollection)
	LrFunctionContext.postAsyncTaskWithContext("CreateTask", function(context)
		context:addFailureHandler(PhotoShelterHelper.failureHandler)
		context:addFailureHandler(function(status, theError)
			pt.createStatus = LOC("$$$/PhotoShelter/ExportDialog/Create/Error=Unable to create collection")
			pt.creating = false
			PhotoShelterUploadExportDialogSections.newItemDialog(pt, forCollection, false)
		end)

		pt.createStatus = ""
		pt.creating = true

		LRLib_Spinner.start(pt, "creating", false)

		-- Try to create the collection/gallery
		local parentID = nil

		if #pt.folderPath > 0 then
			parentID = pt.folderPath[#pt.folderPath]
		end

		local success, idOrMsg

		if forCollection then
			success, idOrMsg = PhotoShelterAPI.createCollection(name, description, pt.showListed, visibility, parentID,
																													pt.authToken)
		else
			success, idOrMsg = PhotoShelterAPI.createGallery(name, description, pt.showListed, visibility, parentID,
																											 pt.authToken)
		end

		if success then
			-- Update the folder view
			local childrenTable

			if not parentID then
				childrenTable = pt.root
			else
				childrenTable = pt.itemIDMap[parentID].children
			end

			local newItem = {
				name = name,
				id = idOrMsg,
				listed = pt.showListed
			}

			if forCollection then
				newItem.type = "collection"
				newItem.children = {}
			else
				newItem.type = "gallery"
			end

			pt.itemIDMap[idOrMsg] = newItem
			childrenTable[#childrenTable + 1] = newItem

			table.sort(childrenTable, function(a, b)
				return a.name < b.name
			end)

			pt.folderPath[#pt.folderPath + 1] = idOrMsg
			pt.folderPath = pt.folderPath

			setUpScroller(pt)
		else
			pt.createStatus = idOrMsg
			-- We failed to create, reopen the create dialog with the error message
			pt.creating = false

			PhotoShelterUploadExportDialogSections.newItemDialog(pt, forCollection, false)
		end

		pt.creating = false
	end)
end

function PhotoShelterUploadExportDialogSections.newItemDialog(pt, forCollection, reinit)
	LrFunctionContext.callWithContext("CollectionDialog", function(context)
		local f = LrView.osFactory()
		local bind = LrView.bind
		local share = LrView.share

		Trapper.TrapContextErrors(context)

		-- Set up the default parameters
		if reinit then
			pt.dlg_name = ""
			pt.dlg_description = ""
			pt.createStatus = ""
		end

		local parent

		if #pt.folderPath == 0 then
			pt.dlg_visibility = "private"
			parent = "Root"
		else
			pt.dlg_visibility = "inherit"
			parent = selectedItemName(#pt.folderPath + 1)
		end

		local createText, createTitle, createButton

		if forCollection then
			createTitle = LOC("$$$/PhotoShelter/ExportDialog/CreateCollection/Title=Create a new collection")
			createText = LOC("$$$/PhotoShelter/ExportDialog/CreateCollection/New=Create new collection:")
			createButton = LOC("$$$/PhotoShelter/ExportDialog/CreateCollection/Create=Create Collection")
		else
			createTitle = LOC("$$$/PhotoShelter/ExportDialog/CreateGallery/Title=Create a new gallery")
			createText = LOC("$$$/PhotoShelter/ExportDialog/CreateGallery/New=Create new gallery:")
			createButton = LOC("$$$/PhotoShelter/ExportDialog/CreateGallery/Create=Create Gallery")
		end

		local c = f:column {
			bind_to_object = pt,
			spacing = f:control_spacing(),
			width = 500,
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = createText,
					width = share "lhs",
					alignment = "right"
				},
				f:edit_field {
					fill_horizontal = 1,
					value = bind "dlg_name",
					immediate = true
				}
			},
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/In=In:"),
					width = share "lhs",
					alignment = "right"
				},
				f:static_text {
					title = parent
				}
			},
			LrView.conditionalItem(pt.contributor == false, f:column {
				fill_horizontal = 1,
				spacing = f:control_spacing(),
				f:row {
					fill_horizontal = 1,
					spacing = f:label_spacing(),
					f:static_text {
						title = LOC("$$$/PhotoShelter/ExportDialog/Create/Description=Description:"),
						width = share "lhs",
						alignment = "right"
					},
					f:edit_field {
						fill_horizontal = 1,
						value = bind "dlg_description",
						height_in_lines = 5,
						wraps = true,
						allow_newlines = true
					}
				},
				f:row {
					fill_horizontal = 1,
					spacing = f:label_spacing(),
					f:static_text {
						title = LOC("$$$/PhotoShelter/ExportDialog/Create/VisibleTo=Visibility:"),
						width = share "lhs",
						alignment = "right"
					},
					f:column {
						fill_horizontal = 1,
						spacing = f:control_spacing(),
						LrView.conditionalItem(#pt.folderPath > 0, f:radio_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Inherit=Inherit from Parent"),
							value = bind "dlg_visibility",
							checked_value = "inherit"
						}),
						f:radio_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Everyone=Everyone"),
							value = bind "dlg_visibility",
							checked_value = "everyone"
						},
						f:radio_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Permission=Those with Permission"),
							value = bind "dlg_visibility",
							checked_value = "permission"
						},
						f:radio_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Me=Only Me"),
							value = bind "dlg_visibility",
							checked_value = "private"
						}
					}
				}
			})
		}

		local result = LrDialogs.presentModalDialog {
			title = createTitle,
			contents = c,
			actionVerb = createButton,
			save_frame = "newItemDialogPosition",
			accessoryView = f:row {
				fill_horizontal = 1,
				f:static_text {
					fill_horizontal = 1,
					bind_to_object = pt,
					title = bind "createStatus",
					text_color = LrColor("red"),
					place_vertical = 1,
				},
			},
			actionBinding = {
				enabled = bind {
					key = "dlg_name",
					bind_to_object = pt,
					transform = function(value, fromTable)
						if fromTable then
							return #value > 0
						else
							return LrBinding.kUnsupportedDirection
						end
					end
				},
			},
		}

		if result == "ok" then
			if not pt.contributor then
				createItemOnServer(pt, c, pt.dlg_name, pt.dlg_description, pt.dlg_visibility, forCollection)
			else
				pt.root = { { name = pt.dlg_name, type = "gallery", id = -1, listed = pt.showListed } }
				pt.itemIDMap[-1] = pt.root[1]
				pt.folderPath = { -1 }

				setUpScroller(pt)
			end
		end
	end)
end

-- The export dialog sections
function PhotoShelterUploadExportDialogSections.sectionsForTopOfDialog(_, pt)
	local f = LrView.osFactory()
	local bind = LrView.bind
	local share = LrView.share

	-- Account details
	local accountDetails = {
		title = LOC("$$$/PhotoShelter/ExportDialog/Title=PhotoShelter Account"),
		spacing = f:control_spacing(),
		-- Logo
		f:row {
			f:picture {
				value = LrPathUtils.child(_PLUGIN.path, "PS_Signature.png"),
				frame_width = 0,
				place_horizontal = 0.5,
			}
		},
		f:separator {
			fill_horizontal = 0.8,
			place_horizontal = 0.5,
		},
		f:group_box {
			title = LOC("$$$/PhotoShelter/ExportDialog/Authorization=Authorization"),
			fill_horizontal = 1,
			-- Email
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Email=Email:"),
					alignment = 'right',
					width = share 'labelWidth'
				},
				f:edit_field {
					value = bind 'accountEmail',
					immediate = true,
					fill_horizontal = 1,
				},
			},
			-- Password
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Password=Password:"),
					alignment = 'right',
					width = share 'labelWidth'
				},
				f:password_field {
					value = bind 'password',
					immediate = true,
					fill_horizontal = 1,
					height_in_lines = 1,
				},
			},
			-- Connect/Logout button
			f:view {
				place = "overlapping",
				fill_horizontal = 1,
				f:view {
					fill_horizontal = 1,
					visible = LrBinding.negativeOfKey("connected"),
					margin_bottom = 4,
					f:row {
						fill_horizontal = 1,
						spacing = f:label_spacing(),
						f:push_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Connect=Connect"),
							enabled = bind {
								keys = {
									{ key = "connecting" },
									{ key = "accountEmail" },
									{ key = "password" },
								},
								operation = function(binding, values, fromTable)
									if not values["connecting"] and values["accountEmail"] and #values["accountEmail"] > 0 and values["password"] and #values["password"] > 0 then
										return true
									else
										return false
									end
								end
							},
							place_horizontal = 1,
							action = function()
								connect(pt)
							end
						}
					}
				},
				f:view {
					fill_horizontal = 1,
					visible = bind "connected",
					margin_bottom = 4,
					f:row {
						fill_horizontal = 1,
						spacing = f:label_spacing(),
						f:push_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Logout=Logout"),
							place_horizontal = 1,
							action = function()
								logout(pt)
							end
						}
					}
				}
			}
		}
	}
	-- Organization details
	local orgAccountDetails = {
		title = LOC("$$$/PhotoShelter/ExportDialog/Title=PhotoShelter Account"),
		spacing = f:control_spacing(),
		-- Logo
		f:row {
			f:picture {
				value = LrPathUtils.child(_PLUGIN.path, "Libris_Signature.png"),
				frame_width = 0,
				place_horizontal = 0.5,
			}
		},
		f:separator {
			fill_horizontal = 0.8,
			place_horizontal = 0.5,
		},
		f:group_box {
			title = LOC("$$$/PhotoShelter/ExportDialog/Authorization=Authorization"),
			fill_horizontal = 1,
			-- Email
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Email=Email:"),
					alignment = 'right',
					width = share 'labelWidth'
				},
				f:edit_field {
					value = bind 'accountEmail',
					immediate = true,
					fill_horizontal = 1,
				},
			},
			-- Password
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Password=Password:"),
					alignment = 'right',
					width = share 'labelWidth'
				},
				f:password_field {
					value = bind 'password',
					immediate = true,
					fill_horizontal = 1,
					height_in_lines = 1,
				},
			},
			--  Organization
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Organization=Organization:"),
					alignment = 'right',
					width = share 'labelWidth',
					visible = bind "isMUMember"
				},
				f:popup_menu {
					value = bind "org",
					items = bind "orgs",
					fill_horizontal = 1,
					visible = bind "isMUMember"
				}
			},
			-- Photographer
			f:row {
				fill_horizontal = 1,
				spacing = f:label_spacing(),
				f:static_text {
					title = LOC("$$$/PhotoShelter/ExportDialog/Photographer=Photographer:"),
					alignment = 'right',
					width = share 'labelWidth',
					visible = bind {
						keys = {
							{ key = "isMUMember" },
							{ key = "org" },
						},
						operation = function(binding, values, fromTable)
							if values["isMUMember"] and values["org"] then
								return true
							else
								return false
							end
						end
					},
				},
				f:popup_menu {
					value = bind "photographer",
					items = bind "photographers",
					fill_horizontal = 1,
					visible = bind {
						keys = {
							{ key = "isMUMember" },
							{ key = "org" },
						},
						operation = function(binding, values, fromTable)
							if values["isMUMember"] and values["org"] then
								return true
							else
								return false
							end
						end
					},
				}
			},
			-- Connect/Logout button
			f:view {
				place = "overlapping",
				fill_horizontal = 1,
				f:view {
					fill_horizontal = 1,
					visible = LrBinding.negativeOfKey("connected"),
					margin_bottom = 4,
					f:row {
						fill_horizontal = 1,
						spacing = f:label_spacing(),
						f:push_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Connect=Connect"),
							enabled = bind {
								keys = {
									{ key = "connecting" },
									{ key = "accountEmail" },
									{ key = "password" },
								},
								operation = function(binding, values, fromTable)
									if not values["connecting"] and values["accountEmail"] and #values["accountEmail"] > 0 and values["password"] and #values["password"] > 0 then
										return true
									else
										return false
									end
								end
							},
							place_horizontal = 1,
							action = function()
								connect(pt)
							end
						}
					}
				},
				f:view {
					fill_horizontal = 1,
					visible = bind "connected",
					margin_bottom = 4,
					f:row {
						fill_horizontal = 1,
						spacing = f:label_spacing(),
						f:push_button {
							title = LOC("$$$/PhotoShelter/ExportDialog/Logout=Logout"),
							place_horizontal = 1,
							action = function()
								logout(pt)
							end
						}
					}
				}
			}
		}
	}
	-- Account Folder selection
	local folderSelection = {
		title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/Title=PhotoShelter Folder Selection"),
		spacing = f:control_spacing(),
		f:view {
			place = "overlapping",
			fill_horizontal = 1,
			-- Show log on progression information as we connect
			f:view {
				fill_horizontal = 1,
				visible = LrBinding.negativeOfKey("connected"),
				f:view {
					place_horizontal = 0.5,
					margin_top = 20,
					visible = bind "connecting",
					LRLib_Spinner.view(f, pt)
				},
			},
			-- Connected view
			f:view {
				fill_horizontal = 1,
				visible = bind "connected",
				spacing = f:control_spacing(),
				f:row {
					fill_horizontal = 1,
					-- Left/Right buttons and listed/unlisted selector
					f:push_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/ScrollLeft=<"),
						enabled = bind {
							key = scroller:bindIDForScrollOffset(),
							transform = function(value, fromTable)
								if fromTable then
									return value > 0
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						},
						action = function()
							pt.scrolling = true
							scroller:scroll(-1)
							pt.scrolling = false
						end
					},
					-- Listed/Unlisted
					f:spacer {
						fill_horizontal = 1
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/Listed=Listed on website"),
						bind_to_object = pt,
						value = bind "showListed",
						visible = LrBinding.negativeOfKey("contributor"),
						checked_value = true
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/Unlisted=Unlisted on website"),
						bind_to_object = pt,
						value = bind "showListed",
						visible = LrBinding.negativeOfKey("contributor"),
						checked_value = false
					},
					f:spacer {
						fill_horizontal = 1
					},
					-- Scroll right button
					f:push_button {
						place_horizontal = 1,
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/ScrollRight=>"),
						enabled = bind {
							keys = {
								{ key = scroller:bindIDForScrollOffset() },
								{ key = scroller:bindIDForNumberOfMappings() },
							},
							operation = function(binding, values, fromTable)
								if values[scroller:bindIDForNumberOfMappings()] - values[scroller:bindIDForScrollOffset()] <= NUM_COLS then
									return false
								else
									return true
								end
							end
						},
						action = function()
							pt.scrolling = true
							scroller:scroll(1)
							pt.scrolling = false
						end
					},
				},
				-- Scroller
				scroller:sectionForDialog(f, "ornamentationOnTop"),
				-- Export location
				f:row {
					fill_horizontal = 1,
					spacing = f:label_spacing(),
					f:static_text {
						title = LOC("$$$/PhotoShelter/ExportDialog/UploadinTo=Uploading to:"),
					},
					f:static_text {
						fill_horizontal = 1,
						truncation = "middle",
						font = "<system/bold>",
						title = bind {
							key = "folderPath",
							transform = function(folderPath, fromTable)
								if fromTable then
									local path = ""

									if #folderPath > 0 then
										if pt.showListed then
											path = "Listed"
										else
											path = "Unlisted"
										end

										for _, id in ipairs(folderPath) do
											if pt.itemIDMap[id] then
												path = path .. "/" .. pt.itemIDMap[id].name
											end
										end
									end

									return path
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						}
					}
				},
				-- New collection/gallery and refresh buttons
				f:row {
					fill_horizontal = 1,
					spacing = f:control_spacing(),
					f:push_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/NewGallery=New Gallery"),
						place_vertical = 0.5,
						enabled = bind {
							key = "folderPath",
							transform = function(folderPath, fromTable)
								-- Only enable this button when a collection is selected
								if fromTable then
									if #folderPath == 0 then
										return true
									end
									if pt.itemIDMap[folderPath[#folderPath]] and pt.itemIDMap[folderPath[#folderPath]].type == "collection" then
										return true
									end

									return false
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						},
						action = function()
							PhotoShelterUploadExportDialogSections.newItemDialog(pt, false, true)
						end
					},
					f:push_button {
						enabled = bind {
							key = "folderPath",
							transform = function(folderPath, fromTable)
								-- Only enable this button when a collection is selected
								if fromTable then
									if #folderPath == 0 then
										return true
									end

									if pt.itemIDMap[folderPath[#folderPath]] and pt.itemIDMap[folderPath[#folderPath]].type == "collection" then
										return true
									end

									return false
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						},
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/NewCollection=New Collection"),
						visible = LrBinding.negativeOfKey("contributor"),
						place_vertical = 0.5,
						action = function()
							PhotoShelterUploadExportDialogSections.newItemDialog(pt, true, true)
						end
					},
					f:spacer {
						fill_horizontal = 1
					},
					f:view {
						LRLib_Spinner.view(f, pt)
					},
					f:push_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/Refresh=Refresh"),
						place_vertical = 0.5,
						visible = LrBinding.negativeOfKey("contributor"),
						action = function()
							refreshFoldersInNewTask(pt)
						end
					},
				},
				f:separator {
					fill_horizontal = 1
				},
				f:group_box {
					fill_horizontal = 1,
					spacing = f:control_spacing(),
					f:checkbox {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/MakeSearchable=Make images publicly searchable"),
						value = bind "searchable"
					},
					f:static_text {
						title = " "
					},
					f:static_text {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists=If file already exists on the server:"),
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Upload=Upload file anyway (files of the same name can safely coexist)"),
						value = bind "ifexists",
						checked_value = "upload"
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Overwrite=Overwrite file"),
						value = bind "ifexists",
						checked_value = "overwrite"
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Rename=Rename file before uploading"),
						value = bind "ifexists",
						checked_value = "rename"
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Skip=Skip file (do not upload)"),
						value = bind "ifexists",
						checked_value = "skip"
					}
				}
			}
		}
	}
	-- Organization Folder selection
	local orgFolderSelection = {
		title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/Title=PhotoShelter Folder Selection"),
		spacing = f:control_spacing(),
		f:view {
			place = "overlapping",
			fill_horizontal = 1,
			-- Show log on progression information as we connect
			f:view {
				fill_horizontal = 1,
				visible = LrBinding.negativeOfKey("connected"),
				f:view {
					place_horizontal = 0.5,
					margin_top = 20,
					visible = bind "connecting",
					LRLib_Spinner.view(f, pt)
				},
			},
			-- Connected view
			f:view {
				fill_horizontal = 1,
				visible = bind "connected",
				spacing = f:control_spacing(),
				f:row {
					fill_horizontal = 1,
					-- Left/Right buttons and listed/unlisted selector
					f:push_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/ScrollLeft=<"),
						enabled = bind {
							key = scroller:bindIDForScrollOffset(),
							transform = function(value, fromTable)
								if fromTable then
									return value > 0
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						},
						action = function()
							pt.scrolling = true
							scroller:scroll(-1)
							pt.scrolling = false
						end
					},
					-- Scroll right button
					f:push_button {
						place_horizontal = 1,
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/ScrollRight=>"),
						enabled = bind {
							keys = {
								{ key = scroller:bindIDForScrollOffset() },
								{ key = scroller:bindIDForNumberOfMappings() },
							},
							operation = function(binding, values, fromTable)
								if values[scroller:bindIDForNumberOfMappings()] - values[scroller:bindIDForScrollOffset()] <= NUM_COLS then
									return false
								else
									return true
								end
							end
						},
						action = function()
							pt.scrolling = true
							scroller:scroll(1)
							pt.scrolling = false
						end
					},
				},
				-- Scroller
				scroller:sectionForDialog(f, "ornamentationOnTop"),
				-- Export location
				f:row {
					fill_horizontal = 1,
					spacing = f:label_spacing(),
					f:static_text {
						title = LOC("$$$/PhotoShelter/ExportDialog/UploadinTo=Uploading to:"),
					},
					f:static_text {
						fill_horizontal = 1,
						truncation = "middle",
						font = "<system/bold>",
						title = bind {
							key = "folderPath",
							transform = function(folderPath, fromTable)
								if fromTable then
									local path = "Root"

									if #folderPath > 0 then
										pt.showListed = false
										for _, id in ipairs(folderPath) do
											if pt.itemIDMap[id] then
												path = path .. "/" .. pt.itemIDMap[id].name
											end
										end
									end

									return path
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						}
					}
				},
				-- New collection/gallery and refresh buttons
				f:row {
					fill_horizontal = 1,
					spacing = f:control_spacing(),
					f:push_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/NewGallery=New Gallery"),
						place_vertical = 0.5,
						enabled = bind {
							key = "folderPath",
							transform = function(folderPath, fromTable)
								-- Only enable this button when a collection is selected
								if fromTable then
									if #folderPath == 0 then
										return true
									end

									if pt.itemIDMap[folderPath[#folderPath]] and pt.itemIDMap[folderPath[#folderPath]].type == "collection" then
										return true
									end

									return false
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						},
						action = function()
							PhotoShelterUploadExportDialogSections.newItemDialog(pt, false, true)
						end
					},
					f:push_button {
						enabled = bind {
							key = "folderPath",
							transform = function(folderPath, fromTable)
								-- Only enable this button when a collection is selected
								if fromTable then
									if #folderPath == 0 then
										return true
									end

									if pt.itemIDMap[folderPath[#folderPath]] and pt.itemIDMap[folderPath[#folderPath]].type == "collection" then
										return true
									end

									return false
								else
									return LrBinding.kUnsupportedDirection
								end
							end
						},
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/NewCollection=New Collection"),
						visible = LrBinding.negativeOfKey("contributor"),
						place_vertical = 0.5,
						action = function()
							PhotoShelterUploadExportDialogSections.newItemDialog(pt, true, true)
						end
					},
					f:spacer {
						fill_horizontal = 1
					},
					f:view {
						LRLib_Spinner.view(f, pt)
					},
					f:push_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/Refresh=Refresh"),
						place_vertical = 0.5,
						visible = LrBinding.negativeOfKey("contributor"),
						action = function()
							refreshFoldersInNewTask(pt)
						end
					},
				},
				f:separator {
					fill_horizontal = 1
				},
				f:group_box {
					fill_horizontal = 1,
					spacing = f:control_spacing(),
					f:checkbox {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/MakeSearchable=Make images publicly searchable"),
						value = bind "searchable"
					},
					f:static_text {
						title = " "
					},
					f:static_text {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists=If file already exists on the server:"),
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Upload=Upload file anyway (files of the same name can safely coexist)"),
						value = bind "ifexists",
						checked_value = "upload"
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Overwrite=Overwrite file"),
						value = bind "ifexists",
						checked_value = "overwrite"
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Rename=Rename file before uploading"),
						value = bind "ifexists",
						checked_value = "rename"
					},
					f:radio_button {
						title = LOC("$$$/PhotoShelter/ExportDialog/FolderSelection/IfExists/Skip=Skip file (do not upload)"),
						value = bind "ifexists",
						checked_value = "skip"
					}
				}
			}
		}
	}

	if pt.LR_isExportForPublish and PhotoShelterInfo.isLibris then
		return { orgAccountDetails }
	elseif pt.LR_isExportForPublish then
		return { accountDetails }
	elseif PhotoShelterInfo.isLibris then
		return { orgAccountDetails, orgFolderSelection }
	else
		return { accountDetails, folderSelection }
	end
end
