local LrPasswords = import 'LrPasswords'
local LrDialogs = import 'LrDialogs'
local LrMD5 = import 'LrMD5'
local LrStringUtils = import 'LrStringUtils'
local LrDate = import 'LrDate'
local LrApplication = import 'LrApplication'
local LrErrors = import 'LrErrors'

require "LRLib_DebugLog"
require "LRLib_Serialise"
require "Info"
require "JSON4Lua"
-- DO NOT ADD PHOTOSHELTER API
-- if u do, lua raise "chunk has many nested level" error.

PhotoShelterHelper = {}

-- Mapping of PhotoShelter IPTC name to LR ones
PhotoShelterHelper.IPTCToLR = {
	title = 'title',
	edit_status = '',
	urgency = '',
	category = 'iptcCategory',
	keyword = 'keywordTagsForExport',
	instructions = 'instructions',
	author = 'creator',
	position = 'creatorJobTitle',
	city = 'city',
	state = 'stateProvince',
	country = 'country',
	country_iso = 'isoCountryCode',
	headline = 'headline',
	trans_ref = 'jobIdentifier',
	credit = 'provider',
	source = 'source',
	copyright = 'copyright',
	caption = 'caption',
	caption_writer = 'descriptionWriter',
	filename = '',
	ci_address = 'creatorAddress',
	ci_city = 'creatorCity',
	ci_country = 'creatorCountry',
	ci_postal_code = 'creatorPostalCode',
	ci_region = 'creatorStateProvince',
	ci_email = 'creatorEmail',
	ci_phone = 'creatorPhone',
	ci_url = 'creatorUrl',
	genre = 'intellectualGenre',
	location = 'location',
	rights = 'rightsUsageTerms',
	scene = 'scene',
	subject_code = 'iptcSubjectCode',
	web_statement = 'copyrightInfoUrl',
	publish_date = '',
	sdate = '',
	rating = 'rating'
}

PhotoShelterHelper.mimeTypeLookup = {
--	arw = 'image/x-sony-arw',
--	cr2 = 'image/x-canon-cr2',
--	crw = 'image/x-canon-crw',
--	dcr = 'image/x-kodak-dcr',
--	dng = 'image/x-adobe-dng',
--	erf = 'image/x-epson-erf',
--	k25 = 'image/x-kodak-k25',
--	kdc = 'image/x-kodak-kdc',
--	mrw = 'image/x-minolta-mrw',
--	nef = 'image/x-nikon-nef',
--	orf = 'image/x-olympus-orf',
--	pef = 'image/x-pentax-pef',
--	raf = 'image/x-fuji-raf',
--	raw = 'image/x-panasonic-raw',
--	sr2 = 'image/x-sony-sr2',
--	srf = 'image/x-sony-srf',
--	x3f = 'image/x-sigma-x3f',
	raw = 'image/x-dcraw',
	dng = 'image/x-adobe-dng',
	tiff = 'image/tiff',
	tif = 'image/tiff',
	jpeg = 'image/jpeg',
	jpg = 'image/jpeg',
	psd = 'image/vnd.adobe.photoshop',
	png = 'image/png'
}

function PhotoShelterHelper.throwJsonDecode(text)
	local success, data = PhotoShelterHelper.JsonDecode(text)

	if not success then
		DBG('Something went wrong decoding JSON', {text=text}, 'ERROR')
		LrErrors.throwUserError('Something went wrong decoding JSON')
	end

	return data
end

function PhotoShelterHelper.JsonDecode(text)
	return pcall(function()
		return JSON4Lua.decode(text)
	end)
end

function PhotoShelterHelper.getPhotoMimeType(photo)
	local mime = photo:getFormattedMetadata('fileType')
	if not mime then
		return 'image/jpeg' -- default
	end
	local fileType = string.lower(mime)
	return PhotoShelterHelper.mimeTypeLookup[fileType] or 'image/jpeg'
end

function PhotoShelterHelper.getSubSecondFromExif(exif)
	local subSecond

	if exif then
		if exif['0x9291'] then
			subSecond = tonumber("0." .. exif['0x9291'].value)
		else
			if exif['0x9292'] then
				subSecond = tonumber("0." .. exif['0x9292'].value)
			end
		end
	end
	return subSecond
end

function PhotoShelterHelper.getCaptureDate(exif)
	local captureDate

	if exif then
		if exif['0x9004'] then
			captureDate = exif['0x9004'].value
		else
			if exif['0x9003'] then
				captureDate = exif['0x9003'].value
			end
		end
	end

	return captureDate
end

function PhotoShelterHelper.convertToTimestamp(date)
	local year, month, day, hour, min, sec, tz = date:match("(%d+)-(%d+)-(%d+) (%d+):(%d+):(%d+) (%a+)")
	--  EXIF timestamp can be malformed and not work in this pattern
	if year and month and day and hour and min and sec then
		local timestamp = LrDate.timeFromComponents(year, month, day, hour, min, sec, 0)
		return timestamp
	else
		return false
	end
end

function PhotoShelterHelper.getSubSecondFromImage(lrPhoto)
	local year, month, day, hour, min, sec, subSec

	local cDate = lrPhoto:getRawMetadata('dateTimeOriginalISO8601')
	if cDate ~= nil then
		local p = "(%d+)-(%d+)-(%d+)T(%d+):(%d+):(%d+).(%d+)"
		year, month, day, hour, min, sec, subSec = cDate:match(p)
		if year == nil then
			subSec = nil
		end
		if subSec then
			subSec = tonumber("0." .. subSec)
		end
	end

	return subSec
end

function PhotoShelterHelper.searchPhotos(filename, date, subSecond)
	local activeCatalog = LrApplication.activeCatalog()
	local photos = {}
	local searchCriteria = {}
	
	if not filename and not date then
		return nil
	end
	
	if filename then
		searchCriteria[#searchCriteria + 1] = {
			criteria = "filename",
			operation = "all",
			value = filename
		};
	end
	
	if date then
		searchCriteria[#searchCriteria + 1] = {
			criteria = "captureTime",
			operation = "==",
			value = date
		}
	end
	
	searchCriteria['combine'] = "union"
	
	local foundPhotos = activeCatalog:findPhotos {
		sort = "captureTime",
		ascending = true,
		searchDesc = searchCriteria
	}

	if #foundPhotos > 0 then
		if subSecond then
			for _, lrPhoto in pairs(foundPhotos) do
				local subsec = PhotoShelterHelper.getSubSecondFromImage(lrPhoto)
				if subSecond == subsec then
					table.insert(photos, lrPhoto)
				end
			end
		else
			photos = foundPhotos
		end

		local results = {}

		for _, photo in pairs(photos) do
			local name = photo:getFormattedMetadata('fileName')
			local _, _, basename = string.find(name, '(.+)%.')
			if filename:find(basename, 1, true) then
				table.insert(results, photo)
			end
		end

		if #results ~= 0 then
			photos = results
		end
	end

	return photos
end

function PhotoShelterHelper.calculateMatch(photoShelterImage, lrPhoto, iptc)
	local score = 0

	-- local isVirtual = lrPhoto:getRawMetadata('isVirtualCopy')
	-- if isVirtual then
	--    score = score + 1
	-- end

	local aspectRatio = lrPhoto:getRawMetadata('aspectRatio')
	local targetAspectRatio = photoShelterImage.Image.width / photoShelterImage.Image.height
	local delta = math.abs(targetAspectRatio - aspectRatio)
	delta = math.min(delta, 10)
	score = score + 10 - delta

	if iptc then
		local title = lrPhoto:getFormattedMetadata('title')

		if iptc.title == title then
			score = score + 10
		end

		local caption = lrPhoto:getFormattedMetadata('caption')

		if iptc.caption == caption then
			score = score + 10
		end
	end

	score = score * 100 / 31

	return score
end

function PhotoShelterHelper.isCollection(set)
	return set:type() == 'LrPublishedCollectionSet' and true or false
end

function PhotoShelterHelper.tableLength (table)
	local count = 0
	for _ in pairs(table) do
		count = count + 1
	end
	return count
end

function PhotoShelterHelper.tableConcat(t1, t2)
	for i = 1, #t2 do
		t1[#t1 + 1] = t2[i]
	end
	return t1
end

function PhotoShelterHelper.contains (table, elem)
	for _, i in ipairs(table) do
		if i == elem then
			return true
		end
	end
	return false
end

function PhotoShelterHelper.removeByKey(table, key)
	local element = table[key]
	table[key] = nil
	return element
end

function PhotoShelterHelper.setLRParent(parent)
	if parent[2] ~= 'root' then
		return parent[2]
	else
		return nil
	end
end

function PhotoShelterHelper.setupPSChild(child)
	local child = child.Gallery or child.Collection
	local childId = child.gallery_id or child.collection_id
	local childVisibility = function()
		if child.access_inherit and child.access_inherit == "t" then
			return "inherit"
		else
			return (child.Visibility and child.Visibility.mode) or child.mode or "private"
		end
	end
	return {
		name = child.name,
		id = childId,
		description = child.description,
		listed = child.f_list == "t",
		visibility = childVisibility()
	}
end

-- Setup object of with LR set setting names from PS Child
function PhotoShelterHelper.setLocalSettings (child)
	return {
		Description = child.description,
		ListedOnWebsite = child.listed,
		Visibility = child.visibility
	}
end

function PhotoShelterHelper.hasPhotos(collectionOrSet)
	if collectionOrSet:type() == 'LrPublishedCollectionSet' then
		for _, c in ipairs(collectionOrSet:getChildCollections()) do
			if #c:getPhotos() > 0 then
				return true
			end
		end

		for _, cs in ipairs(collectionOrSet:getChildCollectionSets()) do
			if PhotoShelterHelper.hasPhotos(cs) then
				return true
			end
		end
	else
		if #collectionOrSet:getPhotos() > 0 then
			return true
		end
	end

	return false
end

--- Return a new name based on the lookup table, name, parent and suffix
--- eg: My Collection (2) if already exists another in the same parent.
function PhotoShelterHelper.suffixName (lookupTable, name, lrParent, scopeData)
	local newName = name

	while lookupTable[LrStringUtils.lower(newName)] do
		if lookupTable[LrStringUtils.lower(newName)]:getParent() ~= lrParent then
			break
		end
		if not scopeData.suffix then
			scopeData.suffix = 0
		end
		newName = name .. " (" .. scopeData.suffix .. ")"
		scopeData.suffix = scopeData.suffix + 1
	end

	return newName
end

function PhotoShelterHelper.createSiblingNamesTable (lrParent, publishService)
	local siblingNames = {}
	local siblings
	local publishService = publishService

	if lrParent ~= nil then
		siblings = lrParent:getChildren()
	else
		local siblingCollections = publishService:getChildCollectionSets()
		local siblingGalleries = publishService:getChildCollections()
		siblings = PhotoShelterHelper.tableConcat(siblingCollections, siblingGalleries)
	end

	for _, sibling in ipairs(siblings) do
		siblingNames[LrStringUtils.lower(sibling:getName())] = sibling
	end

	return siblingNames
end

-- Clear plugin related metadata
function PhotoShelterHelper.clearMetadata(lrPhoto)
	lrPhoto:setPropertyForPlugin(_PLUGIN, 'id', nil)
	lrPhoto:setPropertyForPlugin(_PLUGIN, 'onServer', nil)
	-- todo: check which is the purpose of below properties
	lrPhoto:setPropertyForPlugin(_PLUGIN, 'propRelease', nil)
	lrPhoto:setPropertyForPlugin(_PLUGIN, 'modelRelease', nil)
end

-- Return the ID and Name of a Lr Collection (PS Gallery)
function PhotoShelterHelper.pubCollectionNameAndId(lrCollectionOrSet)
	local name = lrCollectionOrSet:getName()
	local id = lrCollectionOrSet:getRemoteId() or 'MISSING-REMOTE-ID'
	local type = PhotoShelterHelper.isCollection(lrCollectionOrSet) and "Collection" or "Gallery"
	return type .. " " .. name .. " (" .. id .. ")"
end

function PhotoShelterHelper.getRemoteUrl(listed, collectionId, galleryId, photoId, hasLibrisRole)
	local url

	if PhotoShelterInfo.isLibris and hasLibrisRole then
		url = "http://www.photoshelter.com/mu/libris/images/#"
	elseif PhotoShelterInfo.isLibris then
		url = 'http://www.photoshelter.com/mu/images/#'
	else
		url = "http://www.photoshelter.com/mem/images/#"
	end

	if collectionId then
		url = url .. '/' .. collectionId -- concatenation
	elseif hasLibrisRole ~= true then
		if listed then
			url = url .. '/l'
		else
			url = url .. '/u'
		end
	end

	if galleryId then
		url = url .. '/' .. galleryId
	end

	if photoId then
		url = url .. '/' .. photoId
	end

	return url
end

function PhotoShelterHelper.getUserLibrisRole(authToken)
	local success, userId, orgId = PhotoShelterAPI.getUserId(authToken)
	if not success then
		local error = 'Unable to retrieve User ID and Organization ID with current session.'
		LrDialogs.showError(error)
		LrErrors.throwUserError(error)
	end

	if orgId == nil then
		return false
	end

	local secondSuccess, userOrgRole = PhotoShelterAPI.getUserOrgRole(authToken, orgId, userId)
	if not secondSuccess then
		local error = 'Unable to retrieve User Organization Roles with current session.'
		LrDialogs.showError(error)
		LrErrors.throwUserError(error)
	end

	return userOrgRole and true or false
end

function PhotoShelterHelper.failureHandler(status, theError)
	local trimmed = theError:match('<AgErrorText>(.-)</AgErrorText>') or theError
	if trimmed then
		trimmed = trimmed:gsub(".*NOLINE:", "")
		LrDialogs.message(trimmed, nil, "critical")
	else
		local theError = theError:gsub(".*NOLINE:", "")
		LrDialogs.message(theError, nil, "critical")
	end
end

function PhotoShelterHelper.authHeaders(authToken)
	local headers = {
		{ field = "X-PS-Api-Key", value = "4I4HIuxYXTw" }
	}

	if authToken and #authToken > 0 then
		table.insert(headers, { field = "X-PS-Auth-Token", value = authToken })
	end

	return headers
end

function PhotoShelterHelper.setUpDebug(pt, doNotReset)
	DBG = function(msg, debug, lvl)
		if not lvl then
			lvl = 'INFO'
		end

		msg = lvl .. " - " .. msg

		if debug then
			LRLib_DebugLog.log(pt, msg)
			LRLib_DebugLog.logRaw(pt, LRLib_TableEx.dump(debug or {}))
		else
			LRLib_DebugLog.logRaw(pt, msg)
		end
	end

	if not doNotReset then
		LRLib_DebugLog.reset()
	end

	DBG('Log setup done!', nil, 'INFO')
end

function PhotoShelterHelper.booleanToString(value)
	return value and 't' or 'f'
end

function PhotoShelterHelper.validateLogon(settings)
	local password = LrPasswords.retrieve(settings['accountEmail'], settings['accountEmail'])
	local success, dataOrMsg = PhotoShelterAPI.logon(settings.accountEmail, password)

	if success and dataOrMsg then
		local token = dataOrMsg.token
		-- Log on to the organisation if there is one
		if settings.org then
			success, dataOrMsg = PhotoShelterAPI.logonOrganisation(settings.org, token)
		end

		-- If successful, return the token
		if success then
			return token
		end
	end

	-- Unable to log on, tell the user
	LrErrors.throwUserError(dataOrMsg)
end

function PhotoShelterHelper.setImageDataForPublishService(publishService, lrPhoto, key, data)
	-- TODO: check wtf is this unique global name
	local uniqueGlobalName = "Random_" .. math.random(10000, 99999)
	local serialisedPublishData = lrPhoto:getPropertyForPlugin(_PLUGIN, "publishData")

	if serialisedPublishData then
		local success, theError = LrTasks.pcall(loadstring(uniqueGlobalName .. " = " .. serialisedPublishData))
		if not success then
			MSG(theError)
		end
	end

	-- TODO: check wtf is _G
	local publishData = _G[uniqueGlobalName] or {}

	local allData = publishData[publishService.localIdentifier] or {}
	allData[key] = data

	publishData[publishService.localIdentifier] = allData
	serialisedPublishData = LRLib_Serialise.serialise(publishData)

	local catalog = LrApplication.activeCatalog()

	if catalog.hasPrivateWriteAccess then
		lrPhoto:setPropertyForPlugin(_PLUGIN, "publishData", serialisedPublishData)
	else
		catalog:withPrivateWriteAccessDo(
				function()
					lrPhoto:setPropertyForPlugin(_PLUGIN, "publishData", serialisedPublishData)
				end,
				{ timeout = 5 }
		)
	end
end

function PhotoShelterHelper.getSerialisedDevelopSettings(lrPhoto)
	-- The develop settings will change if the photo is modified within Lightroom
	local developSettings = lrPhoto:getDevelopSettings()
	developSettings = LRLib_Serialise.serialise(developSettings)

	-- The last edit time will change if the photo is modified outside of Lightroom
	local lastEditTime = lrPhoto:getRawMetadata("dateTime") or ""

	-- Return an MD5 combining these two pieces of information
	return LrMD5.digest(developSettings .. lastEditTime)
end

function PhotoShelterHelper.getImageDataForPublishService(publishService, lrPhoto, key)
	-- TODO: check wtf is this unique global name
	local uniqueGlobalName = "Random_" .. math.random(10000, 99999)
	local serialisedPublishData = lrPhoto:getPropertyForPlugin(_PLUGIN, "publishData")

	if serialisedPublishData then
		local success, theError = LrTasks.pcall(loadstring(uniqueGlobalName .. " = " .. serialisedPublishData))
		if not success then
			-- todo: check if it prompt two dialogs or just one.
			MSG(theError)
			LrErrors.throwUserError(theError)
		end
	end

	-- TODO: check wtf is _G
	local publishData = _G[uniqueGlobalName] or {}
	local allData = publishData[publishService.localIdentifier] or {}

	return allData[key]
end

-- Iterate over the hierarchy parentSet or get the child collection sets from retrieved publishService
-- @param publishSettings (LrPublishService) The publish service object.
-- @param parentSet (TODO) Check what it is
-- @param callback (function) function to be called with the LrPhoto as first argument.
function PhotoShelterHelper.recurseHierarchy(publishService, parentSet, callback)
	if parentSet == nil or parentSet:type() == 'LrPublishedCollectionSet' then
		-- Get the local collection sets (PhotoShelter collections)
		local collectionSets

		if parentSet then
			collectionSets = parentSet:getChildCollectionSets()
		else
			collectionSets = publishService:getChildCollectionSets()
		end

		-- Get the local collections (PhotoShelter galleries)
		local collections

		if parentSet then
			collections = parentSet:getChildCollections()
		else
			collections = publishService:getChildCollections()
		end

		-- Recurse into collection sets
		for _, collectionSet in ipairs(collectionSets) do
			PhotoShelterHelper.recurseHierarchy(publishService, collectionSet, callback)
		end

		-- Apply the callback to each photo
		for _, publishedCollection in ipairs(collections) do
			local photos = publishedCollection:getPublishedPhotos()

			for _, publishedPhoto in ipairs(photos) do
				callback(publishedPhoto)
			end
		end
	elseif parentSet:type() == 'LrPublishedCollection' then
		local photos = parentSet:getPublishedPhotos()

		for _, publishedPhoto in ipairs(photos) do
			callback(publishedPhoto)
		end
	end
end

-- Given a count of the number of instances of a published photo that will be removed from a publish service,
-- this function will remove the photoshelter metadata for photos that will be completely deleted from the
-- publish service.
--
-- photosWithCount should be as follows:
--
-- {
--    <lrPhotoLocalIdentifier> -> { count = <number that will be deleted>, lrPhoto = <lrPhoto> }
--    ....
-- }
function PhotoShelterHelper.updateMetadata(publishService, photosWithCount)
	-- For each photo, if the total count is the same as the count for the number that will be deleted then the clear the metadata
	function doWork()
		for localIdentifier, info in pairs(photosWithCount) do
			local collections = info.lrPhoto:getContainedPublishedCollections()
			local count = 0

			for _, publishedCollection in pairs(collections) do
				local publishServiceIdentifier = publishedCollection:getService().localIdentifier
				for _, pubPhoto in pairs(publishedCollection:getPublishedPhotos()) do
					if publishService.localIdentifier == publishServiceIdentifier then
						if pubPhoto:getPhoto().localIdentifier == localIdentifier and pubPhoto:getRemoteId() then
							count = count + 1
						end
					end
				end
			end

			if count <= info.count then
				info.lrPhoto:setPropertyForPlugin(_PLUGIN, 'id', nil)
				info.lrPhoto:setPropertyForPlugin(_PLUGIN, 'onServer', nil)
			end
		end
	end

	LrApplication.activeCatalog():withPrivateWriteAccessDo(function()
		doWork()
	end, { timeout = 5 })
end
