local LrFunctionContext = import 'LrFunctionContext'
local LrApplication = import 'LrApplication'
local LrPathUtils = import 'LrPathUtils'
local LrDialogs = import 'LrDialogs'
local LrErrors = import 'LrErrors'
local LrHttp = import 'LrHttp'
local LrFileUtils = import 'LrFileUtils'
local LrView = import "LrView"
local LrBinding = import 'LrBinding'
local LrProgressScope = import 'LrProgressScope'

local share = LrView.share

require "Trapper"
require "Info"
require "PhotoShelterAPI"
require "PhotoShelterHelper"
require "PhotoShelterSync"
require "LRLib_TableEx"
require "LRLib_Serialise"
require "LRLib_DebugLog"

local publishServiceProvider = {}

-- This don't really mather because it is deleted on publishServiceProvider.didCreateNewPublishService
function publishServiceProvider.getCollectionBehaviorInfo(publishSettings)
	return {
		defaultCollectionName = LOC("$$$/PhotoShelter/DefaultCollectionName/default=Default Collection"),
		defaultCollectionCanBeDeleted = true,
		canAddCollection = true,
		maxCollectionSetDepth = 10
	}
end

-- This plug-in defined callback function is called when the user chooses the "Go to Published Collection" context-menu item.
-- If this function is not provided, Lightroom uses the URL	recorded for the published collection via
-- exportSession:recordRemoteCollectionUrl.
-- This is not a blocking call. It is called from within a task created
-- using the LrTasks namespace. In most cases, you should not need to start your own task within this function.
-- @name publishServiceProvider.goToPublishedCollection
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- publishedCollectionInfo: (LrPublishedCollectionInfo) An object containing  publication information for this published collection.
-- photo: (LrPhoto) The photo object.
-- publishedPhoto: (LrPublishedPhoto) The object that contains information previously recorded about this photo's publication.
-- remoteId: (string or number) The ID for this published collection that was stored via exportSession:recordRemoteCollectionId
-- remoteUrl: (optional, string) The URL, if any, that was recorded for the published collection via exportSession:recordRemoteCollectionUrl
function publishServiceProvider.goToPublishedCollection(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)
	local url = info['remoteUrl']
	local remoteId = info['remoteId']
	local collectionId = nil
	local galleryId = nil
	local parentInfo = info.publishedCollectionInfo.parents
	local set = info.publishedCollection
	local listed
	local isCollection = set:type() == 'LrPublishedCollectionSet'

	if isCollection then
		listed = set:getCollectionSetInfoSummary().collectionSettings.ListedOnWebsite
	else
		listed = set:getCollectionInfoSummary().collectionSettings.ListedOnWebsite
	end

	local authToken = PhotoShelterHelper.validateLogon(publishSettings)
	local hasLibrisRole = PhotoShelterHelper.getUserLibrisRole(authToken)

	if parentInfo and #parentInfo >= 1 and set.Collection ~= true then
		galleryId = remoteId
		collectionId = parentInfo[#parentInfo].remoteCollectionId
	else
		-- If the ID begins with a C then it's a collection
		local pos = string.find(remoteId, 'C')

		if pos == 1 then
			collectionId = remoteId
		else
			galleryId = remoteId
		end
	end

	url = PhotoShelterHelper.getRemoteUrl(listed, collectionId, galleryId, nil, hasLibrisRole)

	if url ~= nil then
		LrHttp.openUrlInBrowser(url)
	end
end

-- (optional) This plug-in defined callback function is called when the user chooses the
-- "Go to Published Photo" context-menu item.
-- <p>If this function is not provided, Lightroom invokes the URL recorded for the published photo via
-- <a href="LrExportRendition.html#exportRendition:recordPublishedPhotoUrl"><code>exportRendition:recordPublishedPhotoUrl</code></a>.</p>
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.goToPublishedPhoto
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>publishedCollectionInfo</b>: (<a href="LrPublishedCollectionInfo.html"><code>LrPublishedCollectionInfo</code></a>)
--  	An object containing  publication information for this published collection.</li>
-- <li><b>photo</b>: (<a href="LrPhoto.html"><code>LrPhoto</code></a>) The photo object. </li>
-- <li><b>publishedPhoto</b>: (<a href="LrPublishedPhoto.html"><code>LrPublishedPhoto</code></a>)
-- 	The object that contains information previously recorded about this photo's publication.</li>
-- <li><b>remoteId</b>: (string or number) The ID for this published photo
-- 	that was stored via <a href="LrExportRendition.html#exportRendition:recordPublishedPhotoId"><code>exportRendition:recordPublishedPhotoId</code></a></li>
-- <li><b>remoteUrl</b>: (optional, string) The URL, if any, that was recorded for the published photo via
-- <a href="LrExportRendition.html#exportRendition:recordPublishedPhotoUrl"><code>exportRendition:recordPublishedPhotoUrl</code></a>.</li>
-- </ul>
function publishServiceProvider.goToPublishedPhoto(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)
	local authToken = PhotoShelterHelper.validateLogon(publishSettings)
	local url = info['remoteUrl']
	local photoId = info['remoteId']
	local collectionId = nil
	local galleryId = info.publishedCollectionInfo.remoteId
	local parentInfo = info.publishedCollectionInfo.parents
	local success, listed = PhotoShelterAPI.getGalleryListed(galleryId, authToken)

	if success == false then
		LrErrors.throwUserError('We are unable to read Gallery ' .. info.publishedCollectionInfo.name .. ' information from PhotoShelter. Sorry try again.')
	end

	if parentInfo and #parentInfo >= 1 then
		collectionId = parentInfo[#parentInfo].remoteCollectionId
	end

	local hasLibrisRole = PhotoShelterHelper.getUserLibrisRole(authToken)

	url = PhotoShelterHelper.getRemoteUrl(listed, collectionId, galleryId, photoId, hasLibrisRole)

	if url ~= nil then
		LrHttp.openUrlInBrowser(url)
	else
		LrErrors.throwUserError('We are unable to prepare the Photo URL, please try again.')
	end
end

-- This plug-in defined callback function is called when the user creates
-- a new publish service via the Publish Manager dialog. It allows your plug-in
-- to perform additional initialization.
-- This is not a blocking call. It is called from within a task created
-- using the LrTasks namespace. In most cases, you should not need to start your own task within this function.
-- @name publishServiceProvider.didCreateNewPublishService
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- connectionName: (string) the name of the newly-created service
-- publishService: (LrPublishService) The publish service object.
function publishServiceProvider.didCreateNewPublishService(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	Trapper.TRAP(function()
		PhotoShelterSync.destroyDefaultCollection(info.publishService)
	end)
end

-- This plug-in defined callback function is called when the user creates
-- a new publish service via the Publish Manager dialog. It allows your plug-in
-- to perform additional initialization.
-- This is not a blocking call. It is called from within a task created
-- using theLrTasks namespace. In most cases, you should not need to start your own task within this function.
-- @name publishServiceProvider.didUpdatePublishService
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- connectionName: (string) the name of the newly-created service
-- nPublishedPhotos: (number) how many photos are currently published on the service
-- publishService: (LrPublishService) The publish service object.
-- changedMoreThanName: (boolean) true if any setting other than the name (description) has changed
function publishServiceProvider.didUpdatePublishService(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	local callback = function(publishedPhoto)
		local lrPhoto = publishedPhoto:getPhoto()
		PhotoShelterHelper.setImageDataForPublishService(info.publishService, lrPhoto, "developSettings", "")
	end
	-- Clear our stored develop settings so that images get published with new publish settings
	-- (If we don't do this then we'll only update the IPTC for images that have already been published)
	local function clearDevelopSettings(parentSet)
		PhotoShelterHelper.recurseHierarchy(info.publishService, parentSet, callback)
	end

	if info.changedMoreThanName then
		Trapper.TRAP(function()
			clearDevelopSettings()
		end)
	end
end

-- This plug-in defined callback function is called when the user
-- has attempted to delete one or more published collections defined by your
-- plug-in from Lightroom. It provides an opportunity for you to customize the
-- confirmation dialog.
-- This is not a blocking call. It is called from within a task created
-- using the LrTasks namespace. In most cases, you should not need to start your own task within this function.
-- @name publishServiceProvider.shouldDeletePublishedCollection
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- collections: (array of LrPublishedCollection or LrPublishedCollectionSet) The published collection objects.
-- nPhotos: (number) The number of photos contained in the published collection. Only present if there is a single published collection to be deleted.
-- nChildren: (number) The number of child collections contained within the	published collection set. Only present if there is a single published collection set to be deleted.
-- hasItemsOnService: (boolean) True if one or more photos have been published through the collection(s) to be deleted.
-- @return (string) "ignore", "cancel", "delete", or nil
-- (If you return nil, Lightroom's default dialog will be displayed.)
function publishServiceProvider.shouldDeletePublishedCollection(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)
	DBG('publishServiceProvider.shouldDeletePublishedCollection', nil, 'DEBUG')

	local retval = nil
	local response = LrDialogs.confirm(
			LOC("$$$/PhotoShelter/DeletePublishedCollection/Message=Delete a Published PhotoShelter Gallery"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/Info=Deleting an gallery on PhotoShelter also deletes the contained images\nThis can't be undone!"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionOk=Delete"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionCancel=Cancel"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionOther=Leave Gallery on PhotoShelter")
	)

	if response == 'ok' then
		retval = 'delete'
	end

	if response == 'other' then
		retval = 'ignore'
	end

	if response == 'cancel' then
		retval = 'cancel'
	end

	return retval
end

-- (optional) This plug-in defined callback function is called when the user
-- has attempted to delete one or more photos from the Lightroom catalog that are
-- published through your service. It provides an opportunity for you to customize
-- the confirmation dialog.
-- <p>Do not use this hook to actually delete photo(s). Instead, if the user
-- confirms the deletion for all relevant services. Lightroom will call
-- <a href="#publishServiceProvider.deletePhotosFromPublishedCollection"><code>deletePhotosFromPublishedCollection</code></a>
-- for that purpose.
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.shouldDeletePhotosFromServiceOnDeleteFromCatalog
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param nPhotos (number) The number of photos that are being deleted. At least
-- one of these photos is published through this service; some may only be published
-- on other services or not published at all.
-- @return (string) What action should Lightroom take?
-- <ul>
-- <li><b>"ignore"</b>: Leave the photos on the service and simply forget about them.</li>
-- <li><b>"cancel"</b>: Stop the attempt to delete the photos.
-- <li><b>"delete"</b>: Have Lightroom delete the photos immediately from the service.
-- (Your plug-in will receive a call to its
-- <a href="#publishServiceProvider.deletePhotosFromPublishedCollection"><code>deletePhotosFromPublishedCollection</code></a>
-- in this case.)</li>
-- <li><b>nil</b>: Allow Lightroom's built-in confirmation dialog to be displayed.</li>
-- </ul>
function publishServiceProvider.shouldDeletePhotosFromServiceOnDeleteFromCatalog(publishSettings, nPhotos)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	DBG('shouldDeletePhotosFromServiceOnDeleteFromCatalog')
	local retval = nil

	local response = LrDialogs.confirm(
			LOC("$$$/PhotoShelter/DeletePublishedCollection/Message=Delete Published Photos"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/Info=You are trying to delete one or more Photos that are already published in PhotoShelter, do you wanna proceed with this operation?"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionOk=Yes, Delete"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionCancel=No, Cancel")
	)

	if response == 'ok' then
		retval = 'delete'
	end

	if response == 'cancel' then
		retval = 'cancel'
	end

	return retval
end

-- This plug-in defined callback function is called when one or more photos
-- have been removed from a published collection and need to be removed from
-- the service. If the service you are supporting allows photos to be deleted
-- via its API, you should do that from this function.
-- As each photo is deleted, you should call the deletedCallback
-- function to inform Lightroom that the deletion was successful. This will cause
-- Lightroom to remove the photo from the "Delete Photos to Remove" group in the
-- Library grid.
-- This is not a blocking call. It is called from within a task created
-- using the LrTasks namespace. In most cases, you should not need to start your own task within this function.
-- @name publishServiceProvider.deletePhotosFromPublishedCollection
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param arrayOfPhotoIds (table) The remote photo IDs that were declared by this plug-in
-- when they were published.
-- @param deletedCallback (function) This function must be called for each photo ID
-- as soon as the deletion is confirmed by the remote service. It takes a single
-- argument: the photo ID from the arrayOfPhotoIds array.
-- @param localCollectionId (number) The local identifier for the collection for which
-- photos are being removed.
function publishServiceProvider.deletePhotosFromPublishedCollection(publishSettings, arrayOfPhotoIds, deletedCallback, localCollectionId)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	Trapper.TRAP(function(context)
		DBG('publishServiceProvider.deletePhotosFromPublishedCollection', nil, 'DEBUG')
		local authToken = PhotoShelterHelper.validateLogon(publishSettings)
		local progressScope = LrProgressScope(
				{
					title = LOC "$$$/PhotoShelter/Task/Progress/Delete=Deleting Published Photos",
					functionContext = context,
				}
		)

		progressScope:setPortionComplete(0, #arrayOfPhotoIds)

		local publishedCollection = LrApplication.activeCatalog():getPublishedCollectionByLocalIdentifier(localCollectionId)
		local publishService = publishedCollection:getService()

		local remote_collectionId = publishSettings.LR_publishedCollectionInfo.remoteId

		for i, photoId in ipairs(arrayOfPhotoIds) do
			local currentStep = 0
			local totalSteps = 4
			DBG('deletePhotosFromPublishedCollection ' .. photoId, nil, 'DEBUG')
			if progressScope:isCanceled() then
				break
			end

			repeat
				progressScope:setCaption(LOC("$$$/PhotoShelter/Task/Progress/ReadingPhoto=Reading Photo information"))
				local photo
				-- get photo reference
				for _, pubPhoto in pairs(publishedCollection:getPublishedPhotos()) do
					if pubPhoto:getRemoteId() == photoId then
						photo = pubPhoto:getPhoto()
						break
					end
				end

				local suffix = photoId:sub(-8)
				-- dunno why they check this but will leave because didn't hurt
				-- maybe that is added by LRC when something fail but it till allow user to delete the photo.
				if suffix == "_removed" then
					-- update a progress
					deletedCallback(photoId)
					progressScope:setPortionComplete(i, #arrayOfPhotoIds)
					break
				end
				currentStep = currentStep + 1
				progressScope:setPortionComplete(i + (currentStep / totalSteps), #arrayOfPhotoIds)

				progressScope:setCaption(LOC("$$$/PhotoShelter/Task/Progress/CheckPhotoExists=Checking Photo in PhotoShelter"))
				local exists = PhotoShelterAPI.queryImageExists(photoId, authToken)

				if not exists then
					-- clean and allow LRC to delete it
					publishService.catalog:withPrivateWriteAccessDo(
							function(context)
								photo:setPropertyForPlugin(_PLUGIN, 'id', nil)
								photo:setPropertyForPlugin(_PLUGIN, 'onServer', false)
							end
					)
					progressScope:setPortionComplete(i, #arrayOfPhotoIds)
					deletedCallback(photoId)
					break
				end
				currentStep = currentStep + 1
				progressScope:setPortionComplete(i + (currentStep / totalSteps), #arrayOfPhotoIds)

				-- get parent galleries from photoshelter
				progressScope:setCaption(LOC("$$$/PhotoShelter/Task/Progress/CheckPhotoGalleries=Checking Photo Galleries in PhotoShelter"))
				local success, galleriesOrMessage = PhotoShelterAPI.getParentGalleries(photoId, authToken)
				if not success then
					-- could not be removed so prevent to insert it on removed table
					DBG('Unable to read parent galleries for image ' .. photoId, nil, 'ERROR')
					progressScope:setPortionComplete(i, #arrayOfPhotoIds)
					break
				end

				local galleryFound

				if #galleriesOrMessage > 0 then
					for _, gallery in pairs(galleriesOrMessage) do
						if gallery.gallery_id == remote_collectionId then
							galleryFound = true
							break
						end
					end
				end

				-- if we found this gallery on the current photoshelter parent galleries we proceed to remove photo from gallery
				-- otherwise we just update the galleryReferenceCount and call the deletedCallback
				if galleryFound then
					local result, msg = PhotoShelterAPI.removeImageFromGallery(remote_collectionId, photoId, authToken)

					if not result then
						DBG('We was unable to remove the photo ' .. photoId .. ' from gallery ' .. remote_collectionId,
								{ error = msg }, 'ERROR')
						progressScope:setPortionComplete(i, #arrayOfPhotoIds)
						break
					end
				end
				currentStep = currentStep + 1
				progressScope:setPortionComplete(i + (currentStep / totalSteps), #arrayOfPhotoIds)

				progressScope:setCaption(LOC("$$$/PhotoShelter/Task/Progress/CheckPhotoGalleries=Deleting file: ^1", photoId))
				-- total of galleries before remove, less the currently removed, if it is 0 we should remove the photo from photoshelter.
				local shouldDeleteFile = (#galleriesOrMessage - 1) < 1

				if not shouldDeleteFile then
					-- photo is still related in another gallery
					progressScope:setPortionComplete(i, #arrayOfPhotoIds)
					deletedCallback(photoId)
					break
				end

				-- remove the image on PhotoShelter because
				local success, error = PhotoShelterAPI.removeImage(photoId, authToken)

				if not success then
					DBG('Unable to remove image ' .. photoId .. 'from PhotoShelter', { error = error, photoId = photoId },
							'ERROR')
					progressScope:setPortionComplete(i, #arrayOfPhotoIds)
					LrErrors.throwUserError(error)
				end

				publishService.catalog:withPrivateWriteAccessDo(
						function()
							photo:setPropertyForPlugin(_PLUGIN, 'onServer', nil)
							photo:setPropertyForPlugin(_PLUGIN, 'id', nil)
						end
				)

				currentStep = currentStep + 1
				progressScope:setPortionComplete(i + (currentStep / totalSteps), #arrayOfPhotoIds)
				deletedCallback(photoId)
			until true
		end
	end)
end

-- (optional) This plug-in defined callback function is called whenever a new
-- publish service is created and whenever the settings for a publish service
-- are changed. It allows the plug-in to specify which metadata should be
-- considered when Lightroom determines whether an existing photo should be
-- moved to the "Modified Photos to Re-Publish" status.
-- This is a blocking call.
-- @name publishServiceProvider.metadataThatTriggersRepublish
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @return (table) A table containing one or more of the following elements
-- as key, Boolean true or false as a value, where true means that a change
-- to the value does trigger republish status, and false means changes to the
-- value are ignored:
--
-- default: All built-in metadata that appears in XMP for the file.
-- You can override this default behavior by explicitly naming any of these
-- specific fields:
--
-- rating
-- label
-- title
-- caption
-- gps
-- gpsAltitude
-- creator
-- creatorJobTitle
-- creatorAddress
-- creatorCity
-- creatorStateProvince
-- creatorPostalCode
-- creatorCountry
-- creatorPhone
-- creatorEmail
-- creatorUrl
-- headline
-- iptcSubjectCode
-- descriptionWriter
-- iptcCategory
-- iptcOtherCategories
-- dateCreated
-- intellectualGenre
-- scene
-- location
-- city
-- stateProvince
-- country
-- isoCountryCode
-- jobIdentifier
-- instructions
-- provider
-- source
-- copyright
-- rightsUsageTerms
-- copyrightInfoUrl
-- copyrightStatus
-- keywords
--
-- customMetadata: All plug-in defined custom metadata (defined by any plug-in).
-- (plug-in ID).*: All custom metadata defined by the plug-in with the specified ID.
-- (plug-in ID).(field ID): One specific custom metadata field defined by the plug-in with the specified ID.
function publishServiceProvider.metadataThatTriggersRepublish(publishSettings)
	return {
		default = true
	}
end

-- (optional) This plug-in defined callback function is called when the user
-- creates a new published collection or edits an existing one. It can add
-- additional controls to the dialog box for editing this collection. These controls
-- can be used to configure behaviors specific to this collection (such as
-- privacy or appearance on a web service).
-- <p>This is a blocking call. If you need to start a long-running task (such as
-- network access), create a task using the <a href="LrTasks.html"><code>LrTasks</code></a>
-- namespace.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.viewForCollectionSettings
-- @class function
-- @param f (<a href="LrView.html#LrView.osFactory"><code>LrView.osFactory</code></a> object)
-- A view factory object.
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>collectionSettings</b>: (<a href="LrObservableTable.html"><code>LrObservableTable</code></a>)
-- Plug-in specific settings for this collection. The settings in this table
-- are not interpreted by Lightroom in any way, except that they are stored
-- with the collection. These settings can be accessed via
-- <a href="LrPublishedCollection.html#pubCollection:getCollectionInfoSummary"><code>LrPublishedCollection:getCollectionInfoSummary</code></a>.
-- The values in this table must be numbers, strings, or Booleans.
-- There are some special properties in this table:
-- <p><code>LR_canSaveCollection</code>,
-- which allows you to disable the Edit or Create button in the collection dialog.
-- (If set to true, the Edit / Create button is enabled; if false, it is disabled.)</p>
-- <p><code>LR_liveName</code> will be kept current with the value displayed
-- in the name field of the dialog during the life span of the dialog. This enables
-- a plug-in to add an observer to monitor name changes performed in the dialog.</p>
-- <p><code>LR_canEditName</code> allows the plug-in
-- to control whether the edit field containing the collection name in the dialog is enabled.
-- In the case of new creation, the value defaults to true, meaning that the collection name is
-- editable via the UI, while in the case of a collection being edited, the value defaults in accordance with
-- what the plug-in specifies (or doesn't specify) via 'publishServiceProvider.disableRenamePublishedCollection'.</p></li>
-- <li><b>collectionType</b>: (string) Either "collection" or "smartCollection"
-- (see also: <code>viewForCollectionSetSettings</code>)</li>
-- <li><b>isDefaultCollection</b>: (Boolean) True if this is the default collection.</li>
-- <li><b>name</b>: (string) In the case of editing, the name of this collection (at the time when the edit operation was initiated),
-- otherwise nil.</li>
-- <li><b>parents</b>: (table) An array of information about parents of this collection, in which each element contains:
-- <ul>
-- <li><b>localCollectionId</b>: (number) The local collection ID.</li>
-- <li><b>name</b>: (string) Name of the collection set.</li>
-- <li><b>remoteCollectionId</b>: (number or string) The remote collection ID assigned by the server.</li>
-- </ul>
-- This field is only present when editing an existing published collection.
-- </li>
-- <li><b>pluginContext</b>: (<a href="LrObservableTable.html"><code>LrObservableTable</code></a>)
-- This is a place for your plug-in to store transient state while the collection
-- settings dialog is running. It is passed to your plug-in's
-- <code>endDialogForCollectionSettings</code> callback, and then discarded.</li>
-- <li><b>publishedCollection</b>: (<a href="LrPublishedCollection.html"><code>LrPublishedCollection</code></a>)
-- The published collection object being edited, or nil when creating a new
-- collection.</li>
-- <li><b>publishService</b>: (<a href="LrPublishService.html"><code>LrPublishService</code></a>)
-- 	The publish service object to which this collection belongs.</li>
-- </ul>
-- @return (table) A single view description created from one of the methods in
-- the view factory. (We recommend that <code>f:groupBox</code> be the outermost view.)
function publishServiceProvider.viewForCollectionSettings(f, publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	local bind = import 'LrView'.bind
	local collectionSettings = assert(info.collectionSettings, 'Gallery settings failed to load')
	local topLevel = "unknown"
	local defaultVis = "private"
	local authToken = PhotoShelterHelper.validateLogon(publishSettings)

	local hasLibrisRole = PhotoShelterHelper.getUserLibrisRole(authToken)

	if info.parents then
		if #info.parents > 0 then
			topLevel = "no"
		else
			topLevel = "yes"
		end
	end

	if PhotoShelterInfo.isLibris then
		defaultVis = "inherit"
	end

	collectionSettings.Description = collectionSettings.Description or ""
	collectionSettings.ListedOnWebsite = collectionSettings.ListedOnWebsite or false
	-- This is the searchable option, but we're keeping the old PACT name for compatibility
	collectionSettings.PublicVisible = collectionSettings.PublicVisible or false
	collectionSettings.Visibility = collectionSettings.Visibility or defaultVis
	collectionSettings.PrevPublicVisible = collectionSettings.PublicVisible
	collectionSettings.topLevel = topLevel

	if topLevel == "yes" and collectionSettings.Visibility == "inherit" then
		collectionSettings.Visibility = "private"
	end

	return f:group_box {
		fill_horizontal = 1,
		bind_to_object = collectionSettings,
		f:row {
			spacing = f:label_spacing(),
			fill_horizontal = 1,
			f:static_text {
				title = LOC("$$$/PhotoShelter/CollectionDialog/Description=Description:"),
				alignment = 'right',
				width = share('title_width'),
			},
			f:edit_field {
				width_in_chars = 45,
				height_in_lines = 5,
				value = bind('Description'),
			},
		},
		f:row {
			fill_horizontal = 1,
			spacing = f:label_spacing(),
			f:spacer {
				width = share('title_width'),
			},
			LrView.conditionalItem(topLevel ~= "no" and not hasLibrisRole, f:checkbox {
				title = LOC("$$$/PhotoShelter/CollectionDialog/Listed=Listed on Website"),
				value = bind('ListedOnWebsite'),
			}),
			LrView.conditionalItem(not hasLibrisRole, f:checkbox {
				title = LOC("$$$/PhotoShelter/CollectionDialog/Searchable=Publicly Searchable"),
				value = bind('PublicVisible'),
			}),
		},
		f:spacer {
			height = 10
		},
		f:row {
			fill_horizontal = 1,
			spacing = f:label_spacing(),
			f:static_text {
				title = LOC("$$$/PhotoShelter/ExportDialog/Create/VisibleTo=Visibility:"),
				width = share("title_width"),
				alignment = "right"
			},
			f:column {
				fill_horizontal = 1,
				spacing = f:control_spacing(),
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Inherit=Inherit from Parent"),
					value = bind("Visibility"),
					checked_value = "inherit",
					enabled = LrBinding.keyIsNot("topLevel", "yes"),
				},
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Everyone=Everyone"),
					value = bind("Visibility"),
					checked_value = "everyone"
				},
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Permission=Those with Permission"),
					value = bind("Visibility"),
					checked_value = "permission"
				},
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Me=Only Me"),
					value = bind("Visibility"),
					checked_value = "private"
				}
			}
		}
	}
end

-- (optional) This plug-in defined callback function is called when the user
-- creates a new published collection set or edits an existing one. It can add
-- additional controls to the dialog box for editing this collection set. These controls
-- can be used to configure behaviors specific to this collection set (such as
-- privacy or appearance on a web service).
-- <p>This is a blocking call. If you need to start a long-running task (such as
-- network access), create a task using the <a href="LrTasks.html"><code>LrTasks</code></a>
-- namespace.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.viewForCollectionSetSettings
-- @class function
-- @param f (<a href="LrView.html#LrView.osFactory"><code>LrView.osFactory</code></a> object)
-- A view factory object.
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>collectionSettings</b>: (<a href="LrObservableTable.html"><code>LrObservableTable</code></a>)
-- plug-in specific settings for this collection set. The settings in this table
-- are not interpreted by Lightroom in any way, except that they are stored
-- with the collection set. These settings can be accessed via
-- <a href="LrPublishedCollectionSet.html#pubCollectionSet:getCollectionSetInfoSummary"><code>LrPublishedCollection:getCollectionSetInfoSummary</code></a>.
-- The values in this table must be numbers, strings, or Booleans.
-- There are some special properties in this table:
-- <p><code>LR_canSaveCollection</code>,
-- which allows you to disable the Edit or Create button in the collection set dialog.
-- (If set to true, the Edit / Create button is enabled; if false, it is disabled.)</p>
-- <p><code>LR_liveName</code> will be kept current with the value displayed
-- in the name field of the dialog during the life span of the dialog. This enables
-- a plug-in to add an observer to monitor name changes performed in the dialog.</p>
-- <p><code>LR_canEditName</code> allows the plug-in
-- to control whether the edit field containing the collection set name in the dialog is enabled.
-- In the case of new creation, the value defaults to true, meaning that the collection set name is
-- editable via the UI, while in the case of a collection being edited, the value defaults in accordance with
-- what the plug-in specifies (or doesn't specify) via 'publishServiceProvider.disableRenamePublishedCollection'.</p></li>
-- <li><b>collectionType</b>: (string) "collectionSet"</li>
-- <li><b>isDefaultCollection</b>: (Boolean) true if this is the default collection (will always be false)</li>
-- <li><b>name</b>: (string) In the case of edit, the name of this collection set (at the time when the edit operation was initiated),
-- otherwise nil.</li>
-- <li><b>parents</b>: (table) An array of information about parents of this collection, in which each element contains:
-- <ul>
-- <li><b>localCollectionId</b>: (number) The local collection ID.</li>
-- <li><b>name</b>: (string) Name of the collection set.</li>
-- <li><b>remoteCollectionId</b>: (number or string) The remote collection ID assigned by the server.</li>
-- </ul>
-- This field is only present when editing an existing published collection set. </li>
-- <li><b>pluginContext</b>: (<a href="LrObservableTable.html"><code>LrObservableTable</code></a>)
-- This is a place for your plug-in to store transient state while the collection set
-- settings dialog is running. It will be passed to your plug-in during the
-- <code>endDialogForCollectionSettings</code> and then discarded.</li>
-- <li><b>publishedCollection</b>: (<a href="LrPublishedCollectionSet.html"><code>LrPublishedCollectionSet</code></a>)
-- The published collection set object being edited. Will be nil when creating a new
-- collection Set.</li>
-- <li><b>publishService</b>: (<a href="LrPublishService.html"><code>LrPublishService</code></a>)
-- 	The publish service object.</li>
-- </ul>
-- @return (table) A single view description created from one of the methods in
-- the view factory. (We recommend that <code>f:groupBox</code> be the outermost view.)
function publishServiceProvider.viewForCollectionSetSettings(f, publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)
	local bind = import 'LrView'.bind
	local authToken = PhotoShelterHelper.validateLogon(publishSettings)

	local hasLibrisRole = PhotoShelterHelper.getUserLibrisRole(authToken)
	local collectionSettings = assert(info.collectionSettings, "Collection settings failed to load")
	local topLevel = "unknown"
	local defaultVis = "private"

	if info.parents then
		if #info.parents > 0 then
			topLevel = "no"
		else
			topLevel = "yes"
		end
	end

	if PhotoShelterInfo.isLibris then
		defaultVis = "inherit"
	end

	collectionSettings.Description = collectionSettings.Description or ""
	collectionSettings.ListedOnWebsite = collectionSettings.ListedOnWebsite or false
	collectionSettings.Visibility = collectionSettings.Visibility or defaultVis
	collectionSettings.topLevel = topLevel

	if topLevel == "yes" and collectionSettings.Visibility == "inherit" then
		collectionSettings.Visibility = "private"
	end

	return f:group_box {
		fill_horizontal = 1,
		bind_to_object = collectionSettings,
		f:row {
			spacing = f:label_spacing(),
			f:static_text {
				title = LOC("$$$/PhotoShelter/CollectionDialog/Description=Description:"),
				alignment = 'right',
				width = share('title_width'),
			},
			f:edit_field {
				width_in_chars = 45,
				height_in_lines = 5,
				value = bind('Description'),
			},
		},
		f:row {
			fill_horizontal = 1,
			spacing = f:label_spacing(),
			f:spacer {
				width = share('title_width'),
			},
			LrView.conditionalItem(topLevel ~= "no" and not hasLibrisRole, f:checkbox {
				title = LOC("$$$/PhotoShelter/CollectionDialog/Listed=Listed on Website"),
				value = bind('ListedOnWebsite'),
			}),
		},
		f:spacer {
			height = 10
		},
		f:row {
			fill_horizontal = 1,
			spacing = f:label_spacing(),
			f:static_text {
				title = LOC("$$$/PhotoShelter/ExportDialog/Create/VisibleTo=Visibility:"),
				width = share("title_width"),
				alignment = "right"
			},
			f:column {
				fill_horizontal = 1,
				spacing = f:control_spacing(),
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Inherit=Inherit from Parent"),
					value = bind("Visibility"),
					checked_value = "inherit",
					enabled = LrBinding.keyIsNot("topLevel", "yes"),
				},
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Everyone=Everyone"),
					value = bind("Visibility"),
					checked_value = "everyone"
				},
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Permission=Those with Permission"),
					value = bind("Visibility"),
					checked_value = "permission"
				},
				f:radio_button {
					title = LOC("$$$/PhotoShelter/ExportDialog/Create/Visible/Me=Only Me"),
					value = bind("Visibility"),
					checked_value = "private"
				}
			}
		}
	}
end

-- (optional) This plug-in defined callback function is called when the user
-- has changed the per-collection settings defined via the <code>viewForCollectionSettings</code>
-- callback. It is your opportunity to update settings on your web service to
-- match the new settings.
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>Your code should <b>not</b> use this callback function to clean up from the
-- dialog. This callback is not be called if the user cancels the dialog.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.updateCollectionSettings
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>collectionSettings</b>: (<a href="LrObservableTable.html"><code>LrObservableTable</code></a>)
-- Plug-in specific settings for this collection. The settings in this table
-- are not interpreted by Lightroom in any way, except that they are stored
-- with the collection. These settings can be accessed via
-- <a href="LrPublishedCollection.html#pubCollection:getCollectionInfoSummary"><code>LrPublishedCollection:getCollectionInfoSummary</code></a>.
-- The values in this table must be numbers, strings, or Booleans.
-- <li><b>isDefaultCollection</b>: (Boolean) True if this is the default collection.</li>
-- <li><b>name</b>: (string) The name of this collection.</li>
-- <li><b>parents</b>: (table) An array of information about parents of this collection, in which each element contains:
-- <ul>
-- <li><b>localCollectionId</b>: (number) The local collection ID.</li>
-- <li><b>name</b>: (string) Name of the collection set.</li>
-- <li><b>remoteCollectionId</b>: (number or string) The remote collection ID assigned by the server.</li>
-- </ul> </li>
-- <li><b>publishedCollection</b>: (<a href="LrPublishedCollection.html"><code>LrPublishedCollection</code></a>
-- or <a href="LrPublishedCollectionSet.html"><code>LrPublishedCollectionSet</code></a>)
-- 	The published collection object being edited.</li>
-- <li><b>publishService</b>: (<a href="LrPublishService.html"><code>LrPublishService</code></a>)
-- 	The publish service object to which this collection belongs.</li>
-- </ul>
function publishServiceProvider.updateCollectionSettings(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	Trapper.TRAP(function()
		local authToken = PhotoShelterHelper.validateLogon(publishSettings)

		local exists = PhotoShelterAPI.galleryOrCollectionExists(info.publishedCollection, authToken)

		if not exists then
			LrErrors.throwUserError('Gallery ' .. info.publishedCollection:getName() .. ' was removed on PhotoShelter, please use Sync From PhotoShelter from Plugin Extras menu.')
		end

		local defaultVis = "private"

		if PhotoShelterInfo.isLibris then
			defaultVis = "inherit"
		end

		local collectionSettings = info.collectionSettings or {}
		collectionSettings.Description = collectionSettings.Description or ""
		collectionSettings.ListedOnWebsite = collectionSettings.ListedOnWebsite or false
		collectionSettings.Visibility = collectionSettings.Visibility or defaultVis

		local parentCollectionID = nil
		local parentInfo = info.parents[#info.parents]

		if parentInfo then
			parentCollectionID = parentInfo.remoteCollectionId
		else
			parentCollectionID = "root"
		end

		if #info.parents == 0 and collectionSettings.Visibility == "inherit" then
			collectionSettings.Visibility = "private"
		end

		if info.publishedCollection:getRemoteId() then
			local visibilityWasChanged = collectionSettings.PrevPublicVisible ~= collectionSettings.PublicVisible
			local pubPhotos = info.publishedCollection:getPublishedPhotos()
			local action = 'proceed'

			-- ask user about the action to be taken over the already published photos instead to just update it.
			if visibilityWasChanged and #pubPhotos > 0 then
				action = LrDialogs.promptForActionWithDoNotShow {
					message = LOC("$$$/PhotoShelter/Dialog/Message/SearchableUpdate=Searchable updated"),
					info = LOC("$$$/PhotoShelter/Dialog/Message/SearchableUpdateMessage=You change Searchable option onto Gallery, we will update all Published Photos(^1) to match with this new criteria, do yo wanna proceed with this update?",
										 #pubPhotos),
					actionPrefKey = "proceed",
					verbBtns = {
						{ label = LOC("$$$/PhotoShelter/Dialog/Buttons/Cancel=Cancel"), verb = "cancel", },
						{ label = LOC("$$$/PhotoShelter/Dialog/Buttons/Proceed=Just Gallery"), verb = "proceed-without-update", },
						{ label = LOC("$$$/PhotoShelter/Dialog/Buttons/Proceed=Proceed"), verb = "proceed", },
					}
				}

				if action ~= "cancel" then
					if action == nil then
						action = 'proceed'
					end
				else
					LrErrors.throwCanceled()
				end
			end

			DBG('Updating gallery ' .. info.name .. ' on PhotoShelter', nil, 'DEBUG')
			-- Update existing gallery
			local result, message = PhotoShelterAPI.updateGallery(
					info.publishedCollection:getRemoteId(),
					info.name,
					collectionSettings.Description,
					collectionSettings.ListedOnWebsite,
					collectionSettings.Visibility,
					parentCollectionID,
					authToken
			)

			if result == false then
				LrErrors.throwUserError(message)
			end

			-- Check to see if the user's changed the searchable option
			if visibilityWasChanged and #pubPhotos > 0 and action == 'proceed' then
				for _, pubPhoto in ipairs(pubPhotos) do
					result, message = PhotoShelterAPI.updateImage(
							pubPhoto:getRemoteId(), collectionSettings.PublicVisible, authToken
					)

					if result == false then
						LrErrors.throwUserError(message)
					end
				end
			end

			LrDialogs.showBezel('Gallery updated on PhotoShelter')
		else
			DBG('Creating gallery ' .. info.name .. ' on PhotoShelter', nil, 'DEBUG')
			-- Create new gallery
			local result, idOrMsg = PhotoShelterAPI.createGallery(
					info.name,
					collectionSettings.Description,
					collectionSettings.ListedOnWebsite,
					collectionSettings.Visibility,
					parentCollectionID,
					authToken
			)

			if result then
				if parentCollectionID == nil then
					parentCollectionID = collectionSettings.ListedOnWebsite == 't' and 'l' or 'u'
				end

				info.publishedCollection.catalog:withWriteAccessDo(
						'Store new Collection ID',
						function(context)
							info.publishedCollection:setRemoteId(idOrMsg)
							info.publishedCollection:setRemoteUrl('http://www.photoshelter.com/mem/images/index#/' .. parentCollectionID .. '/' .. idOrMsg)
							LrDialogs.showBezel('Gallery created on PhotoShelter')
						end,
						{ timeout = 5 }
				)
			else
				-- Error
				LrErrors.throwUserError(idOrMsg)
			end
		end
	end)
end


-- (optional) This plug-in defined callback function is called when the user
-- has changed the per-collection set settings defined via the <code>viewForCollectionSetSettings</code>
-- callback. It is your opportunity to update settings on your web service to
-- match the new settings.
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>Your code should <b>not</b> use this callback function to clean up from the
-- dialog. This callback will not be called if the user cancels the dialog.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.updateCollectionSetSettings
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- 	by the user in the Publish Manager dialog. Any changes that you make in
-- 	this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>collectionSettings</b>: (<a href="LrObservableTable.html"><code>LrObservableTable</code></a>)
-- 	Plug-in specific settings for this collection set. The settings in this table
-- 	are not interpreted by Lightroom in any way, except that they are stored
--  with the collection set. These settings can be accessed via
--  <a href="LrPublishedCollectionSet.html#pubCollectionSet:getCollectionSetInfoSummary"><code>LrPublishedCollectionSet:getCollectionSetInfoSummary</code></a>.
--  The values in this table must be numbers, strings, or Booleans.
-- <li><b>isDefaultCollection</b>: (Boolean) True if this is the default collection (always false in this case).</li>
-- <li><b>name</b>: (string) The name of this collection set.</li>
-- <li><b>parents</b>: (table) An array of information about parents of this collection, in which each element contains:
-- <ul>
-- <li><b>localCollectionId</b>: (number) The local collection ID.</li>
-- <li><b>name</b>: (string) Name of the collection set.</li>
-- <li><b>remoteCollectionId</b>: (number or string) The remote collection ID assigned by the server.</li>
-- </ul>
-- </li>
-- <li><b>publishedCollection</b>: (<a href="LrPublishedCollectionSet.html"><code>LrPublishedCollectionSet</code></a>)
-- 	The published collection set object being edited.</li>
-- <li><b>publishService</b>: (<a href="LrPublishService.html"><code>LrPublishService</code></a>)
-- 	The publish service object.</li>
-- </ul>
function publishServiceProvider.updateCollectionSetSettings(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	Trapper.TRAP(function()
		-- no need to verify if we get auth token because if logon is not valid it will raise an error.
		local authToken = PhotoShelterHelper.validateLogon(publishSettings)
		local defaultVis = "private"
		local collectionSettings = assert(info.collectionSettings, 'collection Settings not present')
		local parentCollectionID = nil
		local parentInfo = info.parents[#info.parents]

		local exists = PhotoShelterAPI.galleryOrCollectionExists(info.publishedCollection, authToken)

		if not exists then
			LrErrors.throwUserError('Collection ' .. info.publishedCollection:getName() .. ' was removed on PhotoShelter, please use Sync From PhotoShelter from Plugin Extras menu.')
		end

		if PhotoShelterInfo.isLibris then
			defaultVis = "inherit"
		end

		collectionSettings.Description = collectionSettings.Description or ""
		collectionSettings.ListedOnWebsite = collectionSettings.ListedOnWebsite or false
		collectionSettings.Visibility = collectionSettings.Visibility or defaultVis

		if parentInfo then
			parentCollectionID = parentInfo.remoteCollectionId
		end

		if #info.parents == 0 and collectionSettings.Visibility == "inherit" then
			collectionSettings.Visibility = "private"
		end

		if info.publishedCollection:getRemoteId() then
			DBG('Updating collection ' .. info.name .. ' on PhotoShelter', nil, 'DEBUG')
			-- Update existing collection
			local result, message = PhotoShelterAPI.updateCollection(
					info.publishedCollection:getRemoteId(),
					info.name,
					collectionSettings.Description,
					collectionSettings.ListedOnWebsite,
					collectionSettings.Visibility,
					parentCollectionID,
					authToken
			)

			if result == false then
				LrErrors.throwUserError(message)
			else
				LrDialogs.showBezel("Collection updated in PhotoShelter")
			end
		else
			-- Create new collection
			DBG('Creating new collection ' .. info.name .. ' on PhotoShelter', nil, 'DEBUG')
			local result, idOrMsg = PhotoShelterAPI.createCollection(
					info.name,
					collectionSettings.Description,
					collectionSettings.ListedOnWebsite,
					collectionSettings.Visibility,
					parentCollectionID,
					authToken
			)

			if result then
				info.publishedCollection.catalog:withWriteAccessDo(
						'Store new Collection ID',
						function(context)
							-- Do something that reads or writes catalog
							info.publishedCollection:setRemoteId(idOrMsg)
							info.publishedCollection:setRemoteUrl("http://www.photoshelter.com/mem/images/index#/" .. idOrMsg)
							LrDialogs.showBezel("Collection created in PhotoShelter")
						end,
						{ timeout = 5 }
				)
			else
				-- Error
				LrErrors.throwUserError(idOrMsg)
			end
		end
	end)
end

-- This plug-in callback function is called when the user has re parented a
-- published collection via the Publish Services panel user interface. This is
-- your plug-in's opportunity to make the corresponding change on the service.
-- <p>If your plug-in is unable to update the remote service for any reason,
-- you should throw a Lua error from this function; this causes Lightroom to revert the change.</p>
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.reparentPublishedCollection
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>isDefaultCollection</b>: (Boolean) True if this is the default collection.</li>
-- <li><b>name</b>: (string) The new name being assigned to this collection.</li>
-- <li><b>parents</b>: (table) An array of information about parents of this collection, in which each element contains:
-- <ul>
-- <li><b>localCollectionId</b>: (number) The local collection ID.</li>
-- <li><b>name</b>: (string) Name of the collection set.</li>
-- <li><b>remoteCollectionId</b>: (number or string) The remote collection ID assigned by the server.</li>
-- </ul> </li>
-- <li><b>publishService</b>: (<a href="LrPublishService.html"><code>LrPublishService</code></a>)
-- 	The publish service object.</li>
-- <li><b>publishedCollection</b>: (<a href="LrPublishedCollection.html"><code>LrPublishedCollection</code></a>
-- or <a href="LrPublishedCollectionSet.html"><code>LrPublishedCollectionSet</code></a>)
-- 	The published collection object being renamed.</li>
-- <li><b>remoteId</b>: (string or number) The ID for this published collection
-- 	that was stored via <a href="LrExportSession.html#exportSession:recordRemoteCollectionId"><code>exportSession:recordRemoteCollectionId</code></a></li>
-- <li><b>remoteUrl</b>: (optional, string) The URL, if any, that was recorded for the published collection via
-- <a href="LrExportSession.html#exportSession:recordRemoteCollectionUrl"><code>exportSession:recordRemoteCollectionUrl</code></a>.</li>
-- </ul>
function publishServiceProvider.reparentPublishedCollection(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	if not info.publishedCollection:getRemoteId() then
		-- This hasn't yet been published, allow any movement
		return
	end

	local authToken = PhotoShelterHelper.validateLogon(publishSettings)
	local exists = PhotoShelterAPI.galleryOrCollectionExists(info.publishedCollection, authToken)

	if not exists then
		local elType = PhotoShelterHelper.isCollection(info.publishedCollection) and 'Collection' or 'Gallery'
		LrErrors.throwUserError(elType .. ' ' .. info.publishedCollection:getName() .. ' was removed on PhotoShelter, please use Sync From PhotoShelter from Plugin Extras menu.')
	end

	if info.parents[#info.parents] and info.parents[#info.parents].remoteCollectionId then
		local newParent = info.parents[#info.parents]
		local exists, data = PhotoShelterAPI.queryCollectionById(newParent.remoteCollectionId, authToken)

		if not exists then
			-- new parent will always be a collection set (collection in ps)
			LrErrors.throwUserError('Collection ' .. newParent.name .. ' was removed on PhotoShelter, please use Sync From PhotoShelter from Plugin Extras menu.')
		end
	end

	DBG('Reparent published collection', nil, 'DEBUG')

	if #info.parents > 1 then
		for _, parent in pairs(info.parents) do
			if not parent.remoteCollectionId then
				-- If some of the parents in the tree doesn't have an ID then it hasn't yet been published.
				--
				-- This is hard to handle cleanly until the structure is sync'd, but we don't want to force a sync during an operation like this
				-- in case all hell breaks loose (changing the set structure whilst moving sets isn't wise)
				--
				-- We'll tell the user to sync first.
				LrDialogs.message(
						LOC("$$$/PhotoShelter/Reparent/SyncFirst=Synchronisation required"),
						LOC("$$$/PhotoShelter/Reparent/SyncFirst/Detail=You need to sync the collection structure before reparent it. Out of sync collection: ^1",
								parent.name)
				)

				LrErrors.throwCanceled()
			end
		end
	end

	Trapper.TRAP(function()
		local authToken = PhotoShelterHelper.validateLogon(publishSettings)
		local newParentId
		local oldParentId
		local parentinfo = info.parents[#info.parents]
		local parent = info.publishedCollection:getParent()

		if parentinfo then
			newParentId = parentinfo.remoteCollectionId
		end

		if parent then
			oldParentId = parent:getRemoteId()
		end

		-- LrPublishedCollection is when it is a Gallery in PhotoShelter
		if info.publishedCollection:type() == 'LrPublishedCollection' then
			if newParentId then
				local result, msg = PhotoShelterAPI.addGalleryToCollection(info.remoteId, newParentId, authToken)

				if result == false then
					LrErrors.throwUserError(msg)
				end
			else
				-- new parent is root so update so set f_list
				local collectionSettings = info.publishedCollection:getCollectionInfoSummary().collectionSettings
				local result, message = PhotoShelterAPI.updateGallery(
						info.remoteId,
						info.name,
						collectionSettings.Description,
						collectionSettings.ListedOnWebsite,
						collectionSettings.Visibility,
						"root", -- force update method to set f_list value based on the gallery setting.
						authToken
				)

				if result == false then
					LrErrors.throwUserError(msg)
				end
			end

			local rollbackAdd = function()
				-- rollback the new added gallery otherwise this collection will be in two place at same time.
				-- todo: didn't exists a move collection endpoint?
				local rollbackResult, rollbackMsg = PhotoShelterAPI.removeGalleryFromCollection(
						info.remoteId, newParentId, authToken
				)

				if (rollbackResult == false) then
					LrErrors.throwUserError(msg .. " also we fail on rollback the insertion on the new target parent with follow error: " .. rollbackMsg)
				else
					LrErrors.throwUserError("Unable to move Gallery to the target collection due an error in PhotoShelter")
				end
			end

			if oldParentId then
				local result, msg = PhotoShelterAPI.removeGalleryFromCollection(info.remoteId, oldParentId, authToken)

				if result == false then
					rollbackAdd()
				end
			else
				-- update visibility
				local collectionSettings = info.publishedCollection:getCollectionInfoSummary().collectionSettings
				local result, message = PhotoShelterAPI.updateGallery(
						info.remoteId,
						info.name,
						collectionSettings.Description,
						collectionSettings.ListedOnWebsite,
						collectionSettings.Visibility,
						nil, -- it prevent set f_list, so basically it disapear from "root" level.
						authToken
				)

				if result == false then
					rollbackAdd()
				end
			end

			-- If we've moved the item to the root, then "inherit" isn't a viable option for the visibility.
			-- Get the actual one from the server
			if newParentId == nil then
				local success, permissions = PhotoShelterAPI.getGalleryPermissions(info.remoteId, authToken)

				if success then
					local c = info.publishedCollection
					local collectionSettings = c:getCollectionInfoSummary().collectionSettings or {}
					collectionSettings.Visibility = permissions

					c.catalog:withWriteAccessDo(
							LOC("$$$/PhotoShelter/Sync/Undo=Synchronize PhotoShelter Services"),
							function()
								c:setCollectionSettings(collectionSettings)
								LrDialogs.showBezel('PhotoShelter Gallery successfully moved.')
							end,
							{ timeout = 5 }
					)
				else
					LrDialogs.showBezel('PhotoShelter Gallery successfully moved.')
				end
			end
		else
			local c = info.publishedCollection
			local collectionSettings = c:getCollectionSetInfoSummary().collectionSettings or {}
			local result, msg = PhotoShelterAPI.moveCollectionToCollection(
					info.remoteId, newParentId, collectionSettings.ListedOnWebsite, authToken
			)

			if result == false then
				LrErrors.throwUserError(msg)
			end

			-- If we've moved the item to the root, then "inherit" isn't a viable option for the visibility.
			-- Get the actual one from the server
			if newParentId == nil then
				local success, permissions = PhotoShelterAPI.getCollectionPermissions(info.remoteId, authToken)

				if success then
					local cs = info.publishedCollection
					local collectionSettings = cs:getCollectionSetInfoSummary().collectionSettings or {}
					collectionSettings.Visibility = permissions

					cs.catalog:withWriteAccessDo(
							LOC("$$$/PhotoShelter/Sync/Undo=Synchronize PhotoShelter Services"),
							function()
								cs:setCollectionSetSettings(collectionSettings)
								LrDialogs.showBezel('PhotoShelter Collection successfully moved.')
							end,
							{ timeout = 5 }
					)
				end
			end

			LrDialogs.showBezel('PhotoShelter Collection successfully moved.')
		end
	end)
end

-- (optional) This plug-in defined callback function is called when the user
-- has attempted to delete one or more published collections defined by your
-- plug-in from Lightroom. It provides an opportunity for you to customize the
-- confirmation dialog.
-- <p>Do not use this hook to actually tear down the collection(s). Instead, use
-- <a href="#publishServiceProvider.deletePublishedCollection"><code>deletePublishedCollection</code></a>
-- for that purpose.
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.shouldDeletePublishedCollection
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>collections</b>: (array of <a href="LrPublishedCollection.html"><code>LrPublishedCollection</code></a>
--  or <a href="LrPublishedCollectionSet.html"><code>LrPublishedCollectionSet</code></a>)
-- 	The published collection objects.</li>
-- <li><b>nPhotos</b>: (number) The number of photos contained in the
-- 	published collection. Only present if there is a single published collection
--  to be deleted.</li>
-- <li><b>nChildren</b>: (number) The number of child collections contained within the
-- 	published collection set. Only present if there is a single published collection set
--  to be deleted.</li>
-- <li><b>hasItemsOnService</b>: (boolean) True if one or more photos have been
--  published through the collection(s) to be deleted.</li>
-- </ul>
-- @return (string) "ignore", "cancel", "delete", or nil
-- (If you return nil, Lightroom's default dialog will be displayed.)
function publishServiceProvider.shouldDeletePublishedCollection(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	DBG('publishServiceProvider.shouldDeletePublishedCollection', nil, 'DEBUG')

	local retval = nil
	local response = LrDialogs.confirm(
			LOC("$$$/PhotoShelter/DeletePublishedCollection/Message=Delete a Published PhotoShelter Gallery"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/Info=Deleting an gallery on PhotoShelter also deletes the contained images\nThis can't be undone!"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionOk=Delete"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionCancel=Cancel"),
			LOC("$$$/PhotoShelter/DeletePublishedCollection/ActionOther=Leave Gallery on PhotoShelter")
	)

	if response == 'ok' then
		retval = 'delete'
	end

	if response == 'other' then
		retval = 'ignore'
	end

	if response == 'cancel' then
		retval = 'cancel'
	end

	return retval
end

-- This plug-in callback function is called when the user has deleted a
-- published collection via the Publish Services panel user interface. This is
-- your plug-in's opportunity to make the corresponding change on the service.
-- <p>If your plug-in is unable to update the remote service for any reason,
-- you should throw a Lua error from this function; this causes Lightroom to revert the change.</p>
-- <p>This is not a blocking call. It is called from within a task created
-- using the <a href="LrTasks.html"><code>LrTasks</code></a> namespace. In most
-- cases, you should not need to start your own task within this function.</p>
-- <p>First supported in version 3.0 of the Lightroom SDK.</p>
-- @name publishServiceProvider.deletePublishedCollection
-- @class function
-- @param publishSettings (table) The settings for this publish service, as specified
-- by the user in the Publish Manager dialog. Any changes that you make in
-- this table do not persist beyond the scope of this function call.
-- @param info (table) A table with these fields:
-- <ul>
-- <li><b>isDefaultCollection</b>: (Boolean) True if this is the default collection.</li>
-- <li><b>name</b>: (string) The new name being assigned to this collection.</li>
-- <li><b>parents</b>: (table) An array of information about parents of this collection, in which each element contains:
-- <ul>
-- <li><b>localCollectionId</b>: (number) The local collection ID.</li>
-- <li><b>name</b>: (string) Name of the collection set.</li>
-- <li><b>remoteCollectionId</b>: (number or string) The remote collection ID assigned by the server.</li>
-- </ul> </li>
-- <li><b>publishService</b>: (<a href="LrPublishService.html"><code>LrPublishService</code></a>)
-- 	The publish service object.</li>
-- <li><b>publishedCollection</b>: (<a href="LrPublishedCollection.html"><code>LrPublishedCollection</code></a>
-- or <a href="LrPublishedCollectionSet.html"><code>LrPublishedCollectionSet</code></a>)
-- 	The published collection object being renamed.</li>
-- <li><b>remoteId</b>: (string or number) The ID for this published collection
-- 	that was stored via <a href="LrExportSession.html#exportSession:recordRemoteCollectionId"><code>exportSession:recordRemoteCollectionId</code></a></li>
-- <li><b>remoteUrl</b>: (optional, string) The URL, if any, that was recorded for the published collection via
-- <a href="LrExportSession.html#exportSession:recordRemoteCollectionUrl"><code>exportSession:recordRemoteCollectionUrl</code></a>.</li>
-- </ul>
function publishServiceProvider.deletePublishedCollection(publishSettings, info)
	PhotoShelterHelper.setUpDebug(publishSettings, false)

	DBG('publishServiceProvider.deletePublishedCollection', nil, 'DEBUG')

	Trapper.TRAP(function()
		local authToken = PhotoShelterHelper.validateLogon(publishSettings)

		if not info or not info.remoteId then
			return
		end

		local result = false
		local msg
		local elType = info.publishedCollection:type()
		local elTypeName

		if elType == 'LrPublishedCollectionSet' then
			elTypeName = 'Collection'
			result, msg = PhotoShelterAPI.deleteCollection(info.remoteId, authToken)
		elseif elType == 'LrPublishedCollection' then
			elTypeName = 'Gallery'
			result, msg = PhotoShelterAPI.deleteGallery(info.remoteId, authToken)
		else
			LrErrors.throwUserError('Deleted Element ' .. info.publishedCollection:type() .. ' is not supported yet, please report this.')
		end

		if result == false then
			LrErrors.throwUserError(msg)
		end

		LrDialogs.showBezel(elTypeName .. ' deleted from PhotoShelter')
	end)
end

local showChooseFileId = function(context, files, filename)
	local f = LrView.osFactory()
	local bind = LrView.bind

	local prop = LrBinding.makePropertyTable(context)
	prop.image_id = nil

	local items = {}

	for i, image in ipairs(files) do
		items[#items + 1] = {
			value = image.image_id,
			title = image.image_id,
			value_equal = function(value1, value2)
				return value1 == value2
			end
		}
	end

	local controls = f:column(
			{
				bind_to_object = prop,
				f:simple_list(
						{
							value = bind 'image_id',
							items = items,
							allows_multiple_selection = false
						}
				)
			}
	)

	local result = LrDialogs.presentModalDialog(
			{
				title = LOC("$$$/PhotoShelter/Upload/ExportOverwriteIDChooser=Choose PhotoShelter Photo"),
				resizable = false,
				contents = f:column {
					f:static_text {
						width = 300,
						title = LOC("$$$/PhotoShelter/Upload/ExportOverwriteIDChooserDescription=We found more than one file with same name ^1 so please choose which one do you want override:",
												filename),
					},
					f:spacer {
						height = 10,
					},
					controls,
				}
			}
	)

	return result == "ok", result == "ok" and prop.image_id[#prop.image_id] or nil
end

function publishServiceProvider.processRenderedPhotosForExport(functionContext, exportContext, authToken)
	local exportSession = exportContext.exportSession
	local pt = exportContext.propertyTable

	-- Set progress title
	local nPhotos = exportSession:countRenditions()

	local progressScope = exportContext:configureProgress {
		title = nPhotos > 1
				and LOC("$$$/PhotoShelter/Upload/Progress=Uploading ^1 photos to PhotoShelter", nPhotos)
				or LOC("$$$/PhotoShelter/Upload/Progress/One=Uploading one photo to PhotoShelter"),
	}

	-- Get the list of existing files
	progressScope:setCaption(LOC("$$$/PhotoShelter/Upload/GetImageList=Retrieving existing image list"))

	--  Fix for Export with Previous not having access to the folderPath table, but the serialized version of that table.
	if pt.folderPath == nil then
		pt.folderPath = {}
		local gal_id = string.match(pt.serialisedFolderPath, 'G0000...........')
		table.insert(pt.folderPath, gal_id)
	end

	DBG('Checking photos for export...', nil, 'DEBUG')
	local existingFiles = {}
	local skipped = {}
	if pt.ifexists == "skip" or pt.ifexists == 'overwrite' or pt.ifexists == 'rename' then
		-- the left option is upload anyway
		for _, rendition in exportContext.exportSession:renditions() do
			repeat
				local renditionFilename = LrPathUtils.leafName(rendition.destinationPath)
				DBG('Checking photo ' .. renditionFilename, nil, 'DEBUG')
				local success, matchesOrError = PhotoShelterAPI.queryImagesByName(
						pt.folderPath[#pt.folderPath],
						renditionFilename,
						authToken
				)

				if not success then
					DBG(renditionFilename .. ' check fail', { error = matchesOrError }, 'ERROR')
					rendition:uploadFailed(matchesOrError)
					break
				end

				if matchesOrError.Total > 0 then
					DBG('Found ' .. matchesOrError.Total .. ' for photo ' .. renditionFilename, nil, 'DEBUG')
					if pt.ifexists == 'skip' then
						skipped[#skipped + 1] = renditionFilename
						DBG('do to skip was selected file ' .. renditionFilename .. ' was skipped', nil, 'DEBUG')
						rendition:skipRender()
						break
					end

					if pt.ifexists == 'rename' then
						-- id didn't mather here, just the amount of repeated to make the new name
						existingFiles[renditionFilename] = { image_id = nil, count = matchesOrError.Total }
						break
					end

					if matchesOrError.Total == 1 then
						-- take the only one that match.
						existingFiles[renditionFilename] = { image_id = matchesOrError.Image[1].image_id, count = matchesOrError.Total }
						break
					end

					-- otherwise ask the user about the multiple images with same name, so see which one he desire override.
					local success, id = showChooseFileId(functionContext, matchesOrError.Image, renditionFilename)

					if not success then
						rendition:uploadFailed("user didn't select a photo to override from the matches in gallery")
						rendition:skipRender()
						skipped[#skipped + 1] = renditionFilename
						DBG('user didn\'t select a photo to override from the matches in gallery', nil, 'DEBUG')
						break
					end
					DBG('Photo ' .. renditionFilename .. ' has multiple files on remote, but user choose ' .. id, nil, 'DEBUG')
					existingFiles[renditionFilename] = { image_id = id, count = matchesOrError.Total }
				else
					DBG('Photo ' .. renditionFilename .. ' was not found on photoshelter', nil, 'DEBUG')
				end
			until true
		end
	end

	-- Iterate through photo renditions.
	progressScope:setCaption(LOC("$$$/PhotoShelter/Upload/UploadingPhotos=Uploading images"))

	for i, rendition in exportContext:renditions { stopIfCanceled = true } do
		local currentStep = 0
		local totalSteps = 2
		repeat
			if rendition.wasSkipped then
				DBG('Rendition was skipped', nil, 'DEBUG')
				break
			end

			-- Check for cancellation again after photo has been rendered.
			if progressScope:isCanceled() then
				DBG('Scope cancelled by user.', nil, 'DEBUG')
				return
			end

			-- Wait for next photo to render.
			DBG('Waiting rendition rendering...', nil, 'DEBUG')
			local success, pathOrMessage = rendition:waitForRender()

			if not success then
				DBG('Rendition rendering fail', { error = pathOrMessage }, 'ERROR')
				rendition:uploadFailed(pathOrMessage)
				-- upgrade progress
				progressScope:setPortionComplete(i + 1, nPhotos) -- step done due an error
				break
			end
			DBG('Rendition rendering success', nil, 'DEBUG')
			currentStep = currentStep + 1
			progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)

			local leafName = LrPathUtils.leafName(pathOrMessage)

			-- get the gallery id from source location of the export
			local galleryNameOrID

			if not pt.contributor then
				if #pt.folderPath > 0 then
					galleryNameOrID = pt.folderPath[#pt.folderPath]
				end
			else
				if #pt.folderPath > 0 then
					galleryNameOrID = pt.root[1].name
				end
			end

			local fileName = leafName
			local targetId = galleryNameOrID
			DBG('Retrieving photo mimeType', nil, 'DEBUG')
			local contentType = PhotoShelterHelper.getPhotoMimeType(rendition.photo)
			DBG('Photo mimeType is: ' .. contentType, nil, 'DEBUG')
			local searchable = pt.searchable
			local isReplace = false

			-- file exists so we need to take the action selected on the export dialog
			if pt.ifexists == "rename" then
				if not existingFiles[leafName] then
					fileName = leafName
				else
					local nameNoExt = LrPathUtils.removeExtension(leafName)
					local ext = LrPathUtils.extension(leafName)
					local n = existingFiles[leafName].count + 1
					fileName = nameNoExt .. "_" .. n .. "." .. ext
					DBG('Photo renamed from: ' .. leafName .. ' to: ' .. fileName, nil, 'DEBUG')
				end
			end

			if pt.ifexists == "overwrite" and existingFiles[leafName] then
				targetId = existingFiles[leafName].image_id
				isReplace = true
				DBG('Photo override id: ' .. existingFiles[leafName].image_id, nil, 'DEBUG')
			end

			local success, idOrError = LrFunctionContext.pcallWithContext(
					isReplace and 'ExportReplacePhoto' or 'ExportUploadPhoto',
					function(_, ...)
						local args = { ... }
						if isReplace then
							return PhotoShelterAPI.replace(unpack(args))
						else
							return PhotoShelterAPI.upload(unpack(args))
						end
					end,
					pt, pathOrMessage, fileName, contentType, targetId, searchable, authToken
			)

			if not success then
				DBG('Export: upload photo ' .. pathOrMessage .. ' fail.', { error = idOrError, file = pathOrMessage }, 'ERROR')
				rendition:uploadFailed(idOrError)
				progressScope:setPortionComplete(i + 1, nPhotos) -- step done due an error
				break
			end

			currentStep = currentStep + 1
			progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)
			DBG('Photo ' .. pathOrMessage .. ' export done', nil, 'DEBUG')
		until true -- a way to have something similar to "continue" to interrupt the loop.
	end

	if #skipped > 0 then
		local message
		if #skipped == 1 then
			message = LOC("$$$/PhotoShelter/Upload/Skipped/OneFileSkipped=1 file skipped because is already in target gallery at PhotoShelter")
		else
			message = LOC("$$$/PhotoShelter/Upload/Skipped/FewFilesSkipped=^1 files skipped because are already in target gallery at PhotoShelter",
										#skipped)
		end

		LrDialogs.message(message, table.concat(skipped, "\n"))
	end

	LrDialogs.showBezel('Export operation completed successfully')
end

function publishServiceProvider.processRenderedPhotosForPublish(publishService, functionContext, exportContext, authToken)
	local exportSession = exportContext.exportSession
	local pt = exportContext.propertyTable
	local nPhotos = exportSession:countRenditions()
	local publishService = exportContext.publishService
	local progressScope = exportContext:configureProgress {
		title = LOC "$$$/PhotoShelter/Task/Progress/Publish=Publishing to PhotoShelter",
	}
	progressScope:attachToFunctionContext(functionContext)
	progressScope:setPortionComplete(0, nPhotos)
	-- Iterate through photo renditions.
	progressScope:setCaption(
			nPhotos > 1
					and LOC("$$$/PhotoShelter/Upload/Progress=Uploading ^1 photos to PhotoShelter", nPhotos)
					or LOC "$$$/PhotoShelter/Upload/Progress/One=Uploading one photo to PhotoShelter"
	)

	local publishedCollection = exportContext.publishedCollection
	local publishedCollectionInfo = publishedCollection:getCollectionInfoSummary()
	local publishedCollectionSettings = publishedCollectionInfo.collectionSettings or {}
	local galleryID = publishedCollection:getRemoteId()
	-- The searchable flag is defined in the gallery settings
	local searchable = publishedCollectionSettings.PublicVisible
	-- Get the user's PhotoShelter url
	local userURL = PhotoShelterAPI.getSiteURL(authToken)

	for i, rendition in exportSession:renditions({ stopIfCanceled = true }) do
		repeat -- a way to could have something like a `continue` statement and prevent so dammit nested if
			local photo = rendition.photo
			local mime = PhotoShelterHelper.getPhotoMimeType(photo)
			local remoteIdFromMeta, error = photo:getPropertyForPlugin(_PLUGIN, 'id', nil, true)

			if rendition.wasSkipped then
				progressScope:setPortionComplete(i + 1, nPhotos)
				break
			end

			-- check if the user still exists at photoshelter of is a remanent that for some reason was not cleanup at delete time.
			local exists = PhotoShelterAPI.queryImageExists(remoteIdFromMeta, authToken)

			if not exists then
				remoteIdFromMeta = nil -- move the path to the new publish image.
				local catalog = exportContext.publishService.catalog
				catalog:withPrivateWriteAccessDo(
						function(context)
							-- set remote id on metadata
							rendition.photo:setPropertyForPlugin(_PLUGIN, 'id', nil)
							rendition.photo:setPropertyForPlugin(_PLUGIN, 'onServer', nil)
						end
				)
			end

			local success, pathOrMessage = rendition:waitForRender()
			if not success then
				-- it allow to continue the publish but give alert to the user at the end of the process about all the errors for all images
				rendition:uploadFailed(pathOrMessage)
				progressScope:setPortionComplete(i + 1, nPhotos)
				break
			end

			-- Check for cancellation again after photo has been rendered.
			if progressScope:isCanceled() then
				return
			end

			local leafName = LrPathUtils.leafName(pathOrMessage)
			local currentStep = 0
			local totalSteps = 1

			if not remoteIdFromMeta and error then
				rendition:uploadFailed(error)
				progressScope:setPortionComplete(i + 1, nPhotos)
				break
			end

			if not remoteIdFromMeta then
				totalSteps = 2
				-- upload file
				-- add image to gallery
				DBG('Uploading Photo ' .. pathOrMessage, nil, 'DEBUG')
				local photoShelterPhotoId
				local success, IdOrError = LrFunctionContext.pcallWithContext(
						'UploadPhoto',
						function(_, ...)
							local args = { ... }
							return PhotoShelterAPI.upload(unpack(args))
						end,
						pt, pathOrMessage, leafName, mime,
						galleryID, searchable, authToken
				)

				if not success then
					rendition:uploadFailed(IdOrError)
					DBG('Upload Photo ' .. pathOrMessage .. ' failed. Rendition mark as uploadFailed',
							{ error = IdOrError, file = pathOrMessage }, 'ERROR')
					progressScope:setPortionComplete(i + 1, nPhotos) -- step done due an error
					break
				end

				photoShelterPhotoId = IdOrError
				DBG('Photo ' .. pathOrMessage .. ' uploaded, retrieved ID: ' .. photoShelterPhotoId, nil, 'DEBUG')
				currentStep = currentStep + 1
				progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)
				DBG('Updating photo metadata in local copy' .. pathOrMessage, nil, 'DEBUG')
				local developSettings = PhotoShelterHelper.getSerialisedDevelopSettings(rendition.photo)
				local catalog = exportContext.publishService.catalog
				catalog:withPrivateWriteAccessDo(
						function(context)
							-- set develop setting in metadata to verify next time
							PhotoShelterHelper.setImageDataForPublishService(
									publishService, rendition.photo, "developSettings", developSettings
							)
							-- set remote id on metadata
							rendition.photo:setPropertyForPlugin(_PLUGIN, 'id', photoShelterPhotoId)
							rendition.photo:setPropertyForPlugin(_PLUGIN, 'onServer', true)
							-- set rendition remote id
							rendition:recordPublishedPhotoId(photoShelterPhotoId)
							-- set rendition remote url
							rendition:recordPublishedPhotoUrl(userURL .. "/image/" .. photoShelterPhotoId)
							currentStep = currentStep + 1
							progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)
						end
				)
			else
				local photoUrl = userURL .. "/image/" .. remoteIdFromMeta
				totalSteps = 3
				-- request galleries of image in photoshelter
				-- check if current gallery is on parent list retrieve from photoshelter, if not there, add image to it. if gallery if not published, publish and add to it.

				-- it have a remote id, but it have not a remoteId under this gallery
				local success, galleriesOrMessage = PhotoShelterAPI.getParentGalleries(remoteIdFromMeta, authToken)
				local galleryFound = false

				if not success then
					rendition:uploadFailed(error)
					progressScope:setPortionComplete(i + 1, nPhotos)
					break
				end

				if #galleriesOrMessage > 0 then
					for _, gallery in pairs(galleriesOrMessage) do
						if gallery.gallery_id == galleryID then
							galleryFound = true
							break
						end
					end
				end

				if not galleryFound then
					local success, error = LrFunctionContext.pcallWithContext(
							'AddImageToGallery',
							function(_, ...)
								local args = { ... }
								return PhotoShelterAPI.addImageToGallery(unpack(args))
							end,
							galleryID, remoteIdFromMeta, authToken
					)

					if not success then
						rendition:uploadFailed(error)
						progressScope:setPortionComplete(i + 1, nPhotos)
						break
					end
				end
				progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)

				-- check where is the change
				-- -- check if change is on develop setting, if this is the case need to replace the image and re set the develop settings metadata on file
				local oldDevelopSettings = PhotoShelterHelper.getImageDataForPublishService(
						publishService, photo, "developSettings"
				)

				local developSettings = PhotoShelterHelper.getSerialisedDevelopSettings(photo)

				if (developSettings ~= oldDevelopSettings) then
					-- the image it self has changes, so it need to be uploaded again.
					-- replace image with new one
					local success, idOrError = LrFunctionContext.pcallWithContext(
							'ReplacePhoto',
							function(_, ...)
								local args = { ... }
								PhotoShelterAPI.replace(unpack(args))
							end,
							pt, pathOrMessage, leafName, mime, remoteIdFromMeta, pt.searchable, authToken
					)

					if not success then
						DBG('Replace Photo ' .. pathOrMessage .. ' failed. ', { error = idOrError, file = pathOrMessage }, 'ERROR')
						rendition:uploadFailed(idOrError)
						progressScope:setPortionComplete(i + 1, nPhotos)
						break
					end

					exportContext.publishService.catalog:withPrivateWriteAccessDo(
							function(context)
								-- set develop setting in metadata to verify next time
								PhotoShelterHelper.setImageDataForPublishService(
										publishService, rendition.photo, "developSettings", developSettings
								)
							end
					)
				end
				progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)

				-- -- check if change is on metadata, if this is the case need to update the metadata info in photoshelter
				local success, psIPTCorMessage = PhotoShelterAPI.getIPTC(remoteIdFromMeta, authToken)

				if not success then
					rendition:uploadFailed(psIPTCorMessage)
					progressScope:setPortionComplete(i + 1, nPhotos)
					break
				end

				local lrData = photo:getFormattedMetadata()
				local changed = false
				local newIPTCdata = {}

				for k, v in pairs(PhotoShelterHelper.IPTCToLR) do
					local currentData = lrData[v]

					if currentData then
						local psIPTCitem = psIPTCorMessage[k]

						if psIPTCitem == nil then
							psIPTCitem = ""
						end

						if currentData ~= psIPTCitem then
							newIPTCdata[k] = currentData
							changed = true
						end
					end
				end

				if not changed then
					-- we need set again the remoteId just because is the callback that allow to LRC to understand that its done
					progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)
					-- set rendition remote id
					rendition:recordPublishedPhotoId(remoteIdFromMeta)
					-- set rendition remote url
					rendition:recordPublishedPhotoUrl(photoUrl)
					break
				end

				DBG("Uploading mete data only", nil, 'DEBUG')
				local success, error = PhotoShelterAPI.setIPTC(remoteIdFromMeta, newIPTCdata, authToken)

				if not success then
					rendition:uploadFailed(error)
					progressScope:setPortionComplete(i + 1, nPhotos)
					break
				end

				-- set rendition remote id
				rendition:recordPublishedPhotoId(remoteIdFromMeta)
				-- set rendition remote url
				rendition:recordPublishedPhotoUrl(photoUrl)
				progressScope:setPortionComplete(i + (currentStep / totalSteps), nPhotos)
			end

			LrFileUtils.delete(pathOrMessage)
		until true
	end

	LrDialogs.showBezel('Publish operation completed successfully')
end

function publishServiceProvider.processRenderedPhotos(functionContext, exportContext)
	local pt = exportContext.propertyTable

	-- define global variables DBG with pt as first param of each log generated by DBG("")
	PhotoShelterHelper.setUpDebug(pt, false)

	Trapper.TRAP(function()
		local authToken = PhotoShelterHelper.validateLogon(pt) -- it will throwUserError in case of invalid auth credentials.

		if pt.LR_isExportForPublish then
			publishServiceProvider.processRenderedPhotosForPublish(
					exportContext.publishService, functionContext, exportContext, authToken
			)
		else
			publishServiceProvider.processRenderedPhotosForExport(
					functionContext, exportContext, authToken
			)
		end
	end)

end

PublishSupport = publishServiceProvider