local LrFunctionContext = import 'LrFunctionContext'
local LrApplication = import 'LrApplication'
local LrProgressScope = import 'LrProgressScope'
local LrBinding = import 'LrBinding'
local LrView = import 'LrView'
local LrDialogs = import 'LrDialogs'
local LrErrors = import 'LrErrors'
local LrTasks = import 'LrTasks'
local LrStringUtils = import 'LrStringUtils'
local LrDate = import 'LrDate'

require "PhotoShelterHelper"
require "PhotoShelterAPI"
require "PhotoShelterSync"
require "Trapper"
require "Info"

local function showDialog(context, publishedServices)
	local f = LrView.osFactory()
	local bind = LrView.bind
	local share = LrView.share

	local pt = LrBinding.makePropertyTable(context)

	-- Set up default values
	for _, service in ipairs(publishedServices) do
		pt[service.localIdentifier .. "_sync"] = true
	end

	-- Create the dialog
	local column1 = {
		spacing = f:control_spacing(),
	}

	for _, service in ipairs(publishedServices) do
		table.insert(column1, f:row {
			fill_horizontal = 1,
			f:static_text {
				title = service:getName(),
				alignment = "left",
				width = share "column1width",
			}
		})

	end

	local column2 = {
		spacing = f:control_spacing(),
	}

	for _, service in ipairs(publishedServices) do
		table.insert(column2, f:row {
			width = share "column2width",
			f:spacer { fill_horizontal = 1 },
			f:checkbox {
				title = LOC "",
				value = bind(service.localIdentifier .. "_sync"),
			},
			f:spacer { fill_horizontal = 1 },
		})
	end

	local c = f:column {
		bind_to_object = pt,
		spacing = f:control_spacing(),
		f:column {
			spacing = f:control_spacing(),
			f:static_text {
				title = LOC("$$$/PhotoShelter/SyncFrom/InstructionsL1=Synchronize PhotoShelter collections and galleries into Lightroom.")
			},
			f:static_text {
				title = LOC("$$$/PhotoShelter/SyncFrom/InstructionsL2=This process may take a very long time depending on the size of your account and will override any change you made in directly into PhotoShelter.")
			},
			f:static_text {
				title = LOC("$$$/PhotoShelter/SyncFrom/InstructionsL3=Please avoid publishing or making changes to your PhotoShelter account during this process.")
			}
		},
		f:row {
			f:static_text {
				spacing = f:control_spacing(),
				title = LOC "$$$/PhotoShelter/PhotoShelter/SupportLink=Click here to learn more^n",
				mouse_down = function()
					local LrHttp = import 'LrHttp'
					LrHttp.openUrlInBrowser('https://support.photoshelter.com/hc/en-us/articles/203373590-PhotoShelter-s-Official-Adobe-Lightroom-Plugin#exporting-publishing')
				end,
				text_color = import 'LrColor'(0, 0, 1),
			},
		},
		f:row {
			spacing = f:control_spacing(),
			f:column {
				fill_vertical = 1,
				f:spacer { fill_vertical = 1 },
				f:static_text {
					width = share "column1width",
					title = LOC "$$$/PhotoShelter/PhotoShelter/PublishService=Account Name",
				},
				f:spacer { fill_vertical = 1 },
			},
			f:column {
				f:spacer {
					width = share "column1width"
				}
			},
			f:column {
				f:static_text {
					width = share "column2width",
					title = LOC "$$$/PhotoShelter/PhotoShelter/SyncFrom=Synchronize",
					alignment = "center"
				},
			},
		},
		f:separator { fill_horizontal = 1 },
		f:row {
			spacing = f:control_spacing(),
			f:column(column1),
			f:column {
				f:spacer {
					width = share "column1width"
				}
			},
			f:column(column2)
		},

	}

	-- Only display the dialog if there's more that one service to choose from
	return LrDialogs.presentModalDialog {
		title = LOC "$$$/PhotoShelter/SyncFrom/Title=Synchronize from PhotoShelter",
		contents = c,
		actionVerb = LOC "$$$/PhotoShelter/PhotoShelter/Sync=Synchronize",
		save_frame = "syncFromDialogPosition",
	}, pt
end

local function breadthFirstSync(getParentsChildrenCallback, getChildCallback, authToken, publishService, scopeData)
	local queue = { { 'root', 'root' } }
	local head = 1
	local tail = 2
	local progressScope = scopeData.progressScope

	local function enqueue(obj)
		queue[tail] = obj
		tail = tail + 1
	end

	local function dequeue()
		if head == tail then
			return nil
		end
		local obj = queue[head]
		head = head + 1
		return obj
	end

	local parent = dequeue()

	while parent ~= nil do
		local success, children = getParentsChildrenCallback(parent, authToken, publishService)

		if not success then
			LrErrors.throwUserError(children)
			return
		elseif success and not children then
			return
		end

		if type(children) == 'string' then
			local serviceName = PhotoShelterInfo.isLibris and 'Libris' or 'PhotoShelter'
			local message = "Upload images to your " .. serviceName .. " account to create a direct sync between the images in your Lightroom catalog and " .. serviceName .. "."
			LrDialogs.message('There are no images in your ' .. serviceName .. ' account to sync', message)
			return
		end

		for _, child in ipairs(children) do
			if progressScope:isCanceled() then
				return
			end

			if child.Collection then
				local collection = getChildCallback(child, true, parent, publishService, scopeData, authToken)
				enqueue(collection)
			else
				-- child is a Gallery
				getChildCallback(child, false, parent, publishService, scopeData, authToken)
			end
		end
		scopeData.suffix = 2
		parent = dequeue()
	end

	if parent == nil then
		return
	end
end

--- Determine new LR parents for photos that are in multiple galleries on PhotoShelter.
-- @param photo LrPublishedPhoto object
-- @param parents Array of PhotoShelter gallery objects
-- @param photosToReparent
local function assignNewParentToPubPhoto(photo, parents, photosToReparent, scopeData)
	for _, parent in ipairs(parents) do
		if scopeData.LrCollectionIDs[parent.gallery_id] then
			photosToReparent[parent.gallery_id] = photosToReparent[parent.gallery_id] or {}
			table.insert(photosToReparent[parent.gallery_id], photo)
		end
	end
	return photosToReparent
end

--- Mark LR Photos that no longer exist on PhotoShelter for cleanup.
local function markToUnPublishPhoto(photo, publishedGals, photosToRemove)
	DBG("Photo " .. photo:getRemoteId() .. " no longer on PhotoShelter. Removing it from " .. #publishedGals .. " Lightroom Publish Sets", nil, 'DEBUG')
	for _, gallery in ipairs(publishedGals) do
		photosToRemove[gallery] = photosToRemove[gallery] or {}
		table.insert(photosToRemove[gallery], photo:getPhoto())
	end
	return photosToRemove
end

-- Un publish photos and remove PhotoShelter Specific Metadata
local function unPublishPhotos(photosToRemove)
	for gallery, photos in pairs(photosToRemove) do
		-- this will throw error is gallery is an smart one.
		gallery:removePhotos(photos)
		for _, photo in ipairs(photos) do
			PhotoShelterHelper.clearMetadata(photo)
		end
	end
end

--- Check that LrPublishedPhoto will be unique to its destination if re parented there.
-- Helper function to check uniqueness of LrPublishedPhoto within LrPublishedCollection
-- @param publishedPhoto LrPublishedPhoto object
-- @param lrCollection LrPublishedCollection object
local function pubPhotoIsUnique(publishedPhoto, lrCollection)
	local lrPhoto = publishedPhoto:getPhoto()
	local lrPhotoParents = lrPhoto:getContainedPublishedCollections()
	if PhotoShelterHelper.contains(lrPhotoParents, lrCollection) then
		DBG("Photo " .. publishedPhoto:getRemoteId() .. " already exists in " .. PhotoShelterHelper.pubCollectionNameAndId(lrCollection) .. ". Won't republish it", nil, 'DEBUG')
		return false
	else
		DBG("Photo " .. publishedPhoto:getRemoteId() .. " doesn't exist in " .. PhotoShelterHelper.pubCollectionNameAndId(lrCollection) .. ". Republishing it.", nil, 'DEBUG')
		return true
	end
end

--- Add LrPublishedPhoto objects to a new LrPublishedCollection
-- @param photos Array of LrPublishedPhoto objects
-- @param lrCollection LrPublishedCollection object - the destination gallery for republishing.
local function reparentPhotos(photos, lrCollection)
	DBG("Republishing " .. #photos .. " to " .. PhotoShelterHelper.pubCollectionNameAndId(lrCollection) .. ".", nil, 'DEBUG')
	LrApplication.activeCatalog():withWriteAccessDo(
			LOC("$$$/PhotoShelter/Sync/SyncPhotoShelterServices=Synchronize PhotoShelter Services"),
			function()
				lrCollection:addPublishedPhotos(photos)
			end,
			{ timeout = 5 }
	)
end

--- Make sure photos are unique to their re parenting destinations to avoid catalog (database) errors
-- `LrPublishedPhoto` objects can exist in multiple `LrPublishedCollection`s, but it must be unique within that collection
-- @param collection String Gallery_ID string
-- @param photosToReparent Array of LrPublishedPhoto objects
local function reparentUniquePhotos(collection, photosToReparent, scopeData)
	local dedupedPhotosToReparent = {}
	local lrCollection = scopeData.LrCollectionIDs[collection]
	DBG(#photosToReparent .. " photos to republish. Checking target destination " .. PhotoShelterHelper.pubCollectionNameAndId(lrCollection), nil, 'DEBUG')

	for _, publishedPhoto in ipairs(photosToReparent) do
		if pubPhotoIsUnique(publishedPhoto, lrCollection) then
			table.insert(dedupedPhotosToReparent, publishedPhoto)
		end
	end

	if #dedupedPhotosToReparent >= 1 then
		reparentPhotos(dedupedPhotosToReparent, lrCollection)
	end
end

--- Remove published photos from gallery that isn't on PhotoShelter anymore.
-- If a photo is no longer on PhotoShelter, remove the corresponding photo from its
-- LrPublishedCollection
-- @param photos Array of LrPhoto objects
local function psDeletedPhotosUnPublish(photos)
	LrApplication.activeCatalog():withWriteAccessDo(
			LOC("$$$/PhotoShelter/SyncFromRemote/UpdatePublishedPhotoStatus=Update published photo status"),
			function()
				unPublishPhotos(photos)
			end,
			{ timeout = 5 }
	)
end

--- Determine status of photos belonging to LrPublishedCollections marked for deletion
-- If photos still exist on PhotoShelter, find the LrPublishedCollections related to their
-- PhotoShelter parents and - if unique - move them to those LrPublishedCollections
-- @param removalSet LrPublishedCollection
-- @param authToken PSAPI authentication token
local function orphanedPhotosRemoveOrReparent(removalSet, authToken, scopeData)
	local orphanedPubPhotos = removalSet:getPublishedPhotos()
	local photosToReparent = {}
	local photosToRemove = {}

	for _, photo in ipairs(orphanedPubPhotos) do
		repeat
			local success, parents = PhotoShelterAPI.getParentGalleries(photo:getRemoteId(), authToken)

			if not success then
				local exists = PhotoShelterAPI.queryImageExists(photo:getRemoteId(), authToken)
				if not exists then
					local publishedGals = photo:getPhoto():getContainedPublishedCollections()
					markToUnPublishPhoto(photo, publishedGals, photosToRemove)
					break
				end
				-- break it because something is not going right. the photo is still in photoshelter but we are unable to read how many galleries contains it.
				LrErrors.throwUserError('We are unable to read Galleries of image id: ' .. photo:getRemoteId())
			end

			if PhotoShelterHelper.tableLength(parents) >= 1 then
				assignNewParentToPubPhoto(photo, parents, photosToReparent, scopeData)
			end
		until true
	end

	for collection, publishedPhotos in pairs(photosToReparent) do
		reparentUniquePhotos(collection, publishedPhotos, scopeData)
	end

	psDeletedPhotosUnPublish(photosToRemove)
end

--- Double check that a previously published Gallery or Collection is no longer on PhotoShelter
local function confirmRemovalState(set, authToken, scopeData)
	local id = set:getRemoteId()
	local success, data
	DBG("Confirming " .. PhotoShelterHelper.pubCollectionNameAndId(set) .. " removal from PhotoShelter.", nil, 'DEBUG')
	if PhotoShelterHelper.isCollection(set) then
		success, data = PhotoShelterAPI.queryCollectionById(id, authToken)
	else
		success, data = PhotoShelterAPI.queryGalleryById(id, authToken)
	end
	if success then
		DBG(PhotoShelterHelper.pubCollectionNameAndId(set) .. " still on server but not synced", nil, 'DEBUG')
		PhotoShelterHelper.removeByKey(scopeData.publishedSets, set)
		return false
	else
		DBG("Confirmed " .. PhotoShelterHelper.pubCollectionNameAndId(set) .. " no longer on PhotoShelter. Updating published status.", nil, 'DEBUG')
		return true
	end
end

--- Clean up gallery or collection that isn't on PhotoShelter anymore
-- @param child LrPublished
local function removeUnPublishedCollectionOrGallery(child)
	LrApplication.activeCatalog():withWriteAccessDo(
			LOC("$$$/PhotoShelter/SyncFrom/Sync=Synchronize PhotoShelter Services"),
			function()
				child:delete()
			end,
			{ timeout = 5 }
	)
end

--- Remove galleries from LR that no longer exist on PhotoShelter
-- For each LrPublishedCollection or LrPublishedCollectionSet that is leftover
-- after a sync (It's no longer on PhotoShelter) remove it
local function removeUnPublishedGallery (pubGallery, authToken, scopeData)
	-- Double check that gallery is no longer on PhotoShelter
	if not confirmRemovalState(pubGallery, authToken) then
		return
	end

	DBG(PhotoShelterHelper.pubCollectionNameAndId(pubGallery) .. " has photos, determining if they should be removed or published in another gallery", nil, 'DEBUG')

	-- Deal with orphaned children
	orphanedPhotosRemoveOrReparent(pubGallery, authToken, scopeData)
	removeUnPublishedCollectionOrGallery(pubGallery)
end

--- Recursively loop over sets that have not been synced to clean them up
-- Handles cleanup of collections and galleries. Loops over collection(sets) that were previously
-- published with the service, but were not found in the current sync process, and have likely
-- been deleted from PhotoShelter.
-- @param setsToUnPublish Array of LrPublishedCollection and LrPublishedCollectionSet objects
-- @param authToken String PSAPI authentication token
local function unpublishedLrSetsParse(setsToUnPublish, authToken, scopeData)
	for id, set in pairs(setsToUnPublish) do
		-- Check if LrPublishedCollection(Set) is still published (deleted pub collections don't have remoteId - or even the method anymore technically)
		local result = LrTasks.pcall(function()
			set:getRemoteId()
		end)

		if result then
			DBG(PhotoShelterHelper.pubCollectionNameAndId(set) .. " no longer on PhotoShelter. Updating published status.", nil, 'DEBUG')
			-- Recurse into child sets
			local isCollection = PhotoShelterHelper.isCollection(set)
			local hasPhotos = PhotoShelterHelper.hasPhotos(set)
			local setChildren = isCollection and set:getChildren() or {}
			local hasChildren = PhotoShelterHelper.tableLength(setChildren) > 0

			if isCollection then
				if hasPhotos or hasChildren then
					-- recursively remove childrens and after that remove it.
					unpublishedLrSetsParse(set:getChildren(), authToken, scopeData)
				end
				-- remote it because it is the last one
				removeUnPublishedCollectionOrGallery(set)
			else
				-- remove gallery - it will confirm that this gallery is not in photoshelter before proceed with the delete.
				removeUnPublishedGallery(set, authToken, scopeData)
			end
		else
			DBG(set:type() .. ' with id: ' .. id .. ' already removed', nil, 'DEBUG')
		end
	end
end

-- Check to see if metadata fields for the child have changed and if so update them
local function checkAndUpdateChild(lrChild, psChild, lrParent, isCollection, scopeData)
	local childSettings = (isCollection and lrChild:getCollectionSetInfoSummary().collectionSettings or lrChild:getCollectionInfoSummary().collectionSettings)

	if lrChild:getName() ~= psChild.name or
			childSettings.Description ~= psChild.description or
			childSettings.ListedOnWebsite ~= psChild.listed or
			childSettings.Visibility ~= psChild.visibility or
			lrChild:getParent() ~= lrParent
	then
		LrApplication.activeCatalog():withWriteAccessDo(
				LOC("$$$/PhotoShelter/Sync/Undo=Synchronize PhotoShelter Services"),
				function()
					DBG('Changes found ... updating ' .. lrChild:getName(), nil, 'DEBUG')
					lrChild:setName(psChild.name)

					if lrChild:getParent() ~= lrParent then
						lrChild:setParent(lrParent)
					end

					childSettings.Description = psChild.description
					childSettings.ListedOnWebsite = psChild.listed
					childSettings.Visibility = psChild.visibility

					if isCollection then
						lrChild:setCollectionSetSettings(childSettings)
					else
						lrChild:setCollectionSettings(childSettings)
					end
				end,
				{ timeout = 5 }
		)
	end

	PhotoShelterHelper.removeByKey(scopeData.publishedSets, lrChild:getRemoteId())

	return lrChild
end

local function createLrCollectionSet (publishService, name, id, lightroomParent, photoShelterChild, scopeData)
	local lrChild

	LrApplication.activeCatalog():withWriteAccessDo(
			LOC("$$$/PhotoShelter/Sync/Undo=Synchronize PhotoShelter Services"),
			function()
				lrChild = publishService:createPublishedCollectionSet(name, lightroomParent, false)
				local settings = PhotoShelterHelper.setLocalSettings(photoShelterChild)

				lrChild:setRemoteId(id)
				lrChild:setCollectionSetSettings(settings)

				-- Index collections to check for duplicates
				scopeData.LrCollectionSetIDs[id] = lrChild
				scopeData.LrCollectionSetNames[LrStringUtils.lower(name)] = lrChild
			end,
			{ timeout = 5 }
	)

	return lrChild
end

local function createLrCollection (publishService, name, id, lightroomParent, photoShelterChild, scopeData)
	local lrChild

	LrApplication.activeCatalog():withWriteAccessDo(
			LOC("$$$/PhotoShelter/Sync/Undo=Synchronize PhotoShelter Services"),
			function()
				lrChild = publishService:createPublishedCollection(name, lightroomParent, false)
				local settings = PhotoShelterHelper.setLocalSettings(photoShelterChild)

				lrChild:setRemoteId(id)
				lrChild:setCollectionSettings(settings)

				-- Index the gallery to compare for duplicates
				scopeData.LrCollectionIDs[id] = lrChild
				scopeData.LrCollectionNames[LrStringUtils.lower(name)] = lrChild
			end,
			{ timeout = 5 }
	)

	return lrChild
end

local function locateEquivalentImage(photoShelterImage, authToken)
	local photoShelterImageId = photoShelterImage.image_id
	if not photoShelterImageId or not photoShelterImage.Image then
		return nil
	end
	local photoShelterFilename = photoShelterImage.Image.file_name

	local result, exif = PhotoShelterAPI.getEXIF(photoShelterImageId, authToken)

	if not result then
		return nil
	end

	local iptcData = photoShelterImage.Image.Iptc

	local subSecond = PhotoShelterHelper.getSubSecondFromExif(exif)
	local captureDate = PhotoShelterHelper.getCaptureDate(exif)

	if captureDate then
		local captureTimestamp = PhotoShelterHelper.convertToTimestamp(captureDate)
		-- malformed EXIF captureDate returns false above
		if captureTimestamp then
			captureDate = LrDate.timeToW3CDate(captureTimestamp)
		end
	end

	local foundPhotos = PhotoShelterHelper.searchPhotos(photoShelterImage.Image.file_name, captureDate, subSecond)

	-- If we have matches, narrow them down further
	if #foundPhotos > 1 then
		local maxScore = 0
		local resultList = {}

		for _, photo in pairs(foundPhotos) do
			local matchScore = PhotoShelterHelper.calculateMatch(photoShelterImage, photo, iptcData)

			if matchScore >= maxScore then
				if matchScore > maxScore then
					resultList = {}
				end

				maxScore = matchScore
				table.insert(resultList, photo)
			end
		end

		-- If we have a decent score then use this results list
		if maxScore > 32 then
			foundPhotos = resultList
		end
	end

	-- If we have exactly one match, then we match it up
	if #foundPhotos == 1 then
		return foundPhotos[1]
	end
end

local function syncImages(name, id, lrChild, authToken, scopeData)
	DBG('Attempting to associate LR images for PhotoShelter gallery ' .. name, nil, 'DEBUG')
	local lrPhoto
	local userURL = PhotoShelterAPI.getSiteURL(authToken)
	-- todo: improve this to do something like a while loop to ensure we don't saturate http request on bigger galleries
	local success, psPhotos = PhotoShelterAPI.queryImagesByGallery(id, authToken)

	if not success then
		LrErrors.throwUserError(psPhotos)
		return
	end

	local currentPublishedPhotos = lrChild:getPublishedPhotos()


	-- some photos are not anymore in photoshelter so we will try to discover it an remove the remote id
	-- so they will become an un published image.
	local removedElements = {}
	for _, lrPhoto in pairs(currentPublishedPhotos) do
		local match
		for _, psPhoto in pairs(psPhotos) do
			if psPhoto.image_id == lrPhoto:getRemoteId() then
				match = lrPhoto
				break -- found
			end
		end
		-- if it is not i ps photos it should be deleted
		if not match then
			removedElements[#removedElements + 1] = lrPhoto
		end
	end

	if #removedElements > 0 then
		scopeData.needPublish[#scopeData.needPublish + 1] = lrChild

		local _, error = LrFunctionContext.pcallWithContext(
				'Remove photos from gallery that are no in Gallery at PhotoShelter.',
				function()
					local lrPhotos = {}

					for _, pubPhoto in pairs(removedElements) do
						lrPhotos[#lrPhotos + 1] = pubPhoto:getPhoto()
					end

					LrApplication.activeCatalog():withWriteAccessDo(
							"Remove orphan photos from Gallery: " .. lrChild:getName(),
							function()
								if not lrChild:isSmartCollection() then
									lrChild:removePhotos(lrPhotos)
								else
									for _, lrPhoto in pairs(removedElements) do
										-- when the gallery is smart we need to remove the remote id/url from the photo and the metadata of the photo.
										PhotoShelterHelper.clearMetadata(lrPhoto)
										-- todo: check if this work.
										lrPhoto:setRemoteId("")
										lrPhoto:setRemoteUrl("")
									end
								end
							end, { timeout = 5 })
				end)

		if error then
			LrErrors.throwUserError(error)
		end
	end

	if #psPhotos > 0 then
		for _, photo in ipairs(psPhotos) do
			if scopeData.previouslyPublishedPhotos[photo.image_id] then
				-- We've already got this image ID as a published image.
				-- No need to run the expensive locateEquivalentImage task.
				lrPhoto = scopeData.previouslyPublishedPhotos[photo.image_id]
			else
				lrPhoto = locateEquivalentImage(photo, authToken)
			end

			if lrPhoto then
				-- Found a match, make sure it's up to date
				DBG('Found match for image ' .. photo.image_id, nil, 'DEBUG')
				LrApplication.activeCatalog():withWriteAccessDo(
						'Associate photo with PhotoShelter gallery',
						function()
							local url = userURL .. "/image/" .. photo.image_id
							lrChild:addPhotoByRemoteId(lrPhoto, photo.image_id, url, true)
							lrPhoto:setPropertyForPlugin(_PLUGIN, 'id', photo.image_id)
							lrPhoto:setPropertyForPlugin(_PLUGIN, 'onServer', true)
							lrPhoto:setPropertyForPlugin(
									_PLUGIN, 'propRelease', (photo.property_released == 't' and true or nil)
							)
							lrPhoto:setPropertyForPlugin(_PLUGIN, 'modelRelease', (photo.model_released == 't' and true or nil))
						end,
						{ timeout = 5 }
				)
				table.insert(scopeData.pluginPublishedPhotos, photo.image_id)
			end
		end
	end
end

local function synchronizeChild(child, isCollection, parent, publishService, scopeData, authToken)
	local lrParent = PhotoShelterHelper.setLRParent(parent)
	local lrChild
	local photoShelterChild = PhotoShelterHelper.setupPSChild(child)
	local name = photoShelterChild.name
	local id = photoShelterChild.id
	local progressScope = scopeData.progressScope

	if isCollection then
		if scopeData.LrCollectionSetIDs[id] then
			DBG(' A collection with the id (' .. id .. ') already exists. Checking for updates.', nil, 'DEBUG')
			-- Collection already exists locally - make sure it's up to date
			lrChild = checkAndUpdateChild(
					scopeData.LrCollectionSetIDs[id],
					photoShelterChild,
					lrParent,
					true,
					scopeData
			)
		elseif scopeData.LrCollectionSetNames[LrStringUtils.lower(name)] or scopeData.LrCollectionNames[LrStringUtils.lower(name)] then
			-- A collection with this name already exists, append a suffix if necessary (can't have the same parent)
			DBG(' A collection named ' .. name .. ' already exists. Checking for siblings and appending a suffix')
			name = PhotoShelterHelper.suffixName(
					PhotoShelterHelper.createSiblingNamesTable(lrParent, publishService),
					name,
					lrParent,
					scopeData
			)

			lrChild = createLrCollectionSet(publishService, name, id, lrParent, photoShelterChild, scopeData)
		else
			DBG(' New remote collection' .. name .. ' with ID: ' .. id .. ' creating LR Collection Set', nil, 'DEBUG')
			lrChild = createLrCollectionSet(publishService, name, id, lrParent, photoShelterChild, scopeData)
		end

		return { photoShelterChild, lrChild }
	else
		-- is a Gallery
		if scopeData.LrCollectionIDs[id] then
			-- A gallery with this ID already exists
			DBG(' A gallery with the id (' .. id .. ') already exists. Checking for updates.', nil, 'DEBUG')
			lrChild = checkAndUpdateChild(
					scopeData.LrCollectionIDs[id],
					photoShelterChild,
					lrParent,
					false,
					scopeData
			)
		elseif scopeData.LrCollectionNames[LrStringUtils.lower(name)] or scopeData.LrCollectionSetNames[LrStringUtils.lower(name)] then
			-- A gallery with this name exists, append a suffix if necessary (can't have the same parent)
			DBG(' A gallery named ' .. name .. ' already exists. Checking for siblings and appending a suffix.', nil, 'DEBUG')
			name = PhotoShelterHelper.suffixName(
					PhotoShelterHelper.createSiblingNamesTable(lrParent, publishService),
					name,
					lrParent,
					scopeData
			)
			lrChild = createLrCollection(publishService, name, id, lrParent, photoShelterChild, scopeData)
			-- Handle the case where a gallery has a collection sibling with the same name
			-- Could be relatively common since the move to unite archive and portfolio back in the day
			-- Made galleries of portfolio items that had the same name as archive collections.
		elseif scopeData.LrCollectionSetNames[LrStringUtils.lower(name)] then
			DBG(' A collection named ' .. name .. ' already exists. Checking for siblings and appending a suffix.', nil, 'DEBUG')
			name = PhotoShelterHelper.suffixName(scopeData.LrCollectionSetNames, name, lrParent, scopeData)
			lrChild = createLrCollection(publishService, name, id, lrParent, photoShelterChild, scopeData)
		else
			DBG(' New remote gallery ' .. name .. ' with ID: ' .. id .. ' creating LR Collection', nil, 'DEBUG')
			lrChild = createLrCollection(publishService, name, id, lrParent, photoShelterChild, scopeData)
		end

		scopeData.scannedGalleries = scopeData.scannedGalleries + 1
		progressScope:setPortionComplete(scopeData.scannedGalleries, scopeData.numGalleries)

		syncImages(name, id, lrChild, authToken, scopeData)
	end
end

-- Gets the local LR children collections to build a reference table to the local
-- publish service hierarchy on subsequent synchronizations.
local function getChildrenCollections(parent, authToken, publishService)
	local lrParent = PhotoShelterHelper.setLRParent(parent)

	local collectionSets
	local collections

	if lrParent == nil then
		DBG(' Getting local root children', nil, 'DEBUG')
		collectionSets = publishService:getChildCollectionSets()
		collections = publishService:getChildCollections()
	else
		collectionSets = lrParent:getChildCollectionSets()
		collections = lrParent:getChildCollections()
	end
	for _, child in ipairs(collectionSets) do
		child.Collection = true
	end
	for i = 1, #collections do
		collectionSets[#collectionSets + 1] = collections[i]
	end

	return true, collectionSets
end

-- Builds a reference table to the local hierarchy to help sync
local function buildPublishedHierarchy(child, isCollection, parent, publishService, scopeData)
	local remoteId = child:getRemoteId()
	if isCollection then
		if remoteId then
			scopeData.LrCollectionSetIDs[remoteId] = child
			scopeData.publishedSets[remoteId] = child
		else
			table.insert(scopeData.unpublishedSets, child)
		end
		scopeData.LrCollectionSetNames[LrStringUtils.lower(child:getName())] = child
		return { 'child', child }
	else
		if remoteId then
			scopeData.LrCollectionIDs[remoteId] = child
			scopeData.publishedSets[remoteId] = child
		else
			table.insert(scopeData.unpublishedSets, child)
		end
		scopeData.LrCollectionNames[LrStringUtils.lower(child:getName())] = child
	end
end

local function uploadUnpublishedCollections (publishService, authToken, scopeData)
	local unpublishedSets = scopeData.unpublishedSets

	for _, set in ipairs(unpublishedSets) do
		local isCollection = set:type() == 'LrPublishedCollectionSet'
		local message = isCollection and 'Creating new PhotoShelter Collection' .. set:getName() or 'Creating new PhotoShelter Gallery ' .. set:getName()
		DBG(message, nil, 'DEBUG')
		local collectionSettings

		if isCollection then
			collectionSettings = set:getCollectionSetInfoSummary().collectionSettings or {}
		else
			collectionSettings = set:getCollectionInfoSummary().collectionSettings or {}
		end

		collectionSettings.Description = collectionSettings.Description or ""
		collectionSettings.ListedOnWebsite = collectionSettings.ListedOnWebsite or false
		collectionSettings.Visibility = collectionSettings.Visibility or "inherit"

		local parent = set:getParent()
		local parentID = (parent and parent:getRemoteId() or 'root')

		local childType = (isCollection and 'collection' or 'gallery')
		local success, childCollectionID = PhotoShelterAPI.createGalleryOrCollection(
				childType, set:getName(),
				collectionSettings.Description,
				collectionSettings.ListedOnWebsite,
				collectionSettings.Visibility, parentID,
				authToken
		)

		DBG('New PhotoShelter ' .. childType .. ' ID: ' .. childCollectionID, nil, 'DEBUG')

		if success then
			LrApplication.activeCatalog():withWriteAccessDo(
					LOC("$$$/PhotoShelter/Sync/Undo=Synchronize PhotoShelter Services"),
					function()
						set:setRemoteId(childCollectionID)
					end,
					{ timeout = 5 }
			)

			if isCollection then
				scopeData.LrCollectionSetIDs[childCollectionID] = set
				scopeData.LrCollectionSetNames[set:getName()] = set
			else
				scopeData.LrCollectionIDs[childCollectionID] = set
				scopeData.LrCollectionNames[set:getName()] = set
			end
		end
	end
end

LrFunctionContext.postAsyncTaskWithContext("Force Sync from PhotoShelter", function(context)
	Trapper.TrapContextErrors(context)
	local oldPluginNamespace = "nl.pactsoftware.lightroom.export.PhotoshelterArchive"
	local catalog = LrApplication.activeCatalog()
	local oldPluginPublishedServices = catalog:getPublishServices(oldPluginNamespace)
	local publishedServices = catalog:getPublishServices(PhotoShelterInfo.namespace)
	-- define global variables DBG with pt as first param of each log generated by DBG("")
	PhotoShelterHelper.setUpDebug(
			{
				feature="SFPS",
				currentServices = publishedServices,
				oldServices = oldPluginPublishedServices,
				catalog = catalog,
			},
			false
	)

	if publishedServices and #publishedServices > 0 then
		-- if the response is not ok, just cancel it.
		local dialogResult, pt = showDialog(context, publishedServices)

		if dialogResult ~= "ok" then
			return
		end

		local serviceCount = 0

		local serviceProgressScope = LrProgressScope(
				{
					title = LOC(
							"$$$/PhotoShelter/Sync=Synchronising Services from PhotoShelter"
					),
					functionContext = context
				}
		)
		serviceProgressScope:setCancelable(true)
		serviceProgressScope:setPortionComplete(0, #publishedServices)

		local searchDesc = {}

		-- because in the catalog we could find photos that exists in photoshelter but was upload by the previous
		-- plugin version, so we need to figure out those photos too to could synchronize the pictures.
		for _, oldService in ipairs(oldPluginPublishedServices) do
			local pluginID = oldService:getPluginId()
			searchDesc[#searchDesc + 1] = {
				criteria = 'sdk:' .. pluginID .. '.onServer',
				operation = '==',
				value = true,
			}
		end

		-- append new plugin to the search criteria.
		for _, publishService in ipairs(publishedServices) do
			local selected = pt[publishService.localIdentifier .. "_sync"]

			if selected then
				local pluginID = publishService:getPluginId()
				searchDesc[#searchDesc + 1] = {
					criteria = 'sdk:' .. pluginID .. '.onServer',
					operation = '==',
					value = true,
				}
			end
		end

		searchDesc.combine = "union" -- like an OR so we get photos that belongs to any of those services.

		local psPublishedPhotos = catalog:findPhotos {
			searchDesc = searchDesc
		}

		-- a lookup table of all photos that could match with the current one in photoshelter.
		local previouslyPublishedPhotos = {}

		for _, photo in ipairs(psPublishedPhotos) do
			local remoteIdFromOldMeta, oldPluginMetaError = photo:getPropertyForPlugin(oldPluginNamespace, 'id', nil, true)
			local remoteIdFromMeta, error = photo:getPropertyForPlugin(_PLUGIN, 'id', nil, true)
			local remoteId

			if not oldPluginMetaError then
				remoteId = remoteIdFromOldMeta
			end
			-- no mather if we found from old one, if we get some reference from the new one we will take that
			if not error and remoteIdFromMeta then
				remoteId = remoteIdFromMeta
			end

			if remoteId then
				previouslyPublishedPhotos[remoteId] = photo
			end
		end

		for i, publishService in ipairs(publishedServices) do
			repeat
				local selected = pt[publishService.localIdentifier .. "_sync"]

				if not selected then
					serviceProgressScope:setPortionComplete(i, #publishService)
					break
				end

				local authToken = PhotoShelterHelper.validateLogon(publishService:getPublishSettings())
				local success, colAndGalCount = PhotoShelterAPI.getGalleryAndCollectionCount(authToken)

				if not success then
					LrError.throwUserError('We was unable to retrieve galleries and collections for service ' .. publishService:getName())
				end

				local operationsCount = colAndGalCount
				local progressScope = LrProgressScope(
						{
							title = LOC(
									"$$$/PhotoShelter/Sync=Synchronising ^1 Service (Remote to Local)",
									publishService:getName()
							),
							functionContext = context
						}
				)
				progressScope:setCancelable(true)
				progressScope:setPortionComplete(0, operationsCount)

				local scopeData = {
					progressScope = progressScope,
					numGalleries = colAndGalCount,
					scannedGalleries = 0,
					suffix = nil,
					LrCollectionSetIDs = {},
					LrCollectionSetNames = {},
					LrCollectionIDs = {},
					LrCollectionNames = {},
					publishedSets = {},
					unpublishedSets = {},
					previouslyPublishedPhotos = previouslyPublishedPhotos,
					pluginPublishedPhotos = {},
					needPublish = {}
				}

				if serviceProgressScope and not serviceProgressScope:isCanceled() and progressScope and not progressScope:isCanceled() then
					local localCollectionCount = #publishService:getChildCollectionSets()
					local localGalleriesCount = #publishService:getChildCollections()

					-- if we detect that local service contain collections and galleries we create a lookup table to reference
					-- it later.
					if (localCollectionCount + localGalleriesCount) > 0 then
						-- create lookup table
						breadthFirstSync(getChildrenCollections, buildPublishedHierarchy, nil, publishService, scopeData)
						-- upload any un published collection/gallery found
						uploadUnpublishedCollections(publishService, authToken, scopeData)
					end

					-- iterate the photoshelter galleries to match with the local copy.
					breadthFirstSync(PhotoShelterAPI.getChildren, synchronizeChild, authToken, publishService, scopeData)

					if PhotoShelterHelper.tableLength(scopeData.publishedSets) >= 1 then
						unpublishedLrSetsParse(scopeData.publishedSets, authToken, scopeData)
					end
				end
				progressScope:done()
				serviceCount = serviceCount + 1
				serviceProgressScope:setPortionComplete(serviceCount, #publishedServices)

				if #scopeData.needPublish > 0 then
					for _, child in pairs(scopeData.needPublish) do
						if child:type() == 'LrPublishedCollection' then
							child:publishNow()
						end
					end
				end
			until true
		end

		serviceProgressScope:done()
		LrDialogs.showBezel('Successfully synced with PhotoShelter')
	else
		LrDialogs.showError("We are unable to detect any PhotoShelter plugin.")
	end
end)
