local LrFunctionContext = import 'LrFunctionContext'
local LrApplication = import 'LrApplication'
local LrProgressScope = import 'LrProgressScope'
local LrBinding = import 'LrBinding'
local LrView = import 'LrView'
local LrDialogs = import 'LrDialogs'

require "PhotoShelterHelper"
require "PhotoShelterSync"
require "Trapper"
require "Info"

local showDialog = function(context, publishedServices)
	local f = LrView.osFactory()
	local bind = LrView.bind
	local share = LrView.share

	local pt = LrBinding.makePropertyTable(context)

	-- Set up default values
	for _, service in ipairs(publishedServices) do
		pt[service.localIdentifier .. "_sync"] = true
	end

	-- Create the dialog
	local column1 = {
		spacing = f:control_spacing(),
	}

	for _, service in ipairs(publishedServices) do
		table.insert(column1, f:row {
			fill_horizontal = 1,
			f:static_text {
				title = service:getName(),
				alignment = "left",
				width = share "column1width",
			}
		})

	end

	local column2 = {
		spacing = f:control_spacing(),
	}

	for _, service in ipairs(publishedServices) do
		table.insert(column2, f:row {
			width = share "column2width",
			f:spacer { fill_horizontal = 1 },
			f:checkbox {
				title = LOC "",
				value = bind(service.localIdentifier .. "_sync"),
			},
			f:spacer { fill_horizontal = 1 },
		})
	end

	local c = f:column {
		bind_to_object = pt,
		spacing = f:control_spacing(),
		f:row {
			spacing = f:control_spacing(),
			f:static_text {
				title = LOC("$$$/PhotoShelter/SyncAll/Instructions=Synchronize Lightroom collections^nand galleries into PhotoShelter.^n^nThis process may take a very long time depending on^nthe size of your account and will override any change you made in directly into PhotoShelter.^n^nPlease avoid publishing or making changes to your^nPhotoShelter account during this process.")
			}
		},
		f:row {
			f:static_text {
				spacing = f:control_spacing(),
				title = LOC "$$$/PhotoShelter/SyncLightroom/SupportLink=Click here to learn more^n",
				mouse_down = function()
					local LrHttp = import 'LrHttp'
					LrHttp.openUrlInBrowser('https://support.photoshelter.com/hc/en-us/articles/203373590-PhotoShelter-s-Official-Adobe-Lightroom-Plugin#exporting-publishing')
				end,
				text_color = import 'LrColor'(0, 0, 1),
			},
		},
		f:row {
			spacing = f:control_spacing(),
			f:column {
				fill_vertical = 1,
				f:spacer { fill_vertical = 1 },
				f:static_text {
					width = share "column1width",
					title = LOC "$$$/PhotoShelter/SyncAll/PublishService=Account Name",
				},
				f:spacer { fill_vertical = 1 },
			},
			f:column {
				f:spacer {
					width = share "column1width"
				}
			},
			f:column {
				f:static_text {
					width = share "column2width",
					title = LOC "$$$/PhotoShelter/SyncAll/SyncHierarchy=Synchronize",
					alignment = "center"
				},
			},
		},
		f:separator { fill_horizontal = 1 },
		f:row {
			spacing = f:control_spacing(),
			f:column(column1),
			f:column {
				f:spacer {
					width = share "column1width"
				}
			},
			f:column(column2)
		},

	}

	-- Only display the dialog if there's more that one service to choose from
	return LrDialogs.presentModalDialog {
		title = LOC "$$$/PhotoShelter/SyncAll/Title=Synchronize to PhotoShelter",
		contents = c,
		actionVerb = LOC "$$$/PhotoShelter/SyncAll/Sync=Synchronize",
		save_frame = "syncAllDialogPosition",
	}, pt
end

LrFunctionContext.postAsyncTaskWithContext("Force Sync to PhotoShelter", function(context)
	Trapper.TrapContextErrors(context)
	local catalog = LrApplication.activeCatalog()
	local publishedServices = catalog:getPublishServices(PhotoShelterInfo.namespace)
	PhotoShelterHelper.setUpDebug(
			{
				feature="STPS",
				currentServices = publishedServices,
				catalog = catalog,
			},
			false
	)

	if publishedServices and #publishedServices > 0 then
		-- if the response is not ok, just cancel it.
		local dialogResult, pt = showDialog(context, publishedServices)
		if dialogResult ~= "ok" then
			return
		end

		local serviceCount = 0

		local serviceProgressScope = LrProgressScope(
				{
					title = LOC(
							"$$$/PhotoShelter/Sync=Synchronising to PhotoShelter Services"
					),
					functionContext = context
				}
		)
		serviceProgressScope:setCancelable(true)
		serviceProgressScope:setPortionComplete(0, #publishedServices)

		for _, publishService in ipairs(publishedServices) do
			local count = 0
			local selected = pt[publishService.localIdentifier .. "_sync"]
			if selected then
				local authToken = PhotoShelterHelper.validateLogon(publishService:getPublishSettings())

				local collections = 0
				local galleries = 0
				local rootGalleries = publishService:getChildCollections()
				local rootCollections = publishService:getChildCollectionSets()

				-- filter galleries by just the published
				for _, gallery in ipairs(rootGalleries) do
					if gallery:getRemoteId() then
						galleries = galleries + 1
					end
				end

				local collectCollections

				collectCollections = function(collection)
					if collection:getRemoteId() then
						if collection:type() == 'LrPublishedCollectionSet' then
							collections = collections + 1
							for _, nestedCollection in ipairs(collection:getChildCollectionSets()) do
								collectCollections(nestedCollection)
							end
							for _, nestedGallery in ipairs(collection:getChildCollections()) do
								if nestedGallery:getRemoteId() then
									galleries = galleries + 1
								end
							end
						else
							galleries = galleries + 1
						end
					end
				end

				-- get all collection set from the service at root level.
				for _, collection in ipairs(rootCollections) do
					-- iterate over each one and call the recursive function to pickup nested collection set.
					-- it also take any gallery.
					-- it will skip any non published collection set / collection
					collectCollections(collection)
				end

				local operationsCount = collections + galleries
				local progressScope = LrProgressScope(
						{
							title = LOC(
									"$$$/PhotoShelter/Sync=Synchronising ^1 Service (Local to Remote)",
									publishService:getName()
							),
							functionContext = context
						}
				)
				progressScope:setCancelable(true)
				progressScope:setPortionComplete(0, operationsCount)

				local checkGalleryPhotos = function(gallery, oldGalleryId)
					local galleryId = gallery:getRemoteId()
					local pubPhotos = gallery:getPublishedPhotos()
					local needCallPublish = false

					for _, photo in pairs(pubPhotos) do
						repeat
							local photoId = photo:getRemoteId()
							local exists = PhotoShelterAPI.queryImageExists(photoId, authToken)

							if not exists then
								-- clean and allow LRC to delete it
								publishService.catalog:withPrivateWriteAccessDo(
										function()
											-- clean to ensure publish process update it.
											photo:setRemoteId(nil)
											photo:setRemoteUrl('')
											photo:setPropertyForPlugin(_PLUGIN, 'id', nil)
											photo:setPropertyForPlugin(_PLUGIN, 'onServer', false)
										end
								)
								needCallPublish = true -- need call publish on the gallery to push the images.
								break -- continue with next image
							else
								local success, galleriesOrMessage = PhotoShelterAPI.getParentGalleries(photoId, authToken)
								if not success then
									-- could not be removed so prevent to insert it on removed table
									DBG('Unable to read parent galleries for image ' .. photoId, nil, 'DEBUG')
									break
								end

								local galleryFound = false

								if #galleriesOrMessage > 0 then
									for _, gallery in pairs(galleriesOrMessage) do
										if gallery.gallery_id == oldGalleryId then
											galleryFound = true
										end
									end
								end

								-- if we found this gallery on the current photoshelter parent galleries we proceed to remove photo from gallery
								-- otherwise we just update the galleryReferenceCount and call the deletedCallback
								if galleryFound then
									local result, msg = PhotoShelterAPI.removeImageFromGallery(oldGalleryId, photoId, authToken)

									if not result then
										DBG('We was unable to remove the photo ' .. photoId .. ' from gallery ' .. oldGalleryId, {error=msg}, 'ERROR')
									end
								end

								local success, error = PhotoShelterAPI.addImageToGallery(galleryId, photoId, authToken)
								if not success then
									DBG('We was unable to add photo ' .. photoId .. ' to gallery ' .. galleryId, {error=error}, 'ERROR')
								end
							end
						until true;
					end

					if needCallPublish then
						DBG('Calling gallery publish now for gallery ' .. gallery:getName(), nil, 'DEBUG')
						gallery:publishNow()
					end
				end

				local recreateGallery = function(gallery)
					DBG('Creating gallery ' .. gallery:getName() .. ' on PhotoShelter', nil, 'DEBUG')
					local parent = gallery:getParent() -- it take always the fresh information from the parent.
					local parentCollectionID
					
					if parent then
						parentCollectionID = parent:getRemoteId()
					else
						parentCollectionID = "root"
					end
					
					local collectionSettings = gallery:getCollectionInfoSummary().collectionSettings;
					-- Create new gallery
					local oldGalleryId = gallery:getRemoteId()
					local result, idOrMsg = PhotoShelterAPI.createGallery(
							gallery:getName(),
							collectionSettings.Description,
							collectionSettings.ListedOnWebsite,
							collectionSettings.Visibility,
							parentCollectionID,
							authToken
					)

					if not result then
						-- Error
						LrErrors.throwUserError(idOrMsg)
					end

					if parentCollectionID == nil then
						parentCollectionID = collectionSettings.ListedOnWebsite == 't' and 'l' or 'u'
					end

					gallery.catalog:withWriteAccessDo(
							'Store new Collection ID',
							function()
								gallery:setRemoteId(idOrMsg)
								gallery:setRemoteUrl('http://www.photoshelter.com/mem/images/index#/' .. parentCollectionID .. '/' .. idOrMsg)
							end,
							{ timeout = 5 }
					)

					checkGalleryPhotos(gallery, oldGalleryId)

					count = count + 1
					progressScope:setPortionComplete(count, operationsCount)
				end

				local recreateCollection

				recreateCollection = function(collection)
					DBG('Creating new collection ' .. collection:getName() .. ' on PhotoShelter', nil, 'DEBUG')
					local parent = collection:getParent() -- it return always fresh information
					local parentCollectionID

					if parent then
						parentCollectionID = parent:getRemoteId()
					else
						parentCollectionID = "root"
					end

					local collectionSettings = collection:getCollectionSetInfoSummary().collectionSettings;
					local result, idOrMsg = PhotoShelterAPI.createCollection(
							collection:getName(),
							collectionSettings.Description,
							collectionSettings.ListedOnWebsite,
							collectionSettings.Visibility,
							parentCollectionID,
							authToken
					)

					if not result then
						-- Error
						LrErrors.throwUserError(idOrMsg)
					end

					collection.catalog:withWriteAccessDo(
							'Store new Collection ID',
							function(context)
								-- Do something that reads or writes catalog
								collection:setRemoteId(idOrMsg)
								collection:setRemoteUrl("http://www.photoshelter.com/mem/images/index#/" .. idOrMsg)
							end,
							{ timeout = 5 }
					)

					for _, gallery in ipairs(collection:getChildCollections()) do
						recreateGallery(gallery)
				end

					for _, collection in ipairs(collection:getChildCollectionSets()) do
						recreateCollection(collection)
					end

					count = count + 1
					progressScope:setPortionComplete(count, operationsCount)
				end

				if serviceProgressScope and not serviceProgressScope:isCanceled() and progressScope and not progressScope:isCanceled() then
					-- root galleries verification
					local verifyGalleries

					verifyGalleries = function(galleries)
						for _, gallery in ipairs(galleries) do
							local parent = gallery:getParent()
							local parentCollectionID
							if parent then
								parentCollectionID = parent:getRemoteId()
							else
								parentCollectionID = "root"
							end

							local collectionSettings = gallery:getCollectionInfoSummary().collectionSettings;
							local exists = PhotoShelterAPI.galleryOrCollectionExists(gallery, authToken);

							if not exists then
								local response = LrDialogs.confirm(
										LOC("$$$/PhotoShelter/MissingGallery/Message=Missing Gallery in PhotoShelter"),
										LOC("$$$/PhotoShelter/MissingGallery/Info=Gallery: ^1 was not found at PhotoShelter but found locally, so you can delete locally or recreate and republish the photos on it. What do you wish?",
												gallery:getName()),
										LOC("$$$/PhotoShelter/MissingGallery/ActionOk=Delete"),
										LOC("$$$/PhotoShelter/MissingGallery/ActionCancel=Cancel All"),
										LOC("$$$/PhotoShelter/MissingGallery/ActionOther=Create and Publish on PhotoShelter")
								)

								if response == 'ok' then
									-- delete gallery
									LrApplication.activeCatalog():withWriteAccessDo(
											LOC("$$$/PhotoShelter/SyncTo/MissingGallery/Delete=Delete Gallery does not exists anymore at PhotoShelter"),
											function()
												gallery:delete()
											end,
											{ timeout = 5 }
									)
								end

								if response == 'other' then
									-- recreate and republish photos
									recreateGallery(gallery) -- root gallery so no new parent.
								end

								if response == 'cancel' then
									-- abort all
									serviceProgressScope:cancel()
									progressScope:cancel()
									LrErrors.throwCanceled()
								end
							else
								local success = PhotoShelterAPI.updateGallery(
										gallery:getRemoteId(),
										gallery:getName(),
										collectionSettings.Description,
										collectionSettings.ListedOnWebsite,
										collectionSettings.Visibility,
										parentCollectionID,
										authToken
								)
								if not success then
									local dialogResult = LrDialogs.confirm(
											LOC("$$$/PhotoShelter/Dialog/ForceSyncTitle=Synchronising ^1 Service Error",
													publishService:getName()),
											LOC(
													"$$$/PhotoShelter/Dialog/ForceSyncDescription=Something went wrong when we try to synchronize Gallery: ^1, do you wanna continue?",
													gallery:getName()
											)
									)
									if dialogResult == 'cancel' then
										progressScope:cancel()
										serviceProgressScope:cancel()
										LrErrors.throwCanceled()
									end
								end
								checkGalleryPhotos(gallery)
								count = count + 1
								progressScope:setPortionComplete(count, operationsCount)
							end
						end
					end

					verifyGalleries(rootGalleries)

					-- root collections verification
					local verifyCollections

					verifyCollections = function(collections)
						for _, collection in ipairs(collections) do
							local parent = collection:getParent()
							local parentCollectionID
							if parent then
								parentCollectionID = parent:getRemoteId()
							else
								parentCollectionID = "root"
							end
							local collectionSettings = collection:getCollectionSetInfoSummary().collectionSettings;
							local exists = PhotoShelterAPI.galleryOrCollectionExists(collection, authToken);

							if exists then
								local success = PhotoShelterAPI.updateCollection(
										collection:getRemoteId(),
										collection:getName(),
										collectionSettings.Description,
										collectionSettings.ListedOnWebsite,
										collectionSettings.Visibility,
										parentCollectionID,
										authToken
								)
								if not success then
									local dialogResult = LrDialogs.confirm(
											LOC("$$$/PhotoShelter/Dialog/ForceSyncTitle=Synchronising ^1 Service Error",
													publishService:getName()),
											LOC(
													"$$$/PhotoShelter/Dialog/ForceSyncDescription=Something went wrong when we try to synchronize Collection: ^1, do you wanna continue?",
													collection:getName()
											)
									)
									if dialogResult == 'cancel' then
										progressScope:cancel()
										serviceProgressScope:cancel()
										LrErrors.throwCanceled()
									end
								end

								verifyGalleries(collection:getChildCollections())
								verifyCollections(collection:getChildCollectionSets())
							else
								local response = LrDialogs.confirm(
										LOC("$$$/PhotoShelter/MissingCollection/Message=Missing Collection in PhotoShelter"),
										LOC("$$$/PhotoShelter/MissingCollection/Info=Collection: ^1 was not found at PhotoShelter but found locally, so you can delete locally or recreate and republish the photos on it. What do you wish?",
												collection:getName()),
										LOC("$$$/PhotoShelter/MissingCollection/ActionOk=Delete"),
										LOC("$$$/PhotoShelter/MissingCollection/ActionCancel=Cancel All"),
										LOC("$$$/PhotoShelter/MissingCollection/ActionOther=Create and Publish on PhotoShelter")
								)

								if response == 'ok' then
									-- delete gallery
									LrApplication.activeCatalog():withWriteAccessDo(
											LOC("$$$/PhotoShelter/SyncTo/MissingCollection/Delete=Delete Collection does not exists anymore at PhotoShelter"),
											function()
												collection:delete()
											end,
											{ timeout = 5 }
									)
								end

								if response == 'other' then
									recreateCollection(collection)
								end

								if response == 'cancel' then
									-- abort all
									serviceProgressScope:cancel()
									progressScope:cancel()
									LrErrors.throwCanceled()
								end
							end

							count = count + 1
							progressScope:setPortionComplete(count, operationsCount)
						end
					end

					verifyCollections(rootCollections)
				end
			end

			serviceCount = serviceCount + 1
			serviceProgressScope:setPortionComplete(serviceCount, #publishedServices)
		end

		serviceProgressScope:done()
		LrDialogs.showBezel('Successfully synced local copy to PhotoShelter', fadeDelay)
	else
		LrDialogs.showError("We are unable to detect any PhotoShelterSync plugin.")
	end
end)
