LrFunctionContext = import 'LrFunctionContext'
LrLocalization    = import 'LrLocalization'
LrApplication     = import 'LrApplication'
LrPathUtils       = import 'LrPathUtils'
LrFileUtils       = import 'LrFileUtils'
LrBinding         = import 'LrBinding'
LrDialogs         = import 'LrDialogs'
LrErrors          = import 'LrErrors'
LrLogger          = import 'LrLogger'
LrColor           = import 'LrColor'
LrPrefs           = import 'LrPrefs'
LrShell           = import 'LrShell'
LrTasks           = import 'LrTasks'
LrDate            = import 'LrDate'
LrHttp            = import 'LrHttp'
LrView            = import 'LrView'
LrFtp             = import 'LrFtp'
LrMD5             = import 'LrMD5'

local PROP
local PREFS

local pluginName = "PhotoShelter"
local pluginVersion = nil

require "LRLib_InterOp"
require "LRLib_TableEx"
require "LRLib_StringUtils"
require "JSON4Lua"
require "PhotoShelterHelper"

-- The namespace

PhotoShelterUpdateChecker = {}

-- Utility functions

-----
----- VERSION CHECKING
-----

local function versionIsNewerThan(old, new)

   local new_date, new_num = string.format("%.2f",new):match('^(%d+)%.(%d+)$')

   if not new_num then
      return false
   end

   if tonumber(new_date) > old.major then
      return true
   end

   if (tonumber(new_date) == old.major) and (tonumber(new_num) > old.minor) then
      return true
   end

   return false

end

local function quote(s)
   return ' "' .. s .. '" '
end

-- Does a raw version check (that is, in the current task) and updates the
-- state appropriately. If 'force' is true, the check is in response to a
-- user having pressed a [Check Now] button. Otherwise, it's an auto-check
-- that we want to be light weight.

local function doRawCheck(args)
   local PROP, force = unpack(args)

   local LrVersion = LrApplication.versionString():gsub('%s+', '+')
   local LrBuild   = LrApplication.versionTable().build
   local OS = WIN_ENV and 'W' or 'M'

   local function update(key, val)
      PROP[key] = val
      PREFS[key] = val
   end

   -- Contact server to find out the latest version number. The user's
   -- current version and a "random unique identifier" (created by
   -- math.random) so that I can count, out of curiosity, how many
   -- individual users are using each plugin. The "random unique identifier"
   -- is used instead of an IP address because it's both more and less constant
   -- (more, because it doesn't change as a user's dynamic IP changes, and
   -- less because it differs among different users behind the same proxy).

   local reg = PREFS.lib_selfupdate_serial or "0-0"
   local user, validation = string.match(reg, "(%d+)-(%d+)")
   local urlToCheck
   if PhotoShelterInfo.isLibris then
     urlToCheck = "http://www.photoshelter.com/plugin-finder?plugin=libris-lightroom"
   else
     urlToCheck = "http://www.photoshelter.com/plugin-finder?plugin=lightroom"
   end
   local reply = LrHttp.get(urlToCheck, nil, 3)

   -- Regardless of the reply, if this is an auto-check (as opposed to a
   -- user-initiated check), we want to mark that we did a check. This
   -- stops the thing from auto-checking over and over in the face of
   -- downed connectivity or a problem with the version server.

   if not force then
      PREFS.VERSION_CHECK_TIME = LrDate.currentTime()
   end

   if not reply or reply == '' then 
      -- Something went wrong. If this was a user-initiated check, let them known
      if force then
         MSG(LOC("$$$/PhotoShelter/UpdateChecker/InternalError=Internal Plugin Error"),
             LOC("$$$/PhotoShelter/UpdateChecker/VersionServer/NoContact=Could not contact the version-check server"),
             "warning")
      end
      return
   end

   -- Got a reply...

   reply = JSON4Lua.decode(reply)
   local latestVersion = reply.STABLE
   local url = reply.STABLE_LOC
   
   if not latestVersion then

      -- We got no reply, so just leave things be unless its a forced
      -- check, in which case we report the problem

      if force then
         LrDialogs.message(LOC("$$$/PhotoShelter/UpdateChecker/InternalError=Internal Plugin Error"),
                           LOC("$$$/PhotoShelter/UpdateChecker/VersionServer/BadReply=Reply from the version server not understood."),
                           "warning")
      end

      return
   end

   --
   -- We have a proper reply. We updated the check time earlier if it wasn't a forced check,
   -- but now we'll do so regardless.
   --
   PREFS.VERSION_CHECK_TIME = LrDate.currentTime()
   
   if versionIsNewerThan(pluginVersion, latestVersion) then

      -- Ours is old
      --
      -- Make sure that the latest version is still compatible with this version of LR...
      
      update('NEW_VERSION_NUMBER', latestVersion)
      showUpgradeDialog(PROP, latestVersion, url)
         
   else

      -- Ours is not old
      update('NEW_VERSION_NUMBER', nil)

      if force then
         LrDialogs.message(LOC("$$$/PhotoShelter/UpdateChecker/Version/Latest=You have the latest version (^1)", string.format("%.2f",latestVersion)), nil, "info")         
      end

   end
end

-- Called to asynchronously check the version server

function asyncVersionCheck(PROP, force)

   if not PROP.doing_version_check
  --    and
  --      (force or LrDate.currentTime() - PREFS().VERSION_CHECK_TIME > check_delta)
   then
      PROP.doing_version_check = true
      local old = PROP.version_checknow_title
      PROP.version_checknow_title = LOC("$$$/PhotoShelter/UpdateChecker/Checking=checking^.")

      LrTasks.startAsyncTask( function()
      
         doRawCheck( { PROP, force } )
         PROP.version_checknow_title = old
         PROP.doing_version_check = nil
      
      end)
   end
end

local function show_upgrade_dialog(context, PROP, PluginDialogPROP, latestVersion, url)

   local v = LrView.osFactory()
   local do_it_now_label = LOC("$$$/PhotoShelter/UpdateChecker/Button/Download=Download and Install")

   PROP.okay = LOC("$$$/PhotoShelter/UpdateChecker/Button/Cancel=Cancel")
   PROP.can_invoke = true
   PROP.abort = false
   PROP.status = ""
   PROP.completed = false

   LrDialogs.presentModalDialog {
      title = LrView.bind {
           bind_to_object = PluginDialogPROP,
           key = 'lib_selfupdate_pluginName',
      },

      cancelVerb = "< exclude >", -- this magic, unpublished setting removes the cancel button, and may not work in the future
      actionVerb = LrView.bind { bind_to_object = PROP, key = 'okay' },
      accessoryView = v:view {
         
         fill_horizontal = 1,
         margin_bottom = 2,
         margin_left = 2,
         place = "overlapping",
         bind_to_object = PROP,
         
         v:static_text {
            fill_horizontal = 1,
            bind_to_object = PROP,
            text_color = LrView.bind { key = 'status_color' },
            title = LrView.bind { key = 'status' },
            visible = LrBinding.keyIsNot("status", "")
         },
        
         v:row {
            v:push_button {
               bind_to_object = PROP,
               title = LOC("$$$/PhotoShelter/UpdateChecker/GetPlugin=Get the plugin"),
               action = function()
                  LrHttp.openUrlInBrowser(url)
               end,
               visible = LrBinding.keyEquals("status", "")
            }
         }
      },   
      contents = v:view {
         margin_bottom = 3,        
         v:static_text {
            width = 550,
            height_in_lines = -1,
            title = LrView.bind {
                 bind_to_object = PluginDialogPROP,
                 key = 'lib_selfupdate_pluginName',
                 transform = function (value, fromTable)
                 
                     local line1 = LOC("$$$/LRLib/SelfUpdateV2/Instructions/1=There is a new version of ^1 available (version ^2).", value, string.format("%.2f", latestVersion))
                     
                     return line1
               end
            }
         }
      }
   }

   PROP.abort = true -- just in case a download was running
end


function showUpgradeDialog(PluginDialogPROP, latestVersion, url)

   LrFunctionContext.postAsyncTaskWithContext("Self update dialog", function(context)
                                                 context:addFailureHandler(failureHandler)
                                                 PROP = LrBinding.makePropertyTable(context)
                                                 show_upgrade_dialog(context, PROP, PluginDialogPROP, latestVersion, url)
                                              end)

end

-- Adds a section to the plugin info provider to configure self updates, and
-- allow the user to request self update now.

function PhotoShelterUpdateChecker.PluginInfoSection(pt)

   local f = LrView.osFactory()
   return {
   
      title = LOC "$$$/PhotoShelter/UpdateChecker/PluginInf/Title=Update Checks",
      
      f:column {
         fill_horizontal = 1,
         
         f:checkbox {
            title = LOC "$$$/PhotoShelter/UpdateChecker/AutoCheck=Check for updates to this plugin when Lightroom starts",
            bind_to_object = pt,
            value = LrView.bind "lib_selfupdate_autocheck",
         },
            
         f:spacer { height = 15 },

         f:row {
            fill_horizontal = 1,
            f:push_button  {
               title =  LOC "$$$/PhotoShelter/UpdateChecker/PluginInfo/CheckNow=Check for updates now",
               place_horizontal = 0.5,
               action = function()
                  asyncVersionCheck( pt, true )
               end
            }
         }
      }
   }

end

-----
----- INITIALISATON
-----

local check_delta = 3600  -- how often to auto-check, in seconds

-- Initialises the self update functionality and checks for a later version if auto-updating is enabled.
-- When checking for plugin updates the version number is checked against the server to see whether a later version is
-- available. 

function PhotoShelterUpdateChecker.init(version)

   PREFS = LRLib_InterOp.prefs()

   pluginVersion = version

   if not PREFS.VERSION_CHECK_TIME then
      PREFS.VERSION_CHECK_TIME = 0
   end

   -- Check for updates
   
   if PREFS.lib_selfupdate_autocheck then 

      LrFunctionContext.postAsyncTaskWithContext("Self update", function(context)
                                                    PROP = LrBinding.makePropertyTable(context)
                                                    PROP.lib_selfupdate_pluginName = name
                                                    doRawCheck( { PROP, false } )
                                                 end)

   end
   
end

--- Function that must be called when selecting the plugin manager's dialog for this plugin

function PhotoShelterUpdateChecker.startDialog(pt)

   if not PREFS then PREFS = LRLib_InterOp.prefs() end

   pt.lib_selfupdate_pluginName = pluginName

   if PREFS.lib_selfupdate_autocheck ~= nil then
      pt.lib_selfupdate_autocheck = PREFS.lib_selfupdate_autocheck
   else
      pt.lib_selfupdate_autocheck = true
      PREFS.lib_selfupdate_autocheck = true
   end

end

--- Function that must be called when deselecting the plugin manager's dialog for this plugin

function PhotoShelterUpdateChecker.endDialog(pt, why)

   if not PREFS then PREFS = LRLib_InterOp.prefs() end

   PREFS.lib_selfupdate_autocheck = pt.lib_selfupdate_autocheck
    
end
